/*****************************************************************************/
/*
                                 Net.c

The networking essentials are based on DEC TCP/IP Services for OpenVMS (aka
UCX). The QIO interface was obviously chosen because of its asynchronous,
non-blocking capabilities. Although some functionalilty could have been
implemented using the BSD sockets abstraction and only the asynchronous I/O
done using the QIO I opted for working out the QIOs. It wasn't too difficult
and avoids the (slight) extra overhead of the sockets layer. The MadGoat
NETLIB port follows the basic outline of the DEC TCP/IP functionality. I
deliberately differentiated the UCX and NETLIB interfaces. It wasn't much
extra work and helps keep obvious the fact of the two APIs. NETLIB is a very
nice package! Wished I'd known about it three years ago. My thanks Matthew.

With resource wait explicitly enabled all sys$qio()s should wait until
resources become available. The only execption is ASTLM, a fundamental
requirement for this server. If SS$_EXQUOTA is returned from a sys$qio() (i.e.
without wait for completion, and therefore depending on AST delivery) then
EXIT THE SERVER with error status!

Some underlying NETLIB supporting TCP/IP packages do not support host-name
resolution at AST delivery level, which this module uses, and hence CANNOT
SUPPORT DNS LOOKUP from v4.4 onwards! (See the NETLIB documentation.)

The 'ServiceStruct' structure has two groups of fields and performs two basic
roles. First, it provides a linked list which can be traversed to determine
from the destination IP address of a request which host was specified in a
(multi-homed) request. Second, for each unique port specified for the same
server the client components of the structure are used for the accept()ing of
requests.  Note that there may be more 'service' structures in the list than
actual sockets created (and client components used) because only one is
allocated against a given port even though the services may specify multiple
host names using that port.

The server identity is derived from the 'ServerHostName' (gethostname()) name
and the 'ServerPort' (which is the first port specified as a service, or the
port specified in the configuration or /PORT= qualifier).

SERVICE PARAMETERS
------------------
The following syntax and values are available for both the [service]
configuration parameter and /SERVICE= qualifier.  Also see SESOLA.C for SSL
parameter processing.

  http:                      system host name, HTTP service
  https:                     system host name, SSL service
  http:port                  system host name, supplied port, HTTP service
  https:port                 system host name, supplied SSL service
  http://host.domain         specified host name, pot 80, HTTP service
  https://host.domain        specified host name port 443, SSL service
  http://host.domain:port    specified host name and port, HTTP service
  https://host.domain:port   specified host name and port, SSL service
  ;cert=                     SSL service's non-default server certificate
  ;key=                      SSL service's non-default server RSA private key
  ;cipher=                   SSL service's specific, non-default cipher list
  ;proxy                     service provide proxy access
  ;connect                   service provides SSL connect proxy access
  ;chain=                    service chains to up-stream proxy service

For example:

  [service]
  http://host.name:8080;proxy;connect
  https://host.name:443;cert=HT_ROOT:[LOCAL]SITE.PEM


VERSION HISTORY
---------------
19-SEP-99  MGD  virtual host modifications
18-AUG-99  MGD  bugfix; parsing certificate/key from service
11-JUN-99  MGD  bugfix; NetTooBusy() sys$fao() charset,
                bugfix; NetDummyReadAst() UCX IOsb by reference
26-MAY-99  MGD  bugfix; NetShutdownSocket() AST handling
04-APR-99  MGD  provide HTTP/0.9 header absorbtion (finally!)
15-JAN-99  MGD  changed AST delivery algorithm
07-NOV-98  MGD  WATCH facility
24-OCT-98  MGD  allow SSL certificate to be specified per-service
08-APR-98  MGD  allow legitimate connects to be CANCELed in NetAcceptAst()
19-JAN-98  MGD  redesigned NetWriteBuffer()
27-NOV-97  MGD  hmmm, rationalized AST processing by making network
                writes always deliver an AST (implicit or explicit)
                (causing ACCVIOs for the last couple of versions)
25-OCT-97  MGD  changes to MsgFor() function
06-SEP-97  MGD  multi-homed hosts and multi-port services
30-AUG-97  MGD  bugfix; get server host name before starting logging
                (woops, problem introduced with NETLIB)
09-AUG-97  MGD  message database
23-JUL-97  MGD  HTTPD v4.3, MultiNet dropped, using the NETLIB Library
01-FEB-97  MGD  HTTPd version 4
27-SEP-96  MGD  add dummy read for early error reporting
12-APR-96  MGD  observed Multinet disconnection/zero-byte behaviour
                (request now aborts if Multinet returns zero bytes)
03-JAN-96  MGD  support for both DEC TCP/IP Services and TGV MultiNet
01-DEC-95  MGD  NetWriteBuffered() for improving network I/O
20-DEC-94  MGD  multi-threaded version
20-JUN-94  MGD  single-threaded version
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <dvidef.h>
#include <iodef.h>
#include <psldef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application-related header files */
#include "wasd.h"

#define WASD_MODULE "NET"

/******************/
/* global storage */
/******************/

boolean NetMultiHomedHost;

int  QueueLength = 5,
     ServiceCount,
     ServerHostNameLength;
     
char  ServerHostName [128],
      ServerHostPort [128],
      Services [256];

struct ServiceStruct  *ServiceListHead,
                      *ServiceListTail;

unsigned long  ServiceHashTableMask,
               ServiceHashTableEntries;
struct ServiceStruct  **ServiceHashTable;

char  ErrorServiceList [] = "service list confusing";

#ifdef IP_UCX

char HttpdIpPackage [] = "Digital-TCPIP";

$DESCRIPTOR (InetDeviceDsc, "UCX$DEVICE");

int  OptionEnabled = 1;

struct {
   int  l_onoff;
   int  l_linger;
} LingerSeconds = { 1, 10 };

struct {
   unsigned short  Length;
   unsigned short  Parameter;
   char  *Address;
} ReuseAddress =
   { sizeof(OptionEnabled), UCX$C_REUSEADDR, &OptionEnabled },
  ReuseAddressSocketOption =
   { sizeof(ReuseAddress), UCX$C_SOCKOPT, &ReuseAddress },
  LingerOnClose =
   { sizeof(LingerSeconds), UCX$C_LINGER, &LingerSeconds },
  LingerOnCloseOption =
   { sizeof(LingerOnClose), UCX$C_SOCKOPT, &LingerOnClose };

unsigned short  TcpSocket [2] =
   { UCX$C_TCP, INET_PROTYP$C_STREAM };

#endif /* IP_UCX */


#ifdef IP_NETLIB

char HttpdIpPackage [] = "MadGoat-NETLIB";

unsigned long  NetLibAddressListSize = NETLIB_ADDRESS_LIST_SIZE,
               NetLibAddressListCount,
               NetLibSocketType = NETLIB_K_TYPE_STREAM,
               NetLibSocketFamily = NETLIB_K_AF_INET,
               NetLibLevel = NETLIB_K_LEVEL_SOCKET,
               NetLibReUseOption = NETLIB_K_OPTION_REUSEADDR,
               NetLibShutdownRecv = NETLIB_K_SHUTDOWN_RECEIVER,
               NetLibShutdownBoth = NETLIB_K_SHUTDOWN_BOTH,
               NetLibOptionOn = 1,
               NetLibSizeofINADDRDEF = sizeof(struct INADDRDEF),
               NetLibSizeofSINDEF = sizeof(struct SINDEF),
               NetLibSizeofOptionOn = sizeof(unsigned long);

#endif /* IP_NETLIB */


/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern boolean  ControlExitRequested;
extern boolean  ControlRestartRequested;
extern boolean  ProtocolHttpsAvailable;
extern boolean  ProtocolHttpsConfigured;
extern int  AdminWatchEnabled;
extern int  AdminWatchItem;
extern int  CurrentConnectCount;
extern int  MaxConcurrentConnections;
extern int  OutputBufferSize;
extern int  ServerPort;
extern char  CliLogFileName[];
extern char  ControlBuffer[];
extern char  ErrorSanityCheck[];
extern char  Utility[];
extern char  *AcceptHostsPtr;
extern char  *RejectHostsPtr;
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;
extern struct MsgStruct  Msgs;

/*****************************************************************************/
/*
Get local host (server) name.
*/ 

NetGetServerHostName ()

{
   register char  *cptr;

   int  status;
   unsigned short  Length;

#ifdef IP_NETLIB
   $DESCRIPTOR (ServerHostNameDsc, ServerHostName);
#endif

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetGetServerHostName()\n");

#ifdef IP_UCX

   if (gethostname (ServerHostName, sizeof(ServerHostName)) != 0)
      ErrorExitVmsStatus (0, strerror(errno), FI_LI);

#endif

#ifdef IP_NETLIB

   status = netlib_get_hostname (&ServerHostNameDsc, &Length);
   if (Debug) fprintf (stdout, "netlib_get_hostname() %%X%08.08X\n", status);
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "netlib_get_hostname()", FI_LI);
   ServerHostName[Length] = '\0';

#endif

   /* looks better if its all in lower case (host name is sometimes upper) */
   for (cptr = ServerHostName; *cptr; cptr++) *cptr = tolower(*cptr);
   ServerHostNameLength = cptr - ServerHostName;
   if (Debug) fprintf (stdout, "|%s|\n", ServerHostName);
}

/*****************************************************************************/
/*
Parse the 'Service's parameter to determine which host(s)/port(s) services need
to be provided for.  This is a comma-separated list with no white-space.  The
service information may comprise of the following optional components:
"[scheme://][hostname][:port][;proxy][;chain=server][;certificate=file]".
*/ 

NetSetService ()

{
   register char  *cptr, *sptr, *zptr;
   register struct ServiceStruct *svptr;

   boolean  ProxyService,
            ProxyServiceFileCache,
            ConnectService,
            SetServerPort;
   int  status,
        Length,
        PortNumber,
        ProxyChainPortNumber,
        RequestScheme;
   char *RequestSchemeNamePtr;
   char  HostName [sizeof(svptr->ServerHostName)],
         ProxyChainHostName [sizeof(svptr->ServerHostName)],
         SeSoLaCert [sizeof(svptr->SeSoLaCert)],
         SeSoLaCipherList [sizeof(svptr->SeSoLaCipherList)],
         SeSoLaKey [sizeof(svptr->SeSoLaKey)];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetSetService()\n");

   SetServerPort = false;

   if (Services[0])
      cptr = Services;
   else
   if (Config.ServicePtr == NULL)
      sprintf (cptr = Services, "%s:%d", ServerHostName, ServerPort);
   else
      cptr = Config.ServicePtr;

   /******************************/
   /* parse comma-separated list */
   /******************************/

   while (*cptr)
   {
      while (*cptr && (ISLWS(*cptr) || *cptr == ',')) cptr++;
      if (!*cptr) break;

      ProxyService = ProxyServiceFileCache = ConnectService = false;
      RequestScheme = PortNumber = ProxyChainPortNumber = 0;
      RequestSchemeNamePtr = NULL;
      HostName[0] = SeSoLaCert[0] = SeSoLaCipherList[0] = SeSoLaKey[0] =
         ProxyChainHostName[0] = '\0';

      if (Debug) fprintf (stdout, "|%s|\n", cptr);

      /*****************************/
      /* service scheme (protocol) */
      /*****************************/

      if (isalpha(*cptr))
      {
         if (strsame (cptr, "http:", 5))
         {
            RequestScheme = SCHEME_HTTP;
            RequestSchemeNamePtr = "http:";
            cptr += 5;
            while (*cptr == '/') cptr++;
         }
         else
         if (strsame (cptr, "https:", 6))
         {
            RequestScheme = SCHEME_HTTPS;
            RequestSchemeNamePtr = "https:";
            if (ProtocolHttpsAvailable) ProtocolHttpsConfigured = true;
            cptr += 6;
            while (*cptr == '/') cptr++;
         }
         else
         if (strsame (cptr, "connect:", 8))
         {
            RequestScheme = SCHEME_CONNECT;
            RequestSchemeNamePtr = "connect:";
            cptr += 8;
            while (*cptr == '/') cptr++;
         }
      }

      /**********************/
      /* host name and port */
      /**********************/

      Length = NetParseHostNamePort (cptr, HostName, sizeof(HostName),
                                     &PortNumber);
      if (Length < 0) ErrorExitVmsStatus (0, ErrorServiceList, FI_LI);
      cptr += Length;

      /****************************/
      /* other service parameters */
      /****************************/

      if (*cptr == ';')
      {
         while (*cptr == ';')
         {
            if (strsame (cptr, ";cache", 6))
            {
               ProxyServiceFileCache = true;
               cptr += 6;
            }
            else
            if (strsame (cptr, ";proxy", 6))
            {
               ProxyService = true;
               cptr += 6;
            }
            else
            if (strsame (cptr, ";chain=", 7))
            {
               cptr += 7;
               Length = NetParseHostNamePort (cptr, ProxyChainHostName,
                                              sizeof(ProxyChainHostName),
                                              &ProxyChainPortNumber);
               if (Length < 0) ErrorExitVmsStatus (0, ErrorServiceList, FI_LI);
               cptr += Length;
            }
            else
            if (strsame (cptr, ";connect", 8))
            {
               ConnectService = true;
               cptr += 8;
            }
            else
            if (strsame (cptr, ";cipher=", 5))
            {
               while (*cptr && *cptr != '=') cptr++;
               if (*cptr) cptr++;
               zptr = (sptr = SeSoLaCipherList) + sizeof(SeSoLaCipherList);
               while (*cptr && !ISLWS(*cptr) && *cptr != ',' && sptr < zptr)
                  *sptr++ = *cptr++;
               if (sptr >= zptr)
                  ErrorExitVmsStatus (0, ErrorServiceList, FI_LI);
               *sptr = '\0';
            }
            else
            if (strsame (cptr, ";key=", 5))
            {
               while (*cptr && *cptr != '=') cptr++;
               if (*cptr) cptr++;
               zptr = (sptr = SeSoLaKey) + sizeof(SeSoLaKey);
               while (*cptr && !ISLWS(*cptr) && *cptr != ',' && sptr < zptr)
                  *sptr++ = toupper(*cptr++);
               if (sptr >= zptr)
                  ErrorExitVmsStatus (0, ErrorServiceList, FI_LI);
               *sptr = '\0';
            }
            else
            {
               /* no keyword is BACKWARD-COMPATIBLE SSL certificate location */
               if (strsame (cptr, ";cert=", 6))
               {
                  while (*cptr && *cptr != '=') cptr++;
                  if (*cptr) cptr++;
               }
               else
               {
                  cptr++;
                  for (sptr = cptr; *sptr && *sptr != '='; sptr++);
                  if (*sptr) ErrorExitVmsStatus (0, ErrorServiceList, FI_LI);
               }
               zptr = (sptr = SeSoLaCert) + sizeof(SeSoLaCert);
               while (*cptr && !ISLWS(*cptr) &&
                      *cptr != ',' && *cptr != ';' && sptr < zptr)
                  *sptr++ = toupper(*cptr++);
               if (sptr >= zptr)
                  ErrorExitVmsStatus (0, ErrorServiceList, FI_LI);
               *sptr = '\0';
            }
         }
      }

      /*****************************/
      /* check service information */
      /*****************************/

      if (!RequestScheme) RequestScheme = SCHEME_HTTP;
      if (RequestScheme == SCHEME_HTTP)
         RequestSchemeNamePtr = "http:";
      else
         RequestSchemeNamePtr = "https:";
      if (!HostName[0]) strcpy (HostName, ServerHostName);
      if (!PortNumber)
      {
         if (RequestScheme == SCHEME_HTTP)
            PortNumber = DEFAULT_HTTP_PORT;
         else
            PortNumber = DEFAULT_HTTPS_PORT;
      }
      if (Debug) fprintf (stdout, "|%s:%d|\n", HostName, PortNumber);

      if (!ProtocolHttpsAvailable && RequestScheme == SCHEME_HTTPS)
      {
         fprintf (stdout,
            "%%%s-W-SERVICE, SSL not available, ignored\n \\%s//%s:%d\\\n",
            Utility, RequestSchemeNamePtr, HostName, PortNumber);
         continue;
      }

      /***********************************/
      /* add service information to list */
      /***********************************/

      /* allocate memory for a new service structure */
      svptr = VmGet (sizeof (struct ServiceStruct));
      if (Debug) fprintf (stdout, "svptr: %d\n", svptr);

      /* add it to the linked list */
      if (ServiceListHead == NULL)
         ServiceListHead = ServiceListTail = svptr;
      else
      {
         ServiceListTail->NextPtr = svptr;
         ServiceListTail = svptr;
      }
      svptr->NextPtr = NULL;
      ServiceCount++;

      status = NetHostNameLookup (HostName,
                                  &svptr->ServerHostName,
                                  &svptr->ServerHostPort,
                                  NULL,
                                  NULL,
                                  &svptr->ServerIpAddressString,
                                  &svptr->ServerIpAddress);
      if (Debug) fprintf (stdout, "NetHostNameLookup() %%X%08.08X\n", status);
      if (VMSnok (status))
         ErrorExitVmsStatus (status, HostName, FI_LI);

      svptr->RequestScheme = RequestScheme;
      svptr->RequestSchemeNamePtr = RequestSchemeNamePtr;
      svptr->ServerPort = PortNumber;
      svptr->ServerHostNameLength = strlen(svptr->ServerHostName);
      strcpy (svptr->SeSoLaCert, SeSoLaCert);
      strcpy (svptr->SeSoLaCipherList, SeSoLaCipherList);
      strcpy (svptr->SeSoLaKey, SeSoLaKey);
      sprintf (svptr->ServerPortString, "%d", svptr->ServerPort);
      sprintf (svptr->ServerHostPort, "%s:%s",
               svptr->ServerHostName, svptr->ServerPortString);

      if (Debug)
         fprintf (stdout, "%08.08X|%s|%d|%s|%s|%d|%s|%s|%s|%s|\n",
                  svptr->ServerIpAddress,
                  svptr->ServerIpAddressString,
                  svptr->RequestScheme, svptr->RequestSchemeNamePtr,
                  svptr->ServerHostName, svptr->ServerPort,
                  svptr->ServerHostPort, svptr->SeSoLaCert,
                  svptr->SeSoLaKey, svptr->SeSoLaCipherList);

      /* process name, etc., generated from the primary port */
      if (!SetServerPort) SetServerPort = ServerPort = svptr->ServerPort;

      /*****************************/
      /* proxy service information */
      /*****************************/

      svptr->ProxyService = ProxyService;
      svptr->ConnectService = ConnectService;
      svptr->ProxyFileCacheEnabled = ProxyServiceFileCache;

      if (ProxyChainHostName[0])
      {
         status = NetHostNameLookup (ProxyChainHostName,
                                     &svptr->ProxyChainHostName,
                                     &svptr->ProxyChainHostPort,
                                     NULL,
                                     NULL,
                                     &svptr->ProxyChainIpAddressString,
                                     &svptr->ProxyChainIpAddress);
         if (Debug)
            fprintf (stdout, "NetHostNameLookup() %%X%08.08X\n", status);
         if (VMSnok (status))
            ErrorExitVmsStatus (status, ProxyChainHostName, FI_LI);

         if (!ProxyChainPortNumber)
            ProxyChainPortNumber = DEFAULT_HTTP_PROXY_PORT;

         svptr->ProxyChainPort = ProxyChainPortNumber;
         svptr->ProxyChainHostNameLength =
            strlen(svptr->ProxyChainHostName);
         sprintf (svptr->ProxyChainPortString, "%d",
                  svptr->ProxyChainPort);
         sprintf (svptr->ProxyChainHostPort, "%s:%s",
                  svptr->ProxyChainHostName,
                  svptr->ProxyChainPortString);
      }

      /********************/
      /* end service loop */
      /********************/

      if (Debug) fprintf (stdout, "cptr |%s|\n", cptr);
      if (*cptr && *cptr != ',')
         ErrorExitVmsStatus (0, ErrorServiceList, FI_LI);
   }

   if (!ServiceCount)
      ErrorExitVmsStatus (0, "No service specified!", FI_LI);

   /* the host and port for the "official" server */
   sprintf (ServerHostPort, "%s:%d", ServerHostName, ServerPort);
   if (Debug) fprintf (stdout, "|%s|\n", ServerHostPort);
}

/*****************************************************************************/
/*
Parse a host name and integer port number from a string. Any comma or
semi-colon will terminate the host name/port.  Return the number of characters
scanned from the string, or -1 to indicate a fatal error.
*/ 

int NetParseHostNamePort
(
char *String,
char *HostNamePtr,
int SizeOfHostName,
int *PortNumberPtr
)
{
   register char  *cptr, *sptr, *zptr;

   boolean  DotDecimal;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetParseHostNamePort() |%s|\n", String);

   cptr = String;
   DotDecimal = false;
   if (isdigit(*cptr))
   {
      /* check it's not a dot-decimal address */
      while (*cptr && !ISLWS(*cptr) &&
             *cptr != '.' && *cptr != ',' && *cptr != ':' && *cptr != ';')
         cptr++;
      if (*cptr == '.') DotDecimal = true;
      cptr = String;
   }
   else
   if (isalpha(*cptr))
      DotDecimal = true;

   if (DotDecimal)
   {
      /* IP host name */
      zptr = (sptr = HostNamePtr) + SizeOfHostName;
      while (*cptr && !ISLWS(*cptr) &&
             *cptr != ',' && *cptr != ':' && *cptr != ';' &&
             sptr < zptr)
         *sptr++ = *cptr++;
      if (sptr >= zptr) return (-1);
      *sptr = '\0';
      if (*cptr == ':') cptr++;
   }

   if (isdigit(*cptr))
   {
      /* IP port */
      *PortNumberPtr = atol(cptr);
      while (*cptr && isdigit(*cptr)) cptr++;
   }

   return (cptr - String);
}

/*****************************************************************************/
/*
It is only necessary to create one socket for each port port on the same
server, even if that port is specified against a different (multi-home) host
name because the accept()ing socket can be interrogated to determine which
host had been specified.  Then in a infinite loop accept connections from
clients.

A hash table is used to expedite access when searching the service list.  The
first entry with a particular hash value is entered into the table.  If a hash
collision occurs the subsequent entry will be further down the list anyway and
so a search from that point and down the linked list can be performed.
*/ 

NetCreateService ()

{
   register char  *cptr, *sptr;
   register struct ServiceStruct  *svptr, *tsvptr;

   int  cnt,
        status,
        PortNumber;
   unsigned short  Length;
   char  *ProtocolPtr;
   
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetCreateService()\n");

   /* hash table needs to be a power of 2 (i.e. 64, 128, 256, 512 ...) */
   ServiceHashTableEntries = ServiceCount;
   for (cnt = 0; ServiceHashTableEntries; cnt++)
      ServiceHashTableEntries >>= 1;
   ServiceHashTableEntries = 1;
   while (--cnt) ServiceHashTableEntries <<= 1;
   /* bigger than it needs to be, gives a better hash spread and hit rate */
   ServiceHashTableEntries <<= 3;

   /* e.g. cache table size of 512 becomes a mask of 0x1ff */
   ServiceHashTableMask = ServiceHashTableEntries - 1;

   if (Debug)
      fprintf (stdout, "HashTableEntries: %d HashTableMask: %08.08X\n",
               ServiceHashTableEntries, ServiceHashTableMask);

   ServiceHashTable =
      VmGet (sizeof(struct ServiceStruct*)*ServiceHashTableEntries);

   /********************************/
   /* process the list of services */
   /********************************/

   for (svptr = ServiceListHead; svptr != NULL; svptr = svptr->NextPtr)
   {
      fprintf (stdout, "%%%s-I-SERVICE, %s//%s\n",
               Utility, svptr->RequestSchemeNamePtr, svptr->ServerHostPort);

      /* generate the name-only hash that will used for quick host compares */
      svptr->ServerHostNameHashValue = NetHostNameHash (svptr->ServerHostName);

      /* generate the name-port hash, if none set update the service table */
      svptr->ServerHostPortHashValue =
         NetHostPortHash (svptr->ServerHostPort, "");
      if (ServiceHashTable[svptr->ServerHostPortHashValue] == NULL)
         ServiceHashTable[svptr->ServerHostPortHashValue] = svptr;

      /**************************************/
      /* check for previous service on port */
      /**************************************/

      for (tsvptr = ServiceListHead; tsvptr != NULL; tsvptr = tsvptr->NextPtr)
      {
         /* skip over this entry */
         if (svptr == tsvptr)
         {
            /* got this far in the list, can't be a socket for this port! */
            tsvptr = NULL;
            break;
         }
         /* break if already a socket created for this port */
         if (svptr->ServerPort == tsvptr->ServerPort) break;
      }

      /* if another service already on the same port then just continue */
      if (tsvptr != NULL) continue;

      /*************************/
      /* create service socket */
      /*************************/

#ifdef IP_UCX

      /* assign a channel to the internet template device */
      if (VMSnok (status =
          sys$assign (&InetDeviceDsc, &svptr->ServerChannel, 0, 0)))
         ErrorExitVmsStatus (status, "sys$assign()", FI_LI);

      /* make the channel a TCP, connection-oriented socket */

      if (Debug) fprintf (stdout, "sys$qiow() IO$_SETMODE\n");
      status = sys$qiow (0, svptr->ServerChannel, IO$_SETMODE,
                         &svptr->ServerIOsb, 0, 0,
                         &TcpSocket, 0, 0, 0, &ReuseAddressSocketOption, 0);

      if (Debug)
         fprintf (stdout, "sys$qiow() %%X%08.08X IOsb %%X%08.08X\n",
                  status, svptr->ServerIOsb.Status);
      if (VMSok (status) && VMSnok (svptr->ServerIOsb.Status))
         status = svptr->ServerIOsb.Status;
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "sys$qiow()", FI_LI);

#endif

#ifdef IP_NETLIB

      if (Debug) fprintf (stdout, "netlib_socket()\n");
      status = netlib_socket (&svptr->ServerNetLibSocket, &NetLibSocketType,
                              &NetLibSocketFamily);
      if (Debug) fprintf (stdout, "netlib_socket() %%X%08.08X\n", status);
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "netlib_socket()", FI_LI);

      status = netlib_setsockopt (&svptr->ServerNetLibSocket, &NetLibLevel,
                                  &NetLibReUseOption,
                                  &NetLibOptionOn, &NetLibSizeofOptionOn,
                                  &svptr->ServerIOsb, 0, 0);
      if (Debug)
         fprintf (stdout, "netlib_setsockopt() %%X%08.08X IOsb %%X%08.08X\n",
                  status, svptr->ServerIOsb.Status);
      if (VMSok (status) && VMSnok (svptr->ServerIOsb.Status))
         status = svptr->ServerIOsb.Status;
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "netlib_setsockopt()", FI_LI);

#endif

      /**************************/
      /* bind the server socket */
      /**************************/

#ifdef IP_UCX

      svptr->ServerSocketNameItem.Length = sizeof(svptr->ServerSocketName);
      svptr->ServerSocketNameItem.Address = &svptr->ServerSocketName;

      svptr->ServerSocketName.sin_family = AF_INET;
      svptr->ServerSocketName.sin_port = htons (svptr->ServerPort);
      svptr->ServerSocketName.sin_addr.s_addr = INADDR_ANY;

      if (Debug) fprintf (stdout, "sys$qiow() IO$_SETMODE\n");
      status = sys$qiow (0, svptr->ServerChannel, IO$_SETMODE,
                         &svptr->ServerIOsb, 0, 0,
                         0, 0, &svptr->ServerSocketNameItem, QueueLength, 0, 0);
      if (Debug)
         fprintf (stdout, "sys$qiow() %%X%08.08X IOsb %%X%08.08X\n",
                  status, svptr->ServerIOsb.Status);
      if (VMSok (status) && VMSnok (svptr->ServerIOsb.Status))
         status = svptr->ServerIOsb.Status;
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "sys$qiow()", FI_LI);

#endif

#ifdef IP_NETLIB

      svptr->ServerNetLibName.sin_w_family = NETLIB_K_AF_INET;
      svptr->ServerNetLibName.sin_w_port =
         netlib_hton_word (&svptr->ServerPort);
      svptr->ServerNetLibName.sin_x_addr.inaddr_l_addr = 0;  /* INADDR_ANY */

      if (Debug) fprintf (stdout, "netlib_bind()\n");
      status = netlib_bind (&svptr->ServerNetLibSocket,
                            &svptr->ServerNetLibName,
                            &NetLibSizeofSINDEF, &svptr->ServerIOsb, 0, 0);
      if (Debug)
         fprintf (stdout, "netlib_bind %%X%08.08X IOsb %%X%08.08X\n",
                  status, svptr->ServerIOsb.Status);
      if (VMSok (status) && VMSnok (svptr->ServerIOsb.Status))
         status = svptr->ServerIOsb.Status;
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "netlib_bind()", FI_LI);

      if (Debug) fprintf (stdout, "netlib_listen()\n");
      status = netlib_listen (&svptr->ServerNetLibSocket, &QueueLength,
                              &svptr->ServerIOsb, 0, 0);
      if (Debug)
         fprintf (stdout, "netlib_listen %%X%08.08X IOsb %%X%08.08X\n",
                  status, svptr->ServerIOsb.Status);
      if (VMSok (status) && VMSnok (svptr->ServerIOsb.Status))
         status = svptr->ServerIOsb.Status;
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "netlib_listen()", FI_LI);

#endif

      /******************************/
      /* check for multi-homed host */
      /******************************/

      for (tsvptr = ServiceListHead; tsvptr != NULL; tsvptr = tsvptr->NextPtr)
      {
         if (svptr->ServerIpAddress != tsvptr->ServerIpAddress)
         {
            NetMultiHomedHost = true;
            break;
         }
      }

      /****************************/
      /* listen for first request */
      /****************************/

      NetAccept (svptr);
   }

   if (Debug) fprintf (stdout, "NetMultiHomedHost: %d\n", NetMultiHomedHost); 

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Scan through the list of services returning true if one is found matching the
supplied IP address (and optionally port), return false otherwise.  If a port
component is not supplied with the parameter it is considered to match any/all
ports.
*/ 

boolean NetServiceConfigured (char *HostNamePort)

{
   register char  *cptr, *sptr;
   register struct ServiceStruct  *svptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetServiceConfigured() |%s|\n", HostNamePort);

   for (svptr = ServiceListHead; svptr != NULL; svptr = svptr->NextPtr)
   {
      if (Debug) fprintf (stdout, "|%s|\n", svptr->ServerHostPort);
      cptr = HostNamePort;
      sptr = svptr->ServerHostPort;
      /* compare the host name portion */
      while (*cptr && *cptr != ':' && *sptr && *sptr != ':')
      {
         if (tolower(*cptr) != tolower(*sptr)) break;
         cptr++;
         sptr++;
      }
      /* continue if the host name portion did not match */
      if ((*cptr && *cptr != ':') || *sptr != ':') continue;
      /* return true if the parameter did not contain a specific port */
      if (!*cptr)
      {
         if (Debug) fprintf (stdout, "match\n");
         return (true);
      }
      cptr++;
      sptr++;
      /* compare the port portion */
      while (*cptr && *sptr)
      {
         if (*cptr != *sptr) break;
         cptr++;
         sptr++;
      }
      /* return true if the ports matched */
      if (!*cptr && !*sptr)
      {
         if (Debug) fprintf (stdout, "match\n");
         return (true);
      }
   }

   if (Debug) fprintf (stdout, "no match\n");
   return (false);
}

/*****************************************************************************/
/*
Generate and return a hash value for a host name and port as a single
colon-delimited string.  If the host-port string does not contain a port
component then use the additional parameter which must contain the string
equivalent of the port number.  The generated value may be used to index into
the service list hash table.
*/ 

#pragma inline(NetHostPortHash)

unsigned long NetHostPortHash
(
char *HostPortString,
char *PortString
)
{
   register unsigned long  HashValue;
   register char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetHostPortHash() |%s|\n", HostPortString);

   /* hash in the host component (forced to lower case) */
   HashValue = 1;
   for (cptr = HostPortString; *cptr && *cptr != ':'; cptr++)
      HashValue = ((tolower(*cptr)*541)^(HashValue*7)) & ServiceHashTableMask;
   if (!*cptr)
   {
      /* host-port did not include a port, used the additional parameter */
      HashValue = ((':'*541)^(HashValue*7)) & ServiceHashTableMask;
      cptr = PortString;
   }
   /* hash in the port component */
   while (*cptr)
   {
      HashValue = ((*cptr*541)^(HashValue*7)) & ServiceHashTableMask;
      cptr++;
   }
   if (Debug) fprintf (stdout, "HashValue: %d\n", HashValue);
   return (HashValue);
}

/*****************************************************************************/
/*
Generate and return a hash value for a host name (excludes any colon-delimited
port number).  THIS HASH VALUE CANNOT BE USED TO INDEX INTO ANY TABLE!!  It is
just for quick elimination of a host string before undergoing a full string
comparison.
*/ 

#pragma inline(NetHostNameHash)

unsigned long NetHostNameHash (char *HostNameString)

{
   register unsigned long  HashValue;
   register char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetHostNameHash() |%s|\n", HostNameString);

   /* generate lower-case hash */
   HashValue = 1;
   for (cptr = HostNameString; *cptr && *cptr != ':'; cptr++)
      HashValue = ((tolower(*cptr)*541)^(HashValue*7)) & 0x0ffff;
   if (Debug) fprintf (stdout, "HashValue: %d\n", HashValue);
   return (HashValue);
}

/*****************************************************************************/
/*
Find a service' host name and port that matches the request's "Host:" request
header line.  Point the request's service pointer at the virtual service's
structure is found.  Return true if a matching service was found, false if not.
A hash table is used to expedite access into the list.  This first entry with
a particular hash value is entered into the table.  If the hash entry does not
match by string comparison then the search just continues down the list.
*/ 

boolean NetVirtualService (struct RequestStruct *rqptr)

{
   register char  *sptr1, *sptr2;
   register struct ServiceStruct  *svptr;

   unsigned long  HashValue;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetVirtualServer() |%s|\n", rqptr->HttpHostPtr);

   if (rqptr->HttpHostPtr == NULL || !rqptr->HttpHostPtr[0])
   {
      /* just use whatever service the request connected to */
      if (Debug) fprintf (stdout, "no \"Host:\"\n");

      if (rqptr->AdminWatchItem &&
          (AdminWatchEnabled & ADMIN_WATCH_CONNECT))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_CONNECT,
                         "ACTUAL !AZ", rqptr->ServicePtr->ServerHostPort); 

      return (true);
   }

   /* not worth the overhead of hash generation for small number of services */
   if (ServiceCount > 4)
   {
      /* generate a hash value for the "Host:" name/port */
      HashValue = NetHostPortHash (rqptr->HttpHostPtr,
                                   rqptr->ServicePtr->ServerPortString);

      /* begin searching the list at the hash table entry */
      svptr = ServiceHashTable[HashValue];
   }
   else
   {
      /* begin with the start of the list */
      svptr = ServiceListHead;
   }

   for (/* above */; svptr != NULL; svptr = svptr->NextPtr)
   {
      if (Debug) fprintf (stdout, "|%s|\n", svptr->ServerHostPort);

      sptr1 = rqptr->HttpHostPtr;
      sptr2 = svptr->ServerHostPort;

      /* compare host names */
      while (*sptr1 && *sptr1 != ':' && *sptr2 && *sptr2 != ':')
         if (tolower (*sptr1++) != tolower (*sptr2++)) break;

      if (Debug) fprintf (stdout, "|%s|%s|\n", sptr1, sptr2);
      if (*sptr1 == ':' && *sptr2 == ':')
      {
         /* the "Host:" contained a port number */
         sptr1++;
         sptr2++;
      }
      else
      if (!*sptr1 && *sptr2 == ':')
      {
         /* the "Host:" did not contain a port number, use the service's */
         sptr1 = rqptr->ServicePtr->ServerPortString;
         sptr2++;
      }
      else
      {
         if (Debug) fprintf (stdout, "no match\n");
         continue;
      }

      /* compare port number strings */
      if (Debug) fprintf (stdout, "|%s|%s|\n", sptr1, sptr2);
      while (*sptr1 && *sptr2) if (*sptr1++ != *sptr2++) break;
      if (*sptr1 || *sptr2)
      {
         if (Debug) fprintf (stdout, "no match\n");
         continue;
      }

      if (Debug) fprintf (stdout, "match |%s|\n", svptr->ServerHostPort);

      if (rqptr->AdminWatchItem &&
          (AdminWatchEnabled & ADMIN_WATCH_CONNECT))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_CONNECT,
                         "VIRTUAL !AZ", svptr->ServerHostPort); 

      /* change the request's service pointer to it's virtual equivalent */
      rqptr->ServicePtr = svptr;

      return (true);
   }

   if (Debug) fprintf (stdout, "end of services!\n");
   return (false);
}

/*****************************************************************************/
/*
Compare the 'ActualHostPort' to the 'VirtualHostPort'.  Host name components
MUST match (case insensitive).  If the virtual specification contains a port
component then both must have port components and both must match.  Returns
true if they do, false if they don't.
*/ 

boolean NetThisVirtualService
(
char *ActualHostPort,
char *VirtualHostPort
)
{
   register char  *sptr1, *sptr2;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetThisVirtualService() |%s|%s|\n",
               ActualHostPort, VirtualHostPort);

   sptr1 = ActualHostPort;
   sptr2 = VirtualHostPort;

   /* compare host names */
   while (*sptr1 && *sptr1 != ':' && *sptr2 && *sptr2 != ':')
      if (tolower (*sptr1++) != tolower (*sptr2++)) return (false);
   if (Debug) fprintf (stdout, "|%s|%s|\n", sptr1, sptr2);

   if (!*sptr2 && (!*sptr1 || *sptr1 == ':'))
   {
      /* virtual does not include a port, match any */
      return (true);
   }

   /* at this stage both should contain port components */
   if (*sptr1 != ':' || *sptr2 != ':') return (false);

   /* compare the port numbers */
   sptr1++;
   sptr2++;
   while (*sptr1 && *sptr2)
      if (*sptr1++ != *sptr2++) return (false);
   if (*sptr1 || *sptr2) return (false);
   return (true);
}

/*****************************************************************************/
/*
Queue an accept() to the listening server socket.
*/ 

NetAccept (struct ServiceStruct *svptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetAccept() %d\n", svptr);

   if (ControlExitRequested || ControlRestartRequested)
   {
      /* server exiting or restarting, no new accepts, just return */
      return;
   }

#ifdef IP_UCX

   /* assign a channel to the internet template device */
   if (VMSnok (status =
       sys$assign (&InetDeviceDsc, &svptr->ClientChannel, 0, 0)))
      ErrorExitVmsStatus (status, "sys$assign()", FI_LI);

   svptr->ClientSocketNameItem.Length = sizeof(svptr->ClientSocketName);
   svptr->ClientSocketNameItem.Address = &svptr->ClientSocketName;
   svptr->ClientSocketNameItem.LengthPtr = &svptr->ClientSocketNameLength;

   if (Debug) fprintf (stdout, "sys$qio() IO$_ACCESS | IO$M_ACCEPT\n");
   status = sys$qio (0, svptr->ServerChannel, IO$_ACCESS | IO$M_ACCEPT,
                     &svptr->ServerIOsb, &NetAcceptAst, svptr,
                     0, 0, &svptr->ClientSocketNameItem, &svptr->ClientChannel,
                     &LingerOnCloseOption, 0);
   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
   if (VMSnok (status)) 
      ErrorExitVmsStatus (status, "sys$qio()", FI_LI);

#endif

#ifdef IP_NETLIB

   if (Debug) fprintf (stdout, "netlib_accept()\n");
   status = netlib_accept (&svptr->ServerNetLibSocket,
                           &svptr->ClientNetLibSocket,
                           &svptr->ClientNetLibName, &NetLibSizeofSINDEF,
                           &svptr->NetLibReturnedLength,
                           &svptr->ServerIOsb, &NetAcceptAst, svptr);
   if (Debug) fprintf (stdout, "netlib_accept() %%X%08.08X\n", status);
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "netlib_accept()", FI_LI);

#endif

}

/*****************************************************************************/
/*
A connection has been accept()ed on the specified server socket.
*/ 

NetAcceptAst (struct ServiceStruct *svptr)

{
   register struct RequestStruct  *rqptr;
   register struct ServiceStruct  *tsvptr;

   boolean WatchThis;
   int  status,
        ServerSocketNameLength;
   unsigned short  PortNumber;
   char  ServerIpAddressString [32];
   struct AnIOsb  IOsb;

#ifdef IP_UCX
   int  SocketNameLength;
   struct sockaddr_in  SocketName;
   struct {
      unsigned int  Length;
      char  *Address;
      unsigned int  *LengthPtr;
   } SocketNameItem;
#endif

#ifdef IP_NETLIB
   static $DESCRIPTOR (IpAddressStringDsc, "");
   unsigned short  Length;
   unsigned long  NetLibReturnedLength;
   struct SINDEF  NetLibName;
#endif

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetAcceptAst() %d\n", svptr);

#ifdef IP_UCX
   if (Debug)
      fprintf (stdout, "sys$qio() IOsb %%X%08.08X\n",
               svptr->ServerIOsb.Status);
#endif

#ifdef IP_NETLIB
   if (Debug)
      fprintf (stdout, "netlib_accept() IOsb %%X%08.08X\n",
               svptr->ServerIOsb.Status);
#endif
      
   if (VMSnok (svptr->ServerIOsb.Status)) 
   {
      if ((ControlExitRequested || ControlRestartRequested) &&
          svptr->ServerIOsb.Status == SS$_ABORT)
      {
         /* server exiting or restarting, no new accepts, just return */
         return;
      }

      if (svptr->ServerIOsb.Status == SS$_CANCEL ||
          svptr->ServerIOsb.Status == SS$_ABORT)
      {
         /* connect dropped, forget it, ready for next connection */
         NetAccept (svptr);
         return;
      }

      /* most often network/system shutting down ... SS$_SHUT */
      ErrorExitVmsStatus (svptr->ServerIOsb.Status, "accept()", FI_LI);
   }

   Accounting.ConnectCount++;

#ifdef IP_UCX

   svptr->ClientPort = svptr->ClientSocketName.sin_port;
   svptr->ClientIpAddress = svptr->ClientSocketName.sin_addr.S_un.S_addr;
   strcpy (svptr->ClientIpAddressString,
           inet_ntoa (svptr->ClientSocketName.sin_addr));
   if (Debug) fprintf (stdout, "|%s|\n", svptr->ClientIpAddressString);

   if (NetMultiHomedHost)
   {
      /************************************/
      /* multiple services, get host/port */
      /************************************/

      SocketNameItem.Length = sizeof(SocketName);
      SocketNameItem.Address = &SocketName;
      SocketNameItem.LengthPtr = &SocketNameLength;

      if (Debug) fprintf (stdout, "sys$qiow() IO$_SENSEMODE\n");
      status = sys$qiow (0, svptr->ClientChannel, IO$_SENSEMODE,
                         &IOsb, 0, 0, 0, 0, &SocketNameItem, 0, 0, 0);
      if (Debug)
         fprintf (stdout, "sys$qiow() %%X%08.08X IOsb %%X%08.08X\n",
                  status, IOsb.Status);
      if (VMSok (status) && VMSnok (IOsb.Status))
         status = IOsb.Status;
      if (VMSnok (status))
      {
         /* hmmm, forget it, ready for next connection */
         NetAccept (svptr);
         return;
      }

      PortNumber = ntohs(SocketName.sin_port);

      if (Debug)
         fprintf (stdout, "|%s|%d|\n",
                  inet_ntoa (SocketName.sin_addr), PortNumber);

      for (tsvptr = ServiceListHead; tsvptr != NULL; tsvptr = tsvptr->NextPtr)
      {
         if (SocketName.sin_addr.s_addr ==
             tsvptr->ServerIpAddress &&
             PortNumber == tsvptr->ServerPort) break;
      }
      /* if none matches then 'tsvptr' is set to NULL */
   }
   else
   if (ServiceCount > 1)
   {
      /* multiple, possibly virtual services */
      tsvptr = svptr;
   }
   else
   {
      /* "they can have any color they want, provided it's black" */
      tsvptr = ServiceListHead;
   }

#endif

#ifdef IP_NETLIB

   IpAddressStringDsc.dsc$a_pointer = svptr->ClientIpAddressString;
   IpAddressStringDsc.dsc$w_length = sizeof(svptr->ClientIpAddressString)-1;

   svptr->ClientPort = svptr->ClientNetLibName.sin_w_port;
   svptr->ClientIpAddress =
      svptr->ClientNetLibName.sin_x_addr.inaddr_l_addr;

   status = netlib_addrtostr (&svptr->ClientNetLibName.sin_x_addr.inaddr_l_addr,
                              &IpAddressStringDsc, &Length);
   if (Debug) fprintf (stdout, "netlib_addrtostr() %%X%08.08X\n", status);
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "netlib_addrtostr()", FI_LI);

   svptr->ClientIpAddressString[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", svptr->ClientIpAddressString);

   if (NetMultiHomedHost)
   {
      /************************************/
      /* multiple services, get host/port */
      /************************************/

      if (Debug) fprintf (stdout, "netlib_getsockname()\n");
      status = netlib_getsockname (&svptr->ClientNetLibSocket,
                                   &NetLibName, &NetLibSizeofSINDEF,
                                   &NetLibReturnedLength, &IOsb, 0, 0);
      if (Debug)
         fprintf (stdout, "netlib_getsockname %%X%08.08X IOsb %%X%08.08X\n",
                  status, IOsb.Status);
      if (VMSok (status) && VMSnok (IOsb.Status))
         status = IOsb.Status;
      if (VMSnok (status))
      {
         /* hmmm, forget it, ready for next connection */
         NetAccept (svptr);
         return;
      }

      PortNumber = netlib_ntoh_word (&NetLibName.sin_w_port);

      if (Debug)
      {
         char  AddressString [32];
         unsigned short  Length;
         $DESCRIPTOR (AddressStringDsc, AddressString);

         netlib_addrtostr (&NetLibName.sin_x_addr, &AddressStringDsc, &Length);
         AddressString[Length] = '\0';
         fprintf (stdout, "|%s|%d|\n", AddressString, PortNumber);
      }

      for (tsvptr = ServiceListHead; tsvptr != NULL; tsvptr = tsvptr->NextPtr)
      {
         if (NetLibName.sin_x_addr.inaddr_l_addr == tsvptr->ServerIpAddress &&
             PortNumber == tsvptr->ServerPort)
            break;
      }
      /* if none matches then 'tsvptr' is set to NULL */
   }
   else
   if (ServiceCount > 1)
   {
      /* multiple, possibly virtual services */
      tsvptr = svptr;
   }
   else
   {
      /* ditto */
      tsvptr = ServiceListHead;
   }

#endif

   if (Debug && tsvptr != NULL)
      fprintf (stdout, "|%s|%s:%d|\n",
               tsvptr->ServerIpAddressString,
               tsvptr->ServerHostName, tsvptr->ServerPort);

   if (svptr->RequestScheme == SCHEME_HTTPS)
      Accounting.ConnectSslCount++;

   if (Config.DnsLookup)
      NetClientAddressLookup (svptr);
   else
      strcpy (svptr->ClientHostName, svptr->ClientIpAddressString);
   if (Debug) fprintf (stdout, "|%s|\n", svptr->ClientHostName);

   if (AdminWatchEnabled)
      WatchThis = AdminWatchFilter (svptr->ClientHostName,
                                    svptr->ClientIpAddressString,
                                    svptr->RequestSchemeNamePtr,
                                    svptr->ServerHostPort,
                                    NULL);
   else
      WatchThis = false;

   if (CurrentConnectCount >= MaxConcurrentConnections)
   {
      /************/
      /* too busy */
      /************/

      if (WatchThis && (AdminWatchEnabled & ADMIN_WATCH_CONNECT))
         AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_CONNECT,
                         "TOO BUSY !AZ,!UL on !AZ//!AZ:!AZ",
                          svptr->ClientHostName, svptr->ClientPort, 
                          svptr->RequestSchemeNamePtr,
                          svptr->ServerIpAddressString,
                          svptr->ServerPortString);

      svptr->ConnectCount++;
      Accounting.ConnectTooBusyCount++;
      Accounting.ResponseStatusCodeCountArray[5]++;

#ifdef IP_UCX
      NetTooBusy (svptr->ClientChannel,
                  svptr->ClientHostName,
                  svptr->ClientIpAddressString);
      NetDummyRead (svptr->ClientChannel);
#endif

#ifdef IP_NETLIB
      NetTooBusy (&svptr->ClientNetLibSocket,
                  svptr->ClientHostName,
                  svptr->ClientIpAddressString);
      NetDummyRead (&svptr->ClientNetLibSocket);
#endif

      /* ready for next connection */
      NetAccept (svptr);

      return;
   }

   /* 'tsvptr' is NULL if the request did not match a known service */
   if (tsvptr == NULL ||
       !ConfigAcceptClientHostName (svptr->ClientIpAddressString,
                                    svptr->ClientHostName))
   {
      /************/
      /* rejected */
      /************/

      if (WatchThis && (AdminWatchEnabled & ADMIN_WATCH_CONNECT))
         AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_CONNECT,
                         "REJECT !AZ,!UL on !AZ//!AZ:!AZ",
                          svptr->ClientHostName, svptr->ClientPort, 
                          svptr->RequestSchemeNamePtr,
                          svptr->ServerIpAddressString,
                          svptr->ServerPortString);

      svptr->ConnectCount++;
      Accounting.ConnectRejectedCount++;
      Accounting.ResponseStatusCodeCountArray[4]++;

#ifdef IP_UCX
      NetAccessDenied (svptr->ClientChannel,
                       svptr->ClientHostName,
                       svptr->ClientIpAddressString);
      NetDummyRead (svptr->ClientChannel);
#endif

#ifdef IP_NETLIB
      NetAccessDenied (&svptr->ClientNetLibSocket,
                       svptr->ClientHostName,
                       svptr->ClientIpAddressString);
      NetDummyRead (&svptr->ClientNetLibSocket);
#endif

      /* ready for next connection */
      NetAccept (svptr);

      return;
   }

   Accounting.ConnectAcceptedCount++;

   /* allocate zeroed dynamic memory for the connection thread */
   rqptr = VmGetRequest ();

   /***************************/
   /* process HTTP connection */
   /***************************/

   /* once the thread structure has been created we have a client! */
   Accounting.ConnectCurrent = ++CurrentConnectCount;
   if (CurrentConnectCount > Accounting.ConnectPeak)
      Accounting.ConnectPeak = CurrentConnectCount;

   /* point at this service' structure */
   rqptr->ServicePtr = tsvptr;

   /* point at and copy the scheme (protocol) the request is using */
   rqptr->RequestScheme = tsvptr->RequestScheme;
   rqptr->RequestSchemeNamePtr = tsvptr->RequestSchemeNamePtr;

   /* copy the client's network information */
   rqptr->ClientPort = svptr->ClientPort;
   rqptr->ClientIpAddress = svptr->ClientIpAddress;
   strcpy (rqptr->ClientIpAddressString, svptr->ClientIpAddressString);
   strcpy (rqptr->ClientHostName, svptr->ClientHostName);

#ifdef IP_UCX
   rqptr->ClientChannel = svptr->ClientChannel;
#endif

#ifdef IP_NETLIB
   rqptr->ClientNetLibSocket = svptr->ClientNetLibSocket;
#endif

   if (WatchThis)
   {
      /* assign a unique number for the life of this request structure */
      rqptr->AdminWatchItem = ++AdminWatchItem;

      if (AdminWatchEnabled & ADMIN_WATCH_CONNECT)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_CONNECT,
                         "ACCEPT !AZ,!UL on !AZ//!AZ:!AZ",
                          svptr->ClientHostName, svptr->ClientPort, 
                          svptr->RequestSchemeNamePtr,
                          svptr->ServerIpAddressString,
                          svptr->ServerPortString);
   }

   RequestBegin (rqptr, true);

   /* ready for next connection */
   NetAccept (svptr);
}

/*****************************************************************************/
/*
Some browsers report a network error (e.g. Windows Netscape Navigator 2.n) if
the server does not read anything from them before returning an error message.
This routine and it's AST ensure a read is done before reporting an error
earlier that reading the request from it.  Needless-to-say, no status checking
is performed and the data just discarded (hope it works consistently :^)
*/ 

#ifdef IP_UCX

NetDummyRead (unsigned short Channel)

{
   static char  DummyBuffer [1024];
   static struct AnIOsb  DummyIOsb;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetDummyRead() Channel: %d\n", Channel);

   status = sys$qio (0, Channel, IO$_READVBLK, &DummyIOsb,
                     &NetDummyReadAst, Channel,
                     DummyBuffer, sizeof(DummyBuffer), 0, 0, 0, 0);

   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);

}
#endif

#ifdef IP_NETLIB

NetDummyRead (unsigned long *NetLibSocketPtr)

{
   static char  DummyBuffer [1024];
   static struct AnIOsb  DummyIOsb;
   static $DESCRIPTOR (DummyBufferDsc, DummyBuffer);

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetDummyRead() *NetLibSocketPtr: %d\n",
               *NetLibSocketPtr);

   status = netlib_read (NetLibSocketPtr, &DummyBufferDsc, 0, 0, 0, 0,
                         &DummyIOsb, &NetDummyReadAst, NetLibSocketPtr);

   if (Debug) fprintf (stdout, "netlib_read %%X%08.08X\n", status);
}
#endif

/*****************************************************************************/
/*
See commentary in NetDummyRead().
*/ 

#ifdef IP_UCX

NetDummyReadAst (unsigned short Channel)

{
   int  status;
   struct  AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetDummyReadAst() Channel: %d\n", Channel);

   status = sys$qiow (0, Channel, IO$_DEACCESS | IO$M_SHUTDOWN, &IOsb, 0, 0,
                      0, 0, 0, UCX$C_DSC_ALL, 0, 0);
   if (Debug)
      fprintf (stdout, "sys$qiow() %%X%08.08X IOsb %%X%08.08X\n",
               status, IOsb.Status);

   status = sys$dassgn (Channel);
   if (Debug) fprintf (stdout, "sys$dassgn() %%X%08.08X\n", status);
}

#endif

#ifdef IP_NETLIB

NetDummyReadAst (unsigned long *NetLibSocketPtr)

{
   int  status;
   struct  AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetDummyReadAst() *NetLibSocketPtr: %d\n",
               *NetLibSocketPtr);

   status = netlib_shutdown (NetLibSocketPtr, &NetLibShutdownBoth, &IOsb, 0, 0);
   if (Debug)
      fprintf (stdout, "netlib_shutdown %%X%08.08X IOsb %%X%08.08X\n",
               status, IOsb.Status);

   status = netlib_close (NetLibSocketPtr);
   if (Debug) fprintf (stdout, "netlib_close() %%X%08.08X\n", status);
}

#endif

/*****************************************************************************/
/*
Using the 'HostName' parameter get the lookup host name into 'HostNamePtr', the
decimal-dot-notation IP address string into 'IpAddressStringPtr' and the 32 bit
IP address into 'IpAddressPtr'.  Any of the '...Ptr' parameters can be NULL and
won't have the respective information returned.
*/ 

NetHostNameLookup
(
char *HostNamePort,
char *HostNamePtr,
char *HostNamePortPtr,
char *PortStringPtr,
int *PortPtr,
char *IpAddressStringPtr,
int *IpAddressPtr
)
{

#ifdef IP_UCX

   static $DESCRIPTOR (HostNamePortFaoDsc, "!AZ:!UL\0");
   static $DESCRIPTOR (PortFaoDsc, "!UL\0");
   static $DESCRIPTOR (StringDsc, "");

   register char  *cptr, *sptr;

   int  idx,
        HostPort;
   char  ch;
   char  HostName [128];
   struct hostent  *HostEntryPtr;
   struct in_addr  HostAddress;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetHostNameLookup() |%s|\n", HostNamePort);

   for (cptr = HostNamePort; *cptr && *cptr != ':'; cptr++);
   if (ch = *cptr) *cptr++ = '\0';
   HostPort = atoi(cptr);
   if (!HostPort) HostPort = 80;
   if (Debug) fprintf (stdout, "%s:%d\n", HostNamePort, HostPort);
   HostEntryPtr = gethostbyname (HostNamePort);
   if (HostEntryPtr == NULL) return (vaxc$errno);
   if (ch) *--cptr = ch;

   for (cptr = HostNamePort; *cptr && *cptr != '.' && *cptr != ':'; cptr++);
   if (*cptr == '.')
   {
      /* fully qualified host name, just copy it in */
      cptr = HostNamePort;
   }
   else
   {
      /* use the resolved name */
      cptr = HostEntryPtr->h_name;
   }
   /* looks better if its all in lower case (sometimes upper) */
   sptr = HostName;
   while (*cptr && *cptr != ':') *sptr++ = tolower(*cptr++);
   *sptr = '\0';
   if (Debug) fprintf (stdout, "HostName |%s|\n", HostName);

   if (HostNamePtr != NULL)
   {
      strcpy (HostNamePtr, HostName);
      if (Debug) fprintf (stdout, "|%s|\n", HostNamePtr);
   }

   if (HostNamePortPtr != NULL)
   {
      StringDsc.dsc$a_pointer = HostNamePortPtr;
      StringDsc.dsc$w_length = 128;
      sys$fao (&HostNamePortFaoDsc, 0, &StringDsc, HostName, HostPort);
      if (Debug) fprintf (stdout, "|%s|\n", HostNamePortPtr);
   }

   if (PortStringPtr != NULL)
   {
      StringDsc.dsc$a_pointer = PortStringPtr;
      StringDsc.dsc$w_length = 32;
      sys$fao (&PortFaoDsc, 0, &StringDsc, HostPort);
      if (Debug) fprintf (stdout, "|%s|\n", PortStringPtr);
   }

   if (PortPtr != NULL)
   {
      *PortPtr = HostPort;
      if (Debug) fprintf (stdout, "%d\n", *PortPtr);
   }

   if (IpAddressPtr != NULL)
   {
      memcpy (IpAddressPtr, HostEntryPtr->h_addr, 4);
      if (Debug) fprintf (stdout, "%08.08X\n", *IpAddressPtr);
   }

   if (IpAddressStringPtr != NULL)
   {
      memcpy (&HostAddress, HostEntryPtr->h_addr, 4);
      strcpy (IpAddressStringPtr, inet_ntoa (HostAddress));
      if (Debug) fprintf (stdout, "|%s|\n", IpAddressStringPtr);
   }

   return (SS$_NORMAL);

#endif

#ifdef IP_NETLIB

   static $DESCRIPTOR (HostNamePortFaoDsc, "!AZ:!UL\0");
   static $DESCRIPTOR (PortFaoDsc, "!UL\0");
   static $DESCRIPTOR (StringDsc, "");

   register char  *cptr, *sptr;

   int  status,
        HostPort;
   unsigned short  Length;
   unsigned long  NetLibSocket;
   char  ch;
   char  HostName [128];
   struct  AnIOsb  IOsb;
   struct INADDRDEF NetLibAddressList [16];
   $DESCRIPTOR (HostNameDsc, HostName);
   $DESCRIPTOR (IpAddressStringDsc, "");

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetHostNameLookup() |%s|\n", HostNamePort);

   /* NETLIB needs a socket to do these, create a temporary one */
   if (Debug) fprintf (stdout, "netlib_socket()\n");
   status = netlib_socket (&NetLibSocket, &NetLibSocketType,
                           &NetLibSocketFamily);
   if (Debug) fprintf (stdout, "netlib_socket() %%X%08.08X\n", status);
   if (VMSnok (status)) return (status);

   for (cptr = HostNamePort; *cptr && *cptr != ':'; cptr++);
   if (ch = *cptr) *cptr++ = '\0';
   HostPort = atoi(cptr);
   if (!HostPort) HostPort = 80;
   if (Debug) fprintf (stdout, "%s:%d\n", HostNamePort, HostPort);
   HostNameDsc.dsc$a_pointer = HostNamePort;
   HostNameDsc.dsc$w_length = strlen(HostNamePort);
   if (ch) *--cptr = ch;

   /* turn the name into a binary address */
   status = netlib_name_to_address (&NetLibSocket, 0, &HostNameDsc,
                                    &NetLibAddressList,
                                    &NetLibAddressListSize,
                                    &NetLibAddressListCount,
                                    &IOsb, 0, 0);
   if (Debug)
      fprintf (stdout,
         "netlib_name_to_address() %%X%08.08X IOsb %%X%08.08X\n",
         status, IOsb.Status);
   if (VMSok (status) && VMSnok (IOsb.Status))
      status = IOsb.Status;
   if (VMSnok (status)) return (status);

   if (PortStringPtr != NULL)
   {
      StringDsc.dsc$a_pointer = PortStringPtr;
      StringDsc.dsc$w_length = 32;
      sys$fao (&PortFaoDsc, 0, &StringDsc, HostPort);
      if (Debug) fprintf (stdout, "|%s|\n", PortStringPtr);
   }

   if (PortPtr != NULL)
   {
      *PortPtr = HostPort;
      if (Debug) fprintf (stdout, "%d\n", *PortPtr);
   }

   if (IpAddressPtr != NULL)
   {
      *IpAddressPtr = NetLibAddressList[0].inaddr_l_addr;
      if (Debug) fprintf (stdout, "%08.08X\n", *IpAddressPtr);
   }

   if (IpAddressStringPtr != NULL)
   {
      /* convert the binary address into a string */
      IpAddressStringDsc.dsc$a_pointer = IpAddressStringPtr;
      IpAddressStringDsc.dsc$w_length = 31;
      status = netlib_addrtostr (&NetLibAddressList[0],
                                 &IpAddressStringDsc, &Length);

      if (Debug) fprintf (stdout, "netlib_addrtostr() %%X%08.08X\n", status);
      if (VMSnok (status)) return (status);
      IpAddressStringPtr[Length] = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", IpAddressStringPtr);
   }

   if (HostNamePtr != NULL || HostNamePortPtr != NULL)
   {
      for (cptr = HostNamePort; *cptr && *cptr != '.' && *cptr != ':'; cptr++);
      if (*cptr == '.')
      {
         /* fully qualified host name, just copy it in */
         sptr = HostName;
         for (cptr = HostNamePort;
              *cptr && *cptr != ':';
              *sptr++ = tolower(*cptr++));
         *sptr = '\0';
         if (Debug) fprintf (stdout, "|%s|\n", HostName);
      }
      else
      {
         /* transform abbreviated name to full ("host" into "host.domain") */
         HostNameDsc.dsc$a_pointer = HostName;
         HostNameDsc.dsc$w_length = sizeof(HostName)-1;
         Length = 0;
         status = netlib_address_to_name (&NetLibSocket, 0,
                     &NetLibAddressList[0].inaddr_l_addr,
                     &NetLibSizeofINADDRDEF, &HostNameDsc, &Length,
                     &IOsb, 0, 0);
         if (Debug)
            fprintf (stdout,
               "netlib_address_to_name() %%X%08.08X IOsb %%X%08.08X\n",
               status, IOsb.Status);
         HostName[Length] = '\0';

         /* looks better if its all in lower case (sometimes upper) */
         for (cptr = HostName; *cptr; cptr++) *cptr = tolower(*cptr);
         if (Debug) fprintf (stdout, "|%s|\n", HostName);
      }
      if (HostNamePtr) strcpy (HostNamePtr, HostName);
   }

   if (HostNamePortPtr != NULL)
   {
      StringDsc.dsc$a_pointer = HostNamePortPtr;
      StringDsc.dsc$w_length = 128;
      sys$fao (&HostNamePortFaoDsc, 0, &StringDsc, HostName, HostPort);
      if (Debug) fprintf (stdout, "|%s|\n", HostNamePortPtr);
   }

   /* get rid of the temporary socket */
   status = netlib_close (&NetLibSocket);
   if (Debug) fprintf (stdout, "netlib_close() %%X%08.08X\n", status);

   return (SS$_NORMAL);

#endif

}

/*****************************************************************************/
/*
Perform the UCX get host name using QIO ACP functionality.  This can execute 
at AST-delivery level.  Doing this within NETLIB needs to cater for the
lowest-common-denominator package, which sometimes will not execute at AST
delivery level, therefore rendering some NETLIB packages incapable of
supporting DNS lookup!
*/ 

#ifdef IP_UCX

NetClientAddressLookup (struct ServiceStruct *svptr) 

{
   static unsigned char ControlSubFunction [4] =
      { INETACP_FUNC$C_GETHOSTBYADDR, INETACP$C_TRANS, 0, 0 };
   static struct dsc$descriptor AddressDsc =
      { 0, DSC$K_DTYPE_T, DSC$K_CLASS_S, 0};
   static struct dsc$descriptor ControlSubFunctionDsc =
      { 4, DSC$K_DTYPE_T, DSC$K_CLASS_S, &ControlSubFunction };
   static struct dsc$descriptor ClientHostNameDsc =
      { 0, DSC$K_DTYPE_T, DSC$K_CLASS_S, 0};

   int  status;
   unsigned short  Length;
   struct  AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetClientAddressLookup()\n");

   AddressDsc.dsc$w_length = sizeof(svptr->ClientSocketName.sin_addr);
   AddressDsc.dsc$a_pointer = &svptr->ClientSocketName.sin_addr;

   ClientHostNameDsc.dsc$w_length = sizeof(svptr->ClientHostName)-1;
   ClientHostNameDsc.dsc$a_pointer = &svptr->ClientHostName;

   Length = 0;
   if (Debug) fprintf (stdout, "sys$qio() IO$_ACPCONTROL\n");
   status = sys$qiow (0, svptr->ServerChannel, IO$_ACPCONTROL,
                      &IOsb, 0, 0, &ControlSubFunctionDsc,
                      &AddressDsc, &Length, &ClientHostNameDsc, 0, 0);
   if (Debug)
      fprintf (stdout, "sys$qio() %%X%08.08X IOsb %%X%08.08X\n",
               status, IOsb.Status);
   if (Length)
      svptr->ClientHostName[Length] = '\0';
   else
      strcpy (svptr->ClientHostName, svptr->ClientIpAddressString);
}
#endif

#ifdef IP_NETLIB

NetClientAddressLookup (struct ServiceStruct *svptr) 

{
   static struct dsc$descriptor ClientHostNameDsc =
      { 0, DSC$K_DTYPE_T, DSC$K_CLASS_S, 0};

   int  status;
   unsigned short  Length;
   struct AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetClientAddressLookup()\n");

   ClientHostNameDsc.dsc$w_length = sizeof(svptr->ClientHostName)-1;
   ClientHostNameDsc.dsc$a_pointer = &svptr->ClientHostName;

   Length = 0;
   status = netlib_address_to_name (&svptr->ClientNetLibSocket, 0,
               &svptr->ClientNetLibName.sin_x_addr.inaddr_l_addr,
               &NetLibSizeofINADDRDEF, &ClientHostNameDsc, &Length,
               &IOsb, 0, 0);
   if (Debug)
      fprintf (stdout, "netlib_address_to_name() %%X%08.08X IOsb %%X%08.08X\n",
               status, IOsb.Status);
   if (Length)
      svptr->ClientHostName[Length] = '\0';
   else
      strcpy (svptr->ClientHostName, svptr->ClientIpAddressString);
}
#endif

/****************************************************************************/
/*
Asynchronously shut the client socket.  This can be done in two modes.  First,
to break the connection under an error condition (effectively a sys$cancel()). 
An AST address of -1 (i.e. no actual AST function to be delivered) should be
supplied for this. Second, to close a socket for a completed request. The AST
function address (should be for RequestEnd()).
*/

int NetShutdownSocket
(
struct RequestStruct *rqptr,
void *AstFunctionPtr
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetShutdownSocket() %d %d\n", rqptr, AstFunctionPtr);

   if (rqptr->NetShutdownAstFunctionPtr == -1)
   {
      /* socket is already being "broken", substitute this AST address */
      rqptr->NetShutdownAstFunctionPtr = AstFunctionPtr;
      return (SS$_NORMAL);
   }
   rqptr->NetShutdownAstFunctionPtr = AstFunctionPtr;

#ifdef IP_UCX

   status = sys$qio (0, rqptr->ClientChannel, IO$_DEACCESS | IO$M_SHUTDOWN,
                     &rqptr->NetShutdownIOsb, NetShutdownSocketAst, rqptr,
                     0, 0, 0, UCX$C_DSC_ALL, 0, 0);
   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);

#endif

#ifdef IP_NETLIB

/*
   30-OCT-98  MGD
   There seems to be a problem with the use of this NETLIB function in
   the following code, but I'm stumped if I can see where it's my doing!
   The workaround has been to call the AST routine explicitly.
   26-MAY-99  MGD
   Appears to have been fixed in NETLIB022.
*/

   status = netlib_shutdown (&rqptr->ClientNetLibSocket,
                             &NetLibShutdownBoth, &rqptr->NetShutdownIOsb,
                             NetShutdownSocketAst, rqptr);
   if (Debug) fprintf (stdout, "netlib_shutdown() %%X%08.08X\n", status);

#endif

   if (VMSok (status)) return (status);

   /* with resource wait enabled the only quota not waited for is ASTLM */
   if (status == SS$_EXQUOTA)
      ErrorExitVmsStatus (status, "sys$qio()", FI_LI);

   /* queuing of shutdown failed, call AST explicitly, status in the IOsb */
   rqptr->NetShutdownIOsb.Status = status;
   rqptr->NetShutdownIOsb.Count = 0;
   SysDclAst (NetShutdownSocketAst, rqptr);
   return (status);
}

/****************************************************************************/
/*
The socket has been shutdown, with IO effectively cancelled.  Close the socket. 
The pointer to the AST can contain one of 3 values;  NULL, no socket close AST
(not actually used by the HTTPd);  -1 to indicate that a socket shutdown is
underway but no AST should be called (used to "break" network connections,
forcing cancelled IO to be delivered);  Any other value as the address of the
AST function.
*/

NetShutdownSocketAst (struct RequestStruct *rqptr)

{
   int  status;
   struct AnIOsb  IOsb;
   void (*AstFunctionPtr)(struct RequestStruct*);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetShutdownSocketAst() %%X%08.08X\n",
               rqptr->NetShutdownIOsb.Status);

   if (rqptr->AdminWatchItem &&
       (AdminWatchEnabled & ADMIN_WATCH_NETWORK))
   {
      if (VMSok(rqptr->NetShutdownIOsb.Status))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_NETWORK,
                         "SHUTDOWN !AZ,!UL",
                         rqptr->ClientHostName, rqptr->ClientPort); 
      else
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_NETWORK,
                         "SHUTDOWN !AZ,!UL %X!8XL !AZ",
                         rqptr->ClientHostName, rqptr->ClientPort, 
                         rqptr->NetShutdownIOsb.Status,
                         SysGetMsg(rqptr->NetShutdownIOsb.Status));
   }

#ifdef IP_UCX

   if (rqptr->ClientChannel)
   {
      status = sys$dassgn (rqptr->ClientChannel);
      if (Debug) fprintf (stdout, "sys$dassgn() %%X%08.08X\n", status);

      rqptr->ClientChannel = 0;
   }

#endif

#ifdef IP_NETLIB

   if (rqptr->ClientNetLibSocket != NULL)
   {
      status = netlib_close (&rqptr->ClientNetLibSocket);
      if (Debug) fprintf (stdout, "netlib_close() %%X%08.08X\n", status);

      rqptr->ClientNetLibSocket = NULL;
   }

#endif

   if ((AstFunctionPtr = rqptr->NetShutdownAstFunctionPtr) == NULL) return;
   rqptr->NetShutdownAstFunctionPtr = NULL;
   if (Debug) fprintf (stdout, "AstFunctionPtr: %d\n", AstFunctionPtr);
   /* if just "breaking" the socket then no real AST destination */
   if (AstFunctionPtr == -1) return;
   (*AstFunctionPtr)(rqptr);
}

/****************************************************************************/
/*
Stop the server from receiving incoming requests.
*/

NetShutdownServerSocket ()

{
   register struct ServiceStruct  *svptr;

   int  status;
   struct  AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetShutdownServerSocket()\n");

   for (svptr = ServiceListHead; svptr != NULL; svptr = svptr->NextPtr)
   {
#ifdef IP_UCX

      status = sys$qiow (0, svptr->ServerChannel, IO$_DEACCESS, &IOsb, 0, 0,
                         0, 0, 0, 0, 0, 0);
      if (Debug)
         fprintf (stdout, "sys$qiow() %%X%08.08X IOsb %%X%08.08X\n",
                  status, IOsb.Status);

#endif

#ifdef IP_NETLIB

      status = netlib_shutdown (&svptr->ServerNetLibSocket,
                                &NetLibShutdownRecv, &IOsb, 0, 0);
      if (Debug)
         fprintf (stdout, "netlib_shutdown %%X%08.08X IOsb %%X%08.08X\n",
                  status, IOsb.Status);

#endif
   }
}

/*****************************************************************************/
/*
Unlike NetWrite() this function writes directly using a client channel/socket,
not using the full request data structure. This is for writing to the client
before any of the request data structures are created (e.g. when the server
denies access). This is a blocking function, but shouldn't introduce much
granularity as it is only used for short messages before a client thread is
set up.
*/ 

#ifdef IP_UCX

int NetDirectWrite
(
unsigned short Channel,
char *DataPtr,
int DataLength
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetDirectWrite()\n");

   status = sys$qiow (0, Channel, IO$_WRITEVBLK, 0, 0, 0,
                      DataPtr, DataLength, 0, 0, 0, 0);

   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);

   return (status);
}
#endif

#ifdef IP_NETLIB

int NetDirectWrite
(
unsigned long *NetLibSocketPtr,
char *DataPtr,
int DataLength
)
{
   static $DESCRIPTOR (DataPtrDsc, "");

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetDirectWrite()\n");

   DataPtrDsc.dsc$a_pointer = DataPtr;
   DataPtrDsc.dsc$w_length = DataLength;

   status = netlib_write (NetLibSocketPtr, &DataPtrDsc, 0, 0, 0, 0, 0);

   if (Debug) fprintf (stdout, "netlib_write() %%X%08.08X\n", status);

   return (status);
}
#endif

/*****************************************************************************/
/*
Called from NetWrite() is a response header needs to be sent before any data.
Response header has now been sent, send the data using the buffered
information about it.
*/ 

NetResponseHeaderAst (struct RequestStruct *rqptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetResponseHeaderAst()\n");

   NetWrite (rqptr,
             rqptr->ResponseHeaderAstFunctionPtr,
             rqptr->ResponseHeaderDataPtr,
             rqptr->ResponseHeaderDataLength);
}

/*****************************************************************************/
/*
Write 'DataLength' bytes located at 'DataPtr' to the client either using either
the "raw" network or via the Secure Sockets Layer.

If 'AstFunctionPtr' zero then use sys$qiow(), waiting for completion. If
an AST completion address is supplied then use sys$qio().  If empty data
buffer is supplied (zero length) then declare an AST to service any AST
routine supplied.  If none then just return.

For responses generated by the server the HTTP header is a separate structure
which can be sent separately. If the HTTP method is "HEAD" only allow bytes in
the header to be sent, absorb any other, explicitly calling the AST completion
routine as necessary (this, in particular, is for scripts that don't recognise
this method).

An "Xray" returns a full response (header and body) as a plain text document
(allows diagnostics, etc.) With an Xray just output a hard-wired, plain-text
response with wait for completion (won't be happening enough for any
significant impact!)

Explicitly declares any AST routine if an error occurs. The calling function
must not do any error recovery if an AST routine has been supplied but the
associated AST routine must!  If an AST was not supplied then the return
status can be checked.
*/ 

int NetWrite
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataLength
)
{
   int  status,
        ResponseHeaderLength;
   char  *ResponseHeaderPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetWrite() %d %d %d %d\n",
               rqptr, AstFunctionPtr, DataPtr, DataLength);
   /** if (Debug) fprintf (stdout, "|%s|\n", DataPtr); **/

   status = SS$_NORMAL;

   if (!rqptr->ResponseHeaderSent)
   {
      /*******************************************/
      /* nothing of response sent to client yet! */
      /*******************************************/

      rqptr->ResponseHeaderSent = true;

      if (rqptr->Xray)
      {
         /*********************************/
         /* header for "Xray" of response */
         /*********************************/

         ResponseHeaderPtr = rqptr->ResponseHeaderPtr;
         ResponseHeaderLength = rqptr->ResponseHeaderLength;

         rqptr->ResponsePreExpired = true;
         HTTP_HEADER_200_PLAIN (rqptr);

         if (rqptr->AdminWatchItem &&
             (AdminWatchEnabled & ADMIN_WATCH_RESPONSE_HEADER))
         {
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_RESPONSE_HEADER,
                            "HEADER !UL bytes", rqptr->ResponseHeaderLength);
            AdminWatchData (rqptr->ResponseHeaderPtr,
                            rqptr->ResponseHeaderLength);
         }

         rqptr->BytesTx += rqptr->ResponseHeaderLength;

         if (rqptr->SeSoLaPtr == NULL)
            status = NetWriteRaw (rqptr, 0,
                                  rqptr->ResponseHeaderPtr,
                                  rqptr->ResponseHeaderLength);
         else
            status = SeSoLaWrite (rqptr, 0,
                                  rqptr->ResponseHeaderPtr,
                                  rqptr->ResponseHeaderLength);

         rqptr->ResponseHeaderPtr = ResponseHeaderPtr;
         rqptr->ResponseHeaderLength = ResponseHeaderLength;
      }

      if (rqptr->ResponseHeaderPtr != NULL &&
          rqptr->HttpVersion != HTTP_VERSION_0_9)
      {
         /************************/
         /* send response header */
         /************************/

         if (Debug)
            fprintf (stdout, "flushing rqptr->ResponseHeaderPtr\n|%s|\n",
                     rqptr->ResponseHeaderPtr);

         if (!rqptr->ResponseHeaderLength)
            rqptr->ResponseHeaderLength = strlen(rqptr->ResponseHeaderPtr);

         if (rqptr->AdminWatchItem &&
             (AdminWatchEnabled & ADMIN_WATCH_RESPONSE_HEADER) &&
             !rqptr->AdminWatchClient)
         {
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_RESPONSE_HEADER,
                            "HEADER !UL bytes", rqptr->ResponseHeaderLength);
            AdminWatchData (rqptr->ResponseHeaderPtr,
                            rqptr->ResponseHeaderLength);
         }

         rqptr->BytesTx += rqptr->ResponseHeaderLength;

         if (AstFunctionPtr != NULL)
         {
            /************************/
            /* AST routine, no wait */
            /************************/

            if (DataLength)
            {
               rqptr->ResponseHeaderAstFunctionPtr = AstFunctionPtr;
               rqptr->ResponseHeaderDataPtr = DataPtr;
               rqptr->ResponseHeaderDataLength = DataLength;
               AstFunctionPtr = &NetResponseHeaderAst;
            }

            if (rqptr->SeSoLaPtr == NULL)
               status = NetWriteRaw (rqptr, AstFunctionPtr,
                                     rqptr->ResponseHeaderPtr,
                                     rqptr->ResponseHeaderLength);
            else
               status = SeSoLaWrite (rqptr, AstFunctionPtr,
                                     rqptr->ResponseHeaderPtr,
                                     rqptr->ResponseHeaderLength);

            /* return, the AST function will sys$qiow() the data */
            return (status);
         }
         else
         {
            /************************/
            /* no AST routine, wait */
            /************************/

            if (rqptr->SeSoLaPtr == NULL)
               status = NetWriteRaw (rqptr, 0,
                                     rqptr->ResponseHeaderPtr,
                                     rqptr->ResponseHeaderLength);
            else
               status = SeSoLaWrite (rqptr, 0,
                                     rqptr->ResponseHeaderPtr,
                                     rqptr->ResponseHeaderLength);

            /* continue on to sys$qiow() the data */
         }
      }
   }

   /********/
   /* data */
   /********/

   rqptr->BytesTx += DataLength;

   if (!DataLength)
   {
      /* without a real network write just fudge the status */
      rqptr->NetWriteIOsb.Status = SS$_NORMAL;
      rqptr->NetWriteIOsb.Count = 0;
      if (AstFunctionPtr == NULL) return (SS$_NORMAL);
      SysDclAst (AstFunctionPtr, rqptr);
      return (SS$_NORMAL);
   }

   if (rqptr->HttpMethod == HTTP_METHOD_HEAD)
   {
      /*****************************/
      /* send only response header */
      /*****************************/

      if (Debug) fprintf (stdout, "method HEAD\n");

      if (rqptr->ResponseHeaderPtr == NULL &&
          rqptr->ResponseHeaderNewlineCount <= 1 &&
          rqptr->ProxyTaskPtr == NULL)
      {
         /**********************/
         /* header from script */
         /**********************/

         register int  cnt, nlcnt;
         register char  *cptr;

         nlcnt = rqptr->ResponseHeaderNewlineCount;
         if (Debug) fprintf (stdout, "nlcnt: %d\n", nlcnt);
         cptr = DataPtr;
         cnt = DataLength;
         while (cnt--)
         {
            if (*cptr == '\n' && ++nlcnt == 2)
            {
               /* two successive end-of-lines, therefore end of header */
               if (Debug) fprintf (stdout, "end-of-header\n");
               cptr++;
               break;
            }
            else
            if (*cptr != '\r' && *cptr != '\n')
               nlcnt = 0;
            cptr++;
         }
         if ((rqptr->ResponseHeaderNewlineCount = nlcnt) == 2)
         {
            /* finally found those two consecutive newlines! */
            rqptr->ResponseHeaderSent = true;
            if (Debug) fprintf (stdout, "DataLength: %d\n", cptr - DataPtr);
            if (DataLength = cptr - DataPtr)
            {
               /* adjust data length to include only the HTTP header */
               DataLength = cptr - DataPtr;
            }
            else
            {
               /* no data in HTTP header left at all */
               rqptr->NetWriteIOsb.Status = STS$K_SUCCESS;
               rqptr->NetWriteIOsb.Count = 0;
               if (AstFunctionPtr != NULL) SysDclAst (AstFunctionPtr, rqptr);
               return (SS$_NORMAL);
            }
         }
      }
      else
      {
         /* HTTP header has been completely sent, absorb anything else */
         rqptr->NetWriteIOsb.Status = STS$K_SUCCESS;
         rqptr->NetWriteIOsb.Count = 0;
         if (AstFunctionPtr != NULL) SysDclAst (AstFunctionPtr, rqptr);
         return (SS$_NORMAL);
      }
   }

   if (rqptr->HttpVersion == HTTP_VERSION_0_9 &&
       rqptr->ResponseHeaderPtr == NULL &&
       rqptr->ResponseHeaderNewlineCount <= 1 &&
       rqptr->ProxyTaskPtr == NULL)
   {
      /*********************************/
      /* absorb any header from script */
      /*********************************/

      /* local storage */
      register int  cnt, nlcnt;
      register char  *cptr;

      nlcnt = rqptr->ResponseHeaderNewlineCount;
      if (Debug) fprintf (stdout, "HTTP/0.9 nlcnt: %d\n", nlcnt);

      cptr = DataPtr;
      cnt = DataLength;
      while (cnt--)
      {
         if (*cptr == '\n' && ++nlcnt == 2)
         {
            /* two successive end-of-lines, therefore end of header */
            if (Debug) fprintf (stdout, "end-of-header\n");
            cptr++;
            break;
         }
         else
         if (*cptr != '\r' && *cptr != '\n')
            nlcnt = 0;
         cptr++;
      }
      if ((rqptr->ResponseHeaderNewlineCount = nlcnt) == 2)
      {
         /* adjust data pointer and length to exclude header */
         rqptr->ResponseHeaderSent = true;
         DataLength = cptr - DataPtr;
         DataPtr = cptr;
         if (Debug)
            fprintf (stdout, "DataPtr: %d DataLength: %d\n",
                     DataPtr, DataLength);
      }
      else
      {
         /* no data in HTTP header left at all */
         rqptr->NetWriteIOsb.Status = STS$K_SUCCESS;
         rqptr->NetWriteIOsb.Count = 0;
         if (AstFunctionPtr != NULL) SysDclAst (AstFunctionPtr, rqptr);
         return (SS$_NORMAL);
      }
      /* HTTP header has been completely absorbed, send everything else */
   }

   /*************/
   /* send data */
   /*************/

   if (rqptr->AdminWatchItem &&
       (AdminWatchEnabled & ADMIN_WATCH_RESPONSE_BODY))
   {
      if (rqptr->ResponseHeaderPtr == NULL)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_RESPONSE_BODY,
                         "STREAM !UL bytes", DataLength);
      else
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_RESPONSE_BODY,
                         "BODY !UL bytes", DataLength);
      AdminWatchDataDump (DataPtr, DataLength);
   }

   if (rqptr->SeSoLaPtr == NULL)
      status = NetWriteRaw (rqptr, AstFunctionPtr, DataPtr, DataLength);
   else
      status = SeSoLaWrite (rqptr, AstFunctionPtr, DataPtr, DataLength);

   return (status);
}

/*****************************************************************************/
/*
Write data to the network. Explicitly declares any AST routine if an error
occurs. The calling function must not do any error recovery if an AST routine
has been supplied but the associated AST routine must! If an AST was not
supplied then the return status can be checked.  AST to NetWriteRaw() which
calls the supplied AST function.
*/

int NetWriteRaw
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataLength
)

{
#ifdef IP_NETLIB
   static $DESCRIPTOR (DataPtrDsc, "");
#endif

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetWriteRaw() %d %d %d %d\n",
               rqptr, AstFunctionPtr, DataPtr, DataLength);
   /** fprintf (stdout, "|%*.*s|\n", DataLength, DataLength, DataPtr); **/

   if (rqptr->NetWriteRawAstFunctionPtr != NULL)
   {
      rqptr->NetWriteIOsb.Status = SS$_BUGCHECK;
      rqptr->NetWriteIOsb.Count = 0;
      if (AstFunctionPtr != NULL) SysDclAst (AstFunctionPtr, rqptr);
      return (rqptr->NetWriteIOsb.Status);
   }
   rqptr->NetWriteRawAstFunctionPtr = AstFunctionPtr;

   if (AstFunctionPtr == NULL)
   {
      /***************/
      /* blocking IO */
      /***************/

#ifdef IP_UCX
      status = sys$qiow (0, rqptr->ClientChannel,
                         IO$_WRITEVBLK, &rqptr->NetWriteIOsb, 0, 0,
                         DataPtr, DataLength, 0, 0, 0, 0);
      if (Debug)
         fprintf (stdout, "sys$qiow() %%X%08.08X %%X%08.08X %d\n",
                  status, rqptr->NetWriteIOsb.Status,
                  rqptr->NetWriteIOsb.Count);
#endif

#ifdef IP_NETLIB
      DataPtrDsc.dsc$a_pointer = DataPtr;
      DataPtrDsc.dsc$w_length = DataLength;

      status = netlib_write (&rqptr->ClientNetLibSocket, &DataPtrDsc, 0, 0,
                             &rqptr->NetWriteIOsb, 0, 0);
      if (Debug)
         fprintf (stdout, "netlib_write() %%X%08.08X %%X%08.08X %d\n",
                  status, rqptr->NetWriteIOsb.Status,
                  rqptr->NetWriteIOsb.Count);
#endif

      /* for blocking IO use the status block for status information */
      if (VMSok (status) && VMSnok (rqptr->NetWriteIOsb.Status))
         status = rqptr->NetWriteIOsb.Status;
   }
   else
   {
      /*******************/
      /* non-blocking IO */
      /*******************/

#ifdef IP_UCX
      status = sys$qio (0, rqptr->ClientChannel,
                        IO$_WRITEVBLK, &rqptr->NetWriteIOsb,
                        &NetWriteRawAst, rqptr,
                        DataPtr, DataLength, 0, 0, 0, 0);
      if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
#endif

#ifdef IP_NETLIB
      DataPtrDsc.dsc$a_pointer = DataPtr;
      DataPtrDsc.dsc$w_length = DataLength;

      status = netlib_write (&rqptr->ClientNetLibSocket, &DataPtrDsc, 0, 0,
                             &rqptr->NetWriteIOsb, &NetWriteRawAst, rqptr);
      if (Debug) fprintf (stdout, "netlib_write() %%X%08.08X\n", status);
#endif
   }

   /****************/
   /* check status */
   /****************/

   if (VMSok (status)) return (status);

   /* if resource wait enabled the only quota not waited for is ASTLM */
   if (status == SS$_EXQUOTA)
      ErrorExitVmsStatus (status, "sys$qio()", FI_LI);

   /* write failed, call AST explicitly, status in the IOsb */
   rqptr->NetWriteIOsb.Status = status;
   rqptr->NetWriteIOsb.Count = 0;
   SysDclAst (NetWriteRawAst, rqptr);
   return (status);
}

/*****************************************************************************/
/*
AST from NetWriteRaw().  Call the AST function.
*/ 

NetWriteRawAst (struct RequestStruct *rqptr)

{
   int  status;
   void (*AstFunctionPtr)(struct RequestStruct*);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetWriteRawAst() %%X%08.08X %d\n",
               rqptr->NetWriteIOsb.Status, rqptr->NetWriteIOsb.Count);

   if (rqptr->AdminWatchItem &&
       (AdminWatchEnabled & ADMIN_WATCH_NETWORK))
   {
      if (VMSok (rqptr->NetWriteIOsb.Status))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_NETWORK,
                         "WRITE %X!8XL !UL bytes",
                         rqptr->NetWriteIOsb.Status,
                         rqptr->NetWriteIOsb.Count);
      else
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_NETWORK,
                         "WRITE %X!8XL !AZ",
                         rqptr->NetWriteIOsb.Status,
                         SysGetMsg(rqptr->NetWriteIOsb.Status));
   }

   if (VMSok (rqptr->NetWriteIOsb.Status))
      rqptr->BytesRawTx += rqptr->NetWriteIOsb.Count;

   if ((AstFunctionPtr = rqptr->NetWriteRawAstFunctionPtr) == NULL) return;
   rqptr->NetWriteRawAstFunctionPtr = NULL;
   (*AstFunctionPtr)(rqptr);
}

/*****************************************************************************/
/*
Wrapper for NetReadRaw().
*/ 

int NetRead
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataSize
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetRead()\n");

   if (rqptr->SeSoLaPtr == NULL)
      status = NetReadRaw (rqptr, AstFunctionPtr, DataPtr, DataSize);
   else
      status = SeSoLaRead (rqptr, AstFunctionPtr, DataPtr, DataSize);
   if (Debug) fprintf (stdout, "status: %%X%08.08X\n", status);
   return (status);
}

/*****************************************************************************/
/*
Queue up a read from the client over the network. If 'AstFunctionPtr' 
is zero then no I/O completion AST routine is called.  If it is non-zero then 
the function pointed to by the parameter is called when the network write 
completes.

Explicitly declares any AST routine if an error occurs. The calling function
must not do any error recovery if an AST routine has been supplied but the
associated AST routine must!  If an AST was not supplied then the return
status can be checked.
*/ 

int NetReadRaw
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataSize
)
{
#ifdef IP_NETLIB
   static $DESCRIPTOR (DataPtrDsc, "");
#endif

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetReadRaw() %d %d %d %d\n",
               rqptr, AstFunctionPtr, DataPtr, DataSize);

   if (rqptr->NetReadRawAstFunctionPtr != NULL)
   {
      rqptr->NetReadIOsb.Status = SS$_BUGCHECK;
      rqptr->NetReadIOsb.Count = 0;
      if (AstFunctionPtr != NULL) SysDclAst (AstFunctionPtr, rqptr);
      return (rqptr->NetReadIOsb.Status);
   }
   rqptr->NetReadRawAstFunctionPtr = AstFunctionPtr;

   if (AstFunctionPtr == NULL)
   {
      /***************/
      /* blocking IO */
      /***************/

#ifdef IP_UCX
      status = sys$qiow (0, rqptr->ClientChannel,
                         IO$_READVBLK, &rqptr->NetReadIOsb, 0, 0,
                         DataPtr, DataSize, 0, 0, 0, 0);
      if (Debug)
         fprintf (stdout, "sys$qiow() %%X%08.08X %%X%08.08X %d\n",
                  status, rqptr->NetReadIOsb.Status,
                  rqptr->NetReadIOsb.Count);
#endif

#ifdef IP_NETLIB
      DataPtrDsc.dsc$a_pointer = DataPtr;
      DataPtrDsc.dsc$w_length = DataSize;

      status = netlib_read (&rqptr->ClientNetLibSocket, &DataPtrDsc, 0, 0, 0, 0,
                            &rqptr->NetReadIOsb, 0, 0);
      if (Debug)
         fprintf (stdout, "netlib_read() %%X%08.08X %%X%08.08X %d\n",
                  status, rqptr->NetReadIOsb.Status,
                  rqptr->NetReadIOsb.Count);
#endif

      /* for blocking IO use the status block for status information */
      if (VMSok (status) && VMSnok (rqptr->NetReadIOsb.Status))
         status = rqptr->NetReadIOsb.Status;
   }
   else
   {
      /*******************/
      /* non-blocking IO */
      /*******************/

#ifdef IP_UCX
      status = sys$qio (0, rqptr->ClientChannel,
                        IO$_READVBLK, &rqptr->NetReadIOsb,
                        &NetReadRawAst, rqptr,
                        DataPtr, DataSize, 0, 0, 0, 0);
      if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
#endif

#ifdef IP_NETLIB
      DataPtrDsc.dsc$a_pointer = DataPtr;
      DataPtrDsc.dsc$w_length = DataSize;

      status = netlib_read (&rqptr->ClientNetLibSocket, &DataPtrDsc, 0, 0, 0, 0,
                            &rqptr->NetReadIOsb, &NetReadRawAst, rqptr);
      if (Debug) fprintf (stdout, "netlib_read() %%X%08.08X\n", status);
#endif
   }

   /****************/
   /* check status */
   /****************/

   /* if I/O successful */
   if (VMSok (status)) return (status);

   /* with resource wait enabled the only quota not waited for is ASTLM */
   if (status == SS$_EXQUOTA)
      ErrorExitVmsStatus (status, "sys$qio()", FI_LI);

   /* queuing of read failed, call AST explicitly, status in the IOsb */
   rqptr->NetReadIOsb.Status = status;
   rqptr->NetReadIOsb.Count = 0;
   SysDclAst (NetReadRawAst, rqptr);
   return (status);
}

/*****************************************************************************/
/*
AST from NetReadRaw().  Call the AST function.
*/ 

NetReadRawAst (struct RequestStruct *rqptr)

{
   int  status;
   void (*AstFunctionPtr)(struct RequestStruct*);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetReadRawAst() %%X%08.08X %d\n",
               rqptr->NetReadIOsb.Status, rqptr->NetReadIOsb.Count);

   if (rqptr->AdminWatchItem &&
       (AdminWatchEnabled & ADMIN_WATCH_NETWORK))
   {
      if (VMSok(rqptr->NetReadIOsb.Status))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_NETWORK,
                         "READ %X!8XL !UL bytes",
                         rqptr->NetReadIOsb.Status,
                         rqptr->NetReadIOsb.Count);
      else
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_NETWORK,
                         "READ %X!8XL !AZ",
                         rqptr->NetReadIOsb.Status,
                         SysGetMsg(rqptr->NetReadIOsb.Status));
   }

   if (VMSok (rqptr->NetReadIOsb.Status))
   {
      rqptr->BytesRawRx += rqptr->NetReadIOsb.Count;
      /* zero bytes with a normal status is a definite no-no (TGV-Multinet) */
      if (!rqptr->NetReadIOsb.Count) rqptr->NetReadIOsb.Status = SS$_ABORT;
   }

   if ((AstFunctionPtr = rqptr->NetReadRawAstFunctionPtr) == NULL) return;
   rqptr->NetReadRawAstFunctionPtr = NULL;
   (*AstFunctionPtr)(rqptr);
}

/*****************************************************************************/
/*
This function buffers output without actually sending it to the client until
ready, either when the first buffer fills or when all output is completely
buffered. It will store any amount of output in a linked list of separately
allocated buffers. This is useful when a function that must not be interrupted
can rapidly store all required output without being slowed by actually
transfering it on the network, then flush it all asynchronously (e.g. the
server administration reports, for instance CacheReport(), which are
blocking).

Providing an AST parameter and a data parameter implies that after the first
buffer fills (and usually overflows into a second) the current buffered output
should be written to the client.

Providing an AST parameter and setting the data parameter to NULL indicates
all the currently buffered contents should be written to the client.

Providing all parameters as NULL and zero, (except 'rqptr') as appropriate,
results in the data buffer initialized (if it didn't exist) or reset (if it
did).
*/ 

NetWriteBuffered
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataLength
)
{
   int  status;
   char  *BufferPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetWriteBuffered() %d %d %d %d\n",
               rqptr, AstFunctionPtr, DataPtr, DataLength);

   /* if all parameters essentially empty then reset the buffers */
   if (AstFunctionPtr == NULL && DataPtr == NULL && DataLength == 0)
   {
      NetWriteBufferedInit (rqptr, true);
      return;
   }

   /* first call, initialize a buffer */
   if (rqptr->OutputBufferStructPtr == NULL)
      NetWriteBufferedInit (rqptr, true);

   /**********************/
   /* buffer this output */
   /**********************/

   if (rqptr->NetWriteBufferedEscapeHtml)
   {
      /* escape HTML-forbidden characters */
      NetWriteBufferedEscapeHtml (rqptr, DataPtr, DataLength);
   }
   else
   {
      while (DataLength)
      {
         if (Debug)
            fprintf (stdout, "%d %d\n",
                     rqptr->OutputBufferRemaining, DataLength);

         if (DataLength <= rqptr->OutputBufferRemaining)
         {
            /* enough space in this buffer */
            memcpy (rqptr->OutputBufferCurrentPtr, DataPtr, DataLength);
            rqptr->OutputBufferCount += DataLength;
            rqptr->OutputBufferCurrentPtr += DataLength;
            rqptr->OutputBufferRemaining -= DataLength;
            DataLength = 0;
         }
         else
         {
            /* fill up any that's left */
            memcpy (rqptr->OutputBufferCurrentPtr,
                    DataPtr,
                    rqptr->OutputBufferRemaining);
            DataPtr += rqptr->OutputBufferRemaining;
            DataLength -= rqptr->OutputBufferRemaining;
            rqptr->OutputBufferCount += rqptr->OutputBufferRemaining;
            /* need another buffer */
            NetWriteBufferedInit (rqptr, false);
         }
      }
   }

   /* if an AST routine supplied then we can think about buffer flushing */
   if (AstFunctionPtr != NULL)
   {
      /* if the first buffer is full or if the flush is being forced */
      if (rqptr->OutputBufferCount > OutputBufferSize || DataPtr == NULL)
      {
         /***********************/
         /* flush the buffer(s) */
         /***********************/

         /* force all buffers to be flushed if 'DataPtr' is NULL */
         if (DataPtr == NULL) rqptr->OutputBufferFlush = true;
         /* buffer the AST function pointer */
         rqptr->OutputBufferAstFunctionPtr = AstFunctionPtr;
         /* get the "write now" pointer to the first buffer in the list */
         rqptr->OutputBufferNowStructPtr =
            LIST_HEAD (&rqptr->OutputBufferList);
         /* fudge this status for NetWriteBufferedNow() */
         rqptr->NetWriteIOsb.Status = SS$_NORMAL;
         NetWriteBufferedNow (rqptr);
         return;
      }

      /************************/
      /* just declare the AST */
      /************************/

      SysDclAst (AstFunctionPtr, rqptr);
   }
}

/*****************************************************************************/
/*
Copy the data into the buffer escaping HTML-forbidden characters.
*/ 

NetWriteBufferedEscapeHtml
(
struct RequestStruct *rqptr,
char *DataPtr,
int DataLength
)
{
   register int  dcnt, bcnt, brcnt;
   register char  *bptr, *dptr, *eptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetWriteBufferedEscapeHtml() %d %d\n",
               DataPtr, DataLength);

   dptr = DataPtr;
   dcnt = DataLength;
   bptr = rqptr->OutputBufferCurrentPtr;
   brcnt = rqptr->OutputBufferRemaining;
   bcnt = rqptr->OutputBufferCount;
   eptr = "";

   while (dcnt)
   {
      /* if there is buffer space left then process this character */
      if (brcnt)
      {
         switch (*dptr)
         {
            case '<' :
               eptr = "&lt;"; dptr++; dcnt--; break;
            case '>' :
               eptr = "&gt;"; dptr++; dcnt--; break;
            case '&' :
               eptr = "&amp;"; dptr++; dcnt--; break;
            default :
               eptr = ""; *bptr++ = *dptr++; dcnt--; brcnt--; bcnt++;
         }

         /* buffer any escape sequence from this character */
         while (*eptr && brcnt)
         {
            *bptr++ = *eptr++;
            brcnt--;
            bcnt++;
         }
      }

      if (brcnt) continue;

      /* need another buffer */
      NetWriteBufferedInit (rqptr, false);
      bptr = rqptr->OutputBufferCurrentPtr;
      brcnt = rqptr->OutputBufferRemaining;

      /* buffer any escape sequence remaining from the previous character */
      while (*eptr && brcnt)
      {
         *bptr++ = *eptr++;
         brcnt--;
         bcnt++;
      }
   }

   rqptr->OutputBufferCurrentPtr = bptr;
   rqptr->OutputBufferRemaining = brcnt;
   rqptr->OutputBufferCount = bcnt;
}

/*****************************************************************************/
/*
This function processes the linked list of output buffers, writing them to the
client, and operates in two distinct modes. First, non-flush mode. In this
mode any full output buffers are written to the client. Any last,
partially-filled buffer is not, instead it is moved to the head of the list to
essentially become the first data in the output buffer to which more can then
be added. In this way when a buffer fills and overflows into a second buffer
the first is written but the second, scantily filled buffer is not. The second
mode is full-flush, where all buffers are written to the client. This is
forced by NetWriteBuffered() whenever an AST routine is supplied but no data
('DataPtr' is NULL).
*/ 

NetWriteBufferedNow (struct RequestStruct *rqptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetWriteBufferedNow() %d %d %d %d %d\n",
               rqptr, rqptr->OutputBufferStructPtr,
               rqptr->OutputBufferRemaining,
               rqptr->OutputBufferCount,
               rqptr->OutputBufferFlush);

   if (VMSnok (rqptr->NetWriteIOsb.Status))
   {
      /* network write has failed (as AST), bail out now */
      if (Debug)
         fprintf (stdout, "NetWriteIOsb.Status %%X%08.08X\n",
                  rqptr->NetWriteIOsb.Status);
      SysDclAst (rqptr->OutputBufferAstFunctionPtr, rqptr);
      return;
   }

   if ((rqptr->OutputBufferFlush &&
        rqptr->OutputBufferCount > OutputBufferSize) ||
       rqptr->OutputBufferCount > OutputBufferSize + OutputBufferSize)
   {
      /*
         If flushing all buffers and at least one full buffer in list, or
         if two or more full buffers in list.
      */
      NetWrite (rqptr, &NetWriteBufferedNow,
                LIST_DATA (rqptr->OutputBufferNowStructPtr),
                OutputBufferSize);
      rqptr->OutputBufferCount -= OutputBufferSize;
      /* set the "write now" buffer pointer to the next buffer */
      rqptr->OutputBufferNowStructPtr =
         LIST_NEXT (rqptr->OutputBufferNowStructPtr);
      return;
   }
   else
   if (rqptr->OutputBufferCount > OutputBufferSize)
   {
      /* if one full buffer, then one partially filled, NOTE the AST! */
      NetWrite (rqptr, rqptr->OutputBufferAstFunctionPtr,
                LIST_DATA (rqptr->OutputBufferNowStructPtr),
                OutputBufferSize);
      rqptr->OutputBufferCount -= OutputBufferSize;
      /*
         Move currently-being-written-to buffer to the head of the list,
         effectively to the front of the buffer.
      */
      ListMoveHead (&rqptr->OutputBufferList, rqptr->OutputBufferStructPtr);
      return;
   }
   else
   {
      /* if last buffer in linked list, write whatever remains in it */
      NetWrite (rqptr, rqptr->OutputBufferAstFunctionPtr,
                LIST_DATA (rqptr->OutputBufferNowStructPtr),
                rqptr->OutputBufferCount);
      /* reset the buffer(s) */
      NetWriteBufferedInit (rqptr, true);
      /* no more buffers therefore flush (if set) is complete */
      rqptr->OutputBufferFlush = false;
      return;
   }
}

/*****************************************************************************/
/*
Get another output buffer.
*/ 

NetWriteBufferedInit
(
struct RequestStruct *rqptr,
boolean ResetBuffers
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "++++++++++++++++++++++++++++++++++\n");
   if (Debug) fprintf (stdout, "NetWriteBufferedInit()\n");

   if (ResetBuffers && !LIST_EMPTY (&rqptr->OutputBufferList))
   {
      /* reset a non-empty list to the first buffer */
      if (Debug) fprintf (stdout, "reset non-empty\n");
      rqptr->OutputBufferStructPtr = LIST_HEAD (&rqptr->OutputBufferList);
      rqptr->OutputBufferPtr = LIST_DATA (rqptr->OutputBufferStructPtr);
      rqptr->OutputBufferCurrentPtr = rqptr->OutputBufferPtr;
      rqptr->OutputBufferRemaining = OutputBufferSize;
      rqptr->OutputBufferCount = 0;
      return;
   }

   if (rqptr->OutputBufferStructPtr != NULL &&
       LIST_HAS_NEXT (rqptr->OutputBufferStructPtr))
   {
      /* reuse a previously allocated buffer from the list */
      if (Debug) fprintf (stdout, "reuse\n");
      rqptr->OutputBufferStructPtr =
         LIST_NEXT (rqptr->OutputBufferStructPtr);
      rqptr->OutputBufferPtr = LIST_DATA (rqptr->OutputBufferStructPtr);
      rqptr->OutputBufferCurrentPtr = rqptr->OutputBufferPtr;
      rqptr->OutputBufferRemaining = OutputBufferSize;
      return;
   }

   /* allocate the first or another buffer to the list */
   if (Debug) fprintf (stdout, "new\n");
   rqptr->OutputBufferStructPtr =
      VmGetHeap (rqptr, sizeof(struct ListEntryStruct) + OutputBufferSize);
   ListAddTail (&rqptr->OutputBufferList, rqptr->OutputBufferStructPtr);
   rqptr->OutputBufferPtr = LIST_DATA (rqptr->OutputBufferStructPtr);
   rqptr->OutputBufferCurrentPtr = rqptr->OutputBufferPtr;
   rqptr->OutputBufferRemaining = OutputBufferSize;
}

/*****************************************************************************/
/*
Report that client host name has been denied access to the server.
*/
 
#ifdef IP_UCX
   NetAccessDenied
(
unsigned short Channel,
char *HostName,
char *IpAddressString
)
#endif
#ifdef IP_NETLIB
   NetAccessDenied
(
unsigned long *NetLibSocketPtr,
char *HostName,
char *IpAddressString
)
#endif
 
{
   static $DESCRIPTOR (MessageFaoDsc,
"HTTP/1.0 403 Forbidden\r\n\
Content-Type: text/html!AZ!AZ\r\n\
\r\n\
<HTML>\n\
<HEAD>\n\
<TITLE>!AZ 403</TITLE>\n\
</HEAD>\n\
<BODY>\n\
!AZ\n\
</BODY>\n\
</HTML>\n");

   short  Length;
   char  Buffer [512];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Config.CharsetDefault[0])
      sys$fao (&MessageFaoDsc, &Length, &BufferDsc,
         "; charset=", Config.CharsetDefault,
         MsgForNoRequest(HostName,IpAddressString,MSG_STATUS_ERROR),
         MsgForNoRequest(HostName,IpAddressString,MSG_GENERAL_ACCESS_DENIED));
   else
      sys$fao (&MessageFaoDsc, &Length, &BufferDsc,
         "", "",
         MsgForNoRequest(HostName,IpAddressString,MSG_STATUS_ERROR),
         MsgForNoRequest(HostName,IpAddressString,MSG_GENERAL_ACCESS_DENIED));
   Buffer[Length] = '\0';

#ifdef IP_UCX
   NetDirectWrite (Channel, Buffer, Length);
#endif
#ifdef IP_NETLIB
   NetDirectWrite (NetLibSocketPtr, Buffer, Length);
#endif
}

/*****************************************************************************/
/*
Report that there is already the limit of client connections.
*/
 
#ifdef IP_UCX
   NetTooBusy
(
unsigned short Channel,
char *HostName,
char *IpAddressString
)
#endif
#ifdef IP_NETLIB
   NetTooBusy
(
unsigned long *NetLibSocketPtr,
char *HostName,
char *IpAddressString
)
#endif
 
{
   static $DESCRIPTOR (MessageFaoDsc,
"HTTP/1.0 502 Busy\r\n\
Content-Type: text/html!AZ!AZ\r\n\
\r\n\
<HTML>\n\
<HEAD>\n\
<TITLE>!AZ 502</TITLE>\n\
</HEAD>\n\
<BODY>\n\
!AZ\n\
</BODY>\n\
</HTML>\n");

   short  Length;
   char  Buffer [512];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Config.CharsetDefault[0])
      sys$fao (&MessageFaoDsc, &Length, &BufferDsc,
         "; charset=", Config.CharsetDefault,
         MsgForNoRequest(HostName,IpAddressString,MSG_STATUS_ERROR),
         MsgForNoRequest(HostName,IpAddressString,MSG_GENERAL_TOO_BUSY));
   else
      sys$fao (&MessageFaoDsc, &Length, &BufferDsc,
         "", "",
         MsgForNoRequest(HostName,IpAddressString,MSG_STATUS_ERROR),
         MsgForNoRequest(HostName,IpAddressString,MSG_GENERAL_TOO_BUSY));
   Buffer[Length] = '\0';

#ifdef IP_UCX
   NetDirectWrite (Channel, Buffer, Length);
#endif
#ifdef IP_NETLIB
   NetDirectWrite (NetLibSocketPtr, Buffer, Length);
#endif
}

/****************************************************************************/

