/*****************************************************************************/
/*
                                  Auth.c


    THE GNU GENERAL PUBLIC LICENSE APPLIES DOUBLY TO ANYTHING TO DO WITH
                    AUTHENTICATION AND AUTHORIZATION!

    This package is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License, or any later
    version.

>   This package is distributed in the hope that it will be useful,
>   but WITHOUT ANY WARRANTY; without even the implied warranty of
>   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
>   GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


Overview
--------

This is "path"-based authorization/authentication.  That is, access is
controlled according to the path specified in the request.  A path can have
capabilities, or access controls, allowing any combination of DELETE, GET,
HEAD, POST and PUT method capabilities.  The world can have capabilities
against a path. An authenticated user also can have capabilities.  The request
capability is the logical AND of the path and world/user capabilities, meaning
the request gets the minimum of the path and user's access.  There are no
access-control files in any data directory.  It is all under the control of the
HTTPd server manager.

AUTHORIZATION IS ALWAYS PERFORMED AGAINST THE REQUEST'S PATHINFO STRING (the
string parsed from the original request path), NOT AGAINST ANY RE-MAPPED PATH.
THERE SHOULD BE NO AMBIGUITY ABOUT WHAT IS BEING AUTHORIZED!

When a request contains a controlled path the authorization for access to that
path is made against a password database for the authentication "realm" the
path belongs to.  A realm (or group) name is a 1 to 31 character string
describing the collection of one or more paths for purposes of authenticating
against a common password file.  A realm/group name should contain only alpha-
numeric, underscore and hyphen characters.  See below for a description of the
difference between realms and groups.  REITERACTION: THE REALM NAME IS WHAT
THE USER-SUPPLIED PASSWORD IS CHECKED AGAINST.

Five realms are reserved, "EXTERNAL", "NONE", "PROMISCUOUS", "VMS" and "WORLD".

  o  "EXTERNAL"     scripts authenticating their own challenge/response,
                    realm, group, user names and plain-text password are
                    available to the script for it's own authorization

  o  "NONE"         any request, is not authenticated,
                    all authorization details are left empty 

  o  "PROMISCUOUS"  only ever mapped when /PROMISCUOUS in use

  o  "VMS"          SYSUAF authentication,
                    via the SYSUAF, or restricted to accounts with identifiers
                    prime/secondary days & network/remote hours will restrict

  o  "WORLD"        any request, is not authenticated,
                    both the realm and username are set to "WORLD"

Note that VMS can have one or more synonym realm names which can be used to
disguise the fact SYSUAF authentication is in use.  Just specify the synonym
name followed immediately by "=VMS", as in the following example: "TESTING=VMS"

A plain-text description can also be associated with a realm name, providing
a more informative message to the browser user during the username/password
dialog.  This must be specified as a double-quote enclosed string preceding the
realm name.  Here are a couple of examples:

  ["the server system's SYSUAF"=VMS]
  ["just a contrived example"=VMS]

Also see the topics "SYSUAF-authentication by WASD Identifier" and
"SYSUAF-authentication by non-WASD Identifier" below.

A path specification begins at the root and usually continues down to an
asterisk indicating a match against any path beginning with that string.  Paths
may overlap.  That is a parent directory may have subdirectory trees that
differ in access controls.  The same path specification may not belong to more
than one realm at a time :^)


User Cancellation of Authentication
-----------------------------------

A user may cancel authentication for any path by submitting a request using
that path and a query string of "?httpd=logout".  An authentication
username/password dialog will be presented by the browser.  The user should
clear both username and password fields and submit.  The same dialog will be
represented at which time it should be cancelled resulting in an authentication
failure report.  The browser should have recognised this and removed the path's
cached authentication information.


Failure Reports
---------------

Authentication failures are reported to the process log.  Three messages can be
generated.  First, %HTTPD-W-AUTHFAIL, if the request is not authenticated due
to there being no such user or a password mismatch.  Isolated instances of this
are only of moderate interest.  Consecutive instances may indicate a user
thrashing about for the correct password, but they usually give up before a
dozen attempts.  The second message, %HTTPD-I-AUTHFAILOK (informational),
occurs if there has been at least one failed attempt to authenticate before a
successful attempt.  The third, %HTTPD-W-AUTHFAILIM, is of greater concern and
occurs if the total number of failed attempts exceeds the configuration limit
(such repeated attempts will always continue to fail for if this limit is
reached successive attempts are automatically failed).


Configuration
-------------

A collection of paths is controlled by a "realm" authentication source, against
which the request username/password is verified, and optionally one or two
"group" sources, from which the username's capabilities are determined (what
HTTP methods the username can apply against the path). 

Authentication sources:

  o  system SYSUAF account
  o  HTA database (WASD-specific, non-plain-text file)
  o  simple, plain-text list of names and unencrypted passwords

Authorization sources:

  o  account possession of specified VMS rights identifiers
     (only for SYSUAF authenticated requests)
  o  HTA database
  o  simple list

A single file contains authorization configuration directives.
The realm directive comprises the authentication source and zero, one or two
grouping sources, in the following format:

  [authentication-source;full-access-group;read-only-group]

Optional method or access keywords (e.g. GET, POST, R+W) may be appended to
specify default access for any paths belonging to the realm.  Multiple "realm"
directives for the same realm name may be included, each with its own trailing
paths.  These multiple declarations just accumulate.

All paths specified after a "realm" directive, up until the next "realm"
directive will belong to that realm and use the specified authentication
source. Paths are specified with a leading slash (i.e. from the root) down to
either the last character or a wildcard asterisk.

HTTP method keywords, or the generic "READ", "R", "READ+WRITE", "R+W", "WRITE"
and "W", control access to the path (in combination with user access methods).
The method keywords are "DELETE", "GET" (implying "HEAD"), "HEAD" "PUT" and
"POST". These keywords are optionally supplied following a path.  If none are
supplied the realm defaults are applied.  The generic keyword "NONE" sets all
access permissions off, including realm defaults.  It is an error not to supply
either. Multiple method keywords may be separated by commas or spaces.  Method
keywords are applied in two groups.  First the group keywords.  Second, an
optional set for controlling world access.  These are delimited from the path
keywords by a semi-colon.

As part of the path access control a list of comma-separated elements may
optionally be supplied.  These elements may be numeric or alphabetic IP host
addresses (with or without asterisk wildcarding allowing wildcard matching of
domains), authenticated username(s) or the request scheme (protocol, "http:" or
"https:").  If supplied this list restricts access to those matching.  For
example:

   *.wasd.dsto.gov.au       would forbid any host outside the WASD subdomain
   131.185.250.*            similarly, but specifying using numeric notation
   *.wasd.dsto.gov.au,~daniel
   https:,*.131.185.250.*   limited to a subnet requesting via SSL service

The reserved word "localhost", or used as "#localhost", refers to the system
the server is executing on.  In this way various functions can be restricted to
browsers executing only on that same system.

Realm names, paths, IP addresses, usernames and access method keywords are all
case insensistive.

Virtual server syntax is the same as for path mapping rules,
[[virtual-host]] (all ports of host) or [[virtual-host:virtual-port]] (
matching specified port of host) for a specific host and [[*]] to resume rules
for all virtual hosts.

Comments may be included by prefixing the line with a "#" characters.  Blank
lines are ignored.  Directives may span multiple lines by ensuring the last
character on any line to be continued is a backslash ("\").

Misconfigured paths, that might otherwise be left without access control, are
reported during server startup, loaded into the configuration and marked
"fail"ed, and will always deny access.  This does not however guarantee that
access control will always work the way intended!!

An example configuration file:

   |[WASD;CONFIDENTIAL]
   |/web/confidential/* get
   |/web/confidential/submissions/* https:,get,post;get
   |/web/authors/* get
   |
   |# realm-context access default
   |[WASD] get,head
   |/web/pub/*
   |/web/pub/submit/* ;get
   |
   |# realm plus full-access and read-only groups
   |[VMS;FULL=LIST;READ=LIST]
   |/web/project/jin/* r+w
   |
   |[WORLD]
   |/web/anyone-can-post-here/* post
   |/web/example/alpha/* *.wasd.dsto.gov.au,read
   |/web/example/numeric/* \
   |   131.185.250.*,read
   |
   |[EXTERNAL]
   |/cgi-bin/footypix*


HTA Databases
-------------

WASD authentication/authorization databases (which have the extension .HTA
accounting for their generic name) are binary-content content files with fixed
512 byte records.  They may only be administered via the server administration
menu.  They provide for encrypted storage of passwords and can also be used to
store groups of usernames for determining group membership.  All HTA databases
must be located in the directory specified by the logical HT_AUTH:

HTA databases are the default for realm and group names.

For example: [VMS;ADMIN;USER]


Simple Lists
------------

Plain-text files may also be used to store collections of usernames.  This is
primarily intended as a simple yet effective means of collecting together names
for full-access and read-only access group membership.  The file should contain
one username at the beginning of each line.  Blank and comment lines are
ignored.  White-space delimited text following the username is ignored (with
the exception noted below).  The format for each line is
"username other text if required ...".  Simple lists have the extension .HTL
and must be stored in the directory specified by the logical name HT_AUTH:

Although it is better to use another, more secure authentication database,
these lists may have have plain-text "passwords" associated with the usernames. 
As these are unencrypted they should not be used for any security-sensitive
purpose, but may suffice for ad hoc or other simple uses.  The format for
specifying a password is "username=password other text if required ...".

Simple lists are indicated by appending "=LIST" to the realm or group name. 

For example: [VMS;ADMIN=LIST;USER=LIST]


VMS Identifiers
---------------

The qualifier /SYSUAF=ID directs the server to allow SYSUAF authentication only
when the VMS account possesses a specified identifier.  NO IDENTIFIER, NO
ACCESS, NO MATTER HOW MANY CORRECT PASSWORDS SUPPLIED.  An excellent
combination is /SYSUAF=(ID,SSL)!

VMS identifiers may be used to control who may authenticate using the SYSUAF,
who is a member of a full-access group, and who is a member of a read-only
group.  These may be any existing or web-specific VMS identifier name possessed
by a given account.  Standard [realm;group-r+w;group-r] syntax is employed,
with a trailing "=ID..." indicating it is an identifier.

For example: [VMS;JIN_PROJECT=ID;JIN_LIBRARIAN=ID]

For a realm specification without group membership requirements (i.e.
[realm;;]) full-access is granted to the username (this is of course adjusted
according to the path specified access level).  When used for determining group
membership, holding the identifier specified in the full-access group (i.e.
[realm;GROUP-R+W;group-r]) provides full read and write access.  If the
full-access identifier is not held by the user but the read-only one is (i.e.
[realm;group-r+w;GROUP-R]) the username receives read access.

NOTE: VMS Identifier identifiers may only be used for specifying group
membership for realms authenticated using the SYSUAF.  Specifying an identifier
group for any other authentication source results in an authorization
configuration error.

If an account holds the reserved identifier WASD_HTTPS_ONLY then it may only
authenticate from the SYSUAF with an SSL (https:) request.  Standard http: is
always denied.

Identifier names may contain 1 to 31 characters.

The following examples should help clarify this description (note that the
identifier names are completely arbitrary).

  [JIN_PROJECT=ID]
  # If an account holds the JIN_PROJECT identifier that username and
  # password may be used for authentication for this realm's paths.
  # Note that this account gets read and write access against the username,
  # subject to the path's access specification (also read+write in this case).
  /web/project/jin/* r+w

  [JIN_PROJECT=ID;JIN_LIBRARIAN=ID;JIN_USER=ID]
  # If an account is a holder of the JIN_PROJECT identifier then the account
  # password may be used for server authentication for this realm's paths.
  # If the account also holds the JIN_LIBRARIAN will it be allowed write
  # access, if JIN_USER then read access, to the realm's paths.
  /web/project/jin/library/* r+w

  [JIN_PROJECT=ID;JIN_CODE=ID]
  # Only if the account possesses both the JIN_PROJECT and JIN_CODE
  # identifiers will the username get read+write access to realm's paths.

  [LOCAL=VMS;JIN_PROJECT=ID]
  # If an account holds the WASD_VMS_R or the WASD_VMS_RE then the username
  # can authenticate against the SYSUAF.  However, only if the account also
  # holds the JIN_PROJECT identifier can they access this realm's paths.
  # Read and/or write depends on the "WASD_VMS_..." identifier held and
  # the path access specification (read-only in this case).
  /web/project/jin/* r

  [JIN_PROJECT=ID;FELIX]
  # If an account is a holder of the JIN_PROJECT identifier then account
  # password may be used for server authentication for this realm's paths.
  # Only if the account also has the WASD_VMS__FELIX identifier will it be
  # allowed to access the realm's paths.
  /web/project/jin/* r+w


SYSUAF-authentication by WASD Identifier
----------------------------------------

THIS FUNCTIONALITY IS DEPRECATED!  In addition to general identifiers, other
"hard-wired" identifiers may be used to control access to and of VMS accounts. 
If a username has been authenticated using using one of the read or write
"hard-wired" identifiers then any group membership must be determined using the
"hard-wired" WASD_VMS__<group> identifier.  Only a realm and one grouping is
allowed.

  o  WASD_VMS_R .......... account has read access
  o  WASD_VMS_RW ......... account has read and write access
  o  WASD_VMS__<group> ... account must possess identifier to access
  o  WASD_VMS_HTTPS ...... account can only SYSUAF authenticate via SSL
  o  WASD_VMS_PWD ........ account can change it's SYSUAF password


SYSUAF-authentication and VMS security profile
----------------------------------------------

The ability to be able to authenticate using the system's SYSUAF is controlled
by the server /SYSUAF qualifier.  By default it is disabled.  (This qualifier
was introduced in v4.4 as a result of creeping paranoia :^)

As of v4.4 it has become possible to control access to files and directories
based on the security profile of a SYSUAF-authenticated remote user. This
feature is controlled using the server /PROFILE qualifier.  By default it is
disabled.

A security-profile is created using sys$create_user_profile() and stored in
the authentication cache. This cached profile is the most efficient method of
implementing this as it obviously removes the need of creating a user profile
each time a resource is checked. If this profile exists in the cache it is
attached to each request structure authenticated via the cache.

When this profile is attached to a request all accesses to files and
directories are first assessed against that user profile using
sys$check_access(). If the user can access the resource (regardless of whether
that is because it is WORLD accessable, or because it is owned by, or
otherwise accessable to the user) SYSPRV is always enabled before accessing
the file/directory. This ensures that the authenticated user is always given
access to the resource (provided the SYSTEM permissions allow it!)

Of course, this functionality only provides access for the server, IT DOES NOT
PROPAGATE TO ANY SCRIPT ACCESS. If scripts must have a similar ability they
should implement their own sys$check_access() (which is not too difficult)
based on the WWW_AUTH_REALM which would be "VMS" indicating
SYSUAF-authentication, and the authenticated name in WWW_REMOTE_USER.

Note: the sys$assume_persona(), et.al., which might seem a more thorough
approach to this functionality, was not possible due to the "independently"
executing subprocesses associated with the server that might be adversely
affected by such an abrupt change in identity!


Controlling server write access
-------------------------------

Write access by the server into VMS directories should be controlled using VMS
ACLs.  World write access should not be given to any server accessed directory.
This is in addition to the path authorization of the server itself of course!
The requirement to have an ACL on the directory prevents inadvertant
mapping/authorization path being able to be written to.  Two different ACLs
control two different grades of access.

1. If the ACL grants CONTROL access to the server account then only
VMS-authenticated usernames with security profiles can potentially write to
the directory, potentially, because a further check is made to assess whether
that VMS account has write access.

This example show a suitable ACL that stays only on the original directory:

  $ SET SECURITY directory.DIR -
    /ACL=(IDENT=HTTP$SERVER,ACCESS=READ+WRITE+EXECUTE+DELETE+CONTROL)

This example shows setting an ACL that will propagate to created
subdirectories:

  $ SET SECURITY directory.DIR -
    /ACL=((IDENT=HTTP$SERVER,OPTIONS=DEFAULT,ACCESS=READ+WRITE+EXECUTE+DELETE+CONTROL), -
          (IDENT=HTTP$SERVER,ACCESS=READ+WRITE+EXECUTE+DELETE+CONTROL))

2. If the ACL grants WRITE access then the directory can be written into by
any authenticated username for the authorized path.

This example show a suitable ACL that stays only on the original directory:

  $ SET SECURITY directory.DIR -
    /ACL=(IDENT=HTTP$SERVER,ACCESS=READ+WRITE+EXECUTE+DELETE)

This example shows setting an ACL that will propagate to created
subdirectories:

  $ SET SECURITY directory.DIR -
    /ACL=((IDENT=HTTP$SERVER,OPTIONS=DEFAULT,ACCESS=READ+WRITE+EXECUTE+DELETE), -
          (IDENT=HTTP$SERVER,ACCESS=READ+WRITE+EXECUTE+DELETE))


Implementation
--------------

A linked-list, binary tree (using the the LIB$..._TREE routines) is used to 
store authentication records.  When a request with an "Authorization:" header 
line encounters a point where authentication is required this binary tree is 
checked for an existing record.  If one does not exist a new one is entered 
into the binary tree, with an empty password field.  The password in the 
authorization line is checked against the password in the record.  If it 
matches the request is authenticated.

If a binary tree record is not found the SYSUAF or REALM-based password is
checked.  If the hashed password matches, the plain password is copied into the
record and used for future authentications.  If not, the authentication fails.

Once a user name is authenticated from the password database, for a limited
period (currently 60 minutes), further requests are authenticated directly
from the binary tree.  This improves performance and reduces file system
references for authentication.  After that period expires the password and any
group access database is again explicitlty checked.  The internal, binary tree
database can be explicitly listed, flushed or re-loaded as required.

A request's authorization flags (->AuthRequestCan) governs what can and can't be
done using authorization.  These are arrived at by bit-wise ANDing of flags
from two sources ...

   o  'AuthGroupCan' flags associated with the authorization realm and path
   o  'AuthUserCan' flags associated with the authenticated user

... this ensuring that the minimum of actions permitted by either path or user
capabilities is reflected in the request capabilities.


VERSION HISTORY
---------------
12-SEP-99  MGD  virtual host modifications
05-AUG-99  MGD  authentication cancellation via "?httpd=logout",
                check UAI_NETWORK/REMOTE_ACCESS_x for access restrictions,
                identifier enabled access allows CAPTIVE and RESTRICTED,
                bugfix; check for "SSL only" after authorization cache hit,
                bugfix; revalidate user minutes update
20-FEB-99  MGD  extend authentication/authorization use of VMS identifiers,
                refine authorization with [realm;group-r+w;group-r],
                ALL paths with problems of any sort load and then always FAIL!
19-DEC-98  MGD  refine authorization WATCH information
07-NOV-98  MGD  WATCH facility
17-OCT-98  MGD  [realm-name=VMS] synonym hiding the fact it's SYSUAF,
                virtual services via "[[virtual-host:virtual-port]]"
29-AUG-98  MGD  move authorization path processing into AuthPathLine(),
                report authentication failures in process log,
                change in behaviour: after cache minutes expires request
                revalidation by the user via browser dialog
16-JUL-98  MGD  /SYSUAF=ID, authentication with possession of an identifier,
                AUTH_RESTRICTED_BY_OTHER indicates non-authentication forbidden
11-MAR-98  MGD  added local redirection kludge ('^'),
                configurable SYSUAF authentication of privileged accounts,
                bugfix; alpha-numeric hosts in access-restriction lists
                rejected as "unknown HTTP method"
08-FEB-98  MGD  provide SSL-only for SYSUAF and authorization in general,
                provide SSL-only for authorized paths (via "https:" or
                "http:" in path access restriction list),
                removed full method descriptions from user auth report
17-AUG-97  MGD  message database,
                SYSUAF-authenticated users security-profile
16-JUL-97  MGD  fixed design flaw with WORLD realm and access restriction list
01-FEB-97  MGD  HTTPd version 4
01-JUL-96  MGD  path/realm-based authorization/authentication
15-MAR-96  MGD  bugfix; some ErrorGeneral() not passing the request pointer
01-DEC-95  MGD  HTTPd version 3
01-APR-95  MGD  initial development for local (SYSUAF) authentication
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <acldef.h>
#include <armdef.h>
#include <descrip.h>
#include <chpdef.h>
#include <iodef.h>
#include <libdef.h>
#include <prvdef.h>
#include <rms.h>
#include <rmsdef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <stdarg.h>

#include <uaidef.h>
/* not defined in VAX C 3.2 <uaidef.h> */
#define UAI$M_RESTRICTED 0x8
#define UAI$C_PURDY_S 3

/* application related header files */
#include "wasd.h"

#define WASD_MODULE "AUTH"

#if DBUG
#define FI_NOLI WASD_MODULE, __LINE__
#else
/* in production let's keep the exact line to ourselves! */
#define FI_NOLI WASD_MODULE, 0
#endif

#define AUTH_RESET_REQUEST \
   rqptr->RemoteUser[0] = rqptr->RemoteUserPassword[0] = '\0'; \
   rqptr->AuthRealmPtr = \
      rqptr->AuthRealmDescriptionPtr = \
      rqptr->AuthGroupReadPtr = \
      rqptr->AuthGroupRestrictListPtr = \
      rqptr->AuthWorldRestrictListPtr = ""; \
   rqptr->AuthGroupCan = \
      rqptr->AuthScheme =  \
      rqptr->AuthSourceGroupWrite = \
      rqptr->AuthSourceGroupRead = \
      rqptr->AuthSourceRealm = \
      rqptr->AuthWorldCan = 0;


/******************/
/* global storage */
/******************/

unsigned long  AuthDefaultCan = AUTH_READONLY_ACCESS;

boolean  AuthorizationEnabled,
         AuthPolicySslOnly = false,
         AuthPolicyAuthorizedOnly = false,
         AuthPolicySysUafSslOnly = false,
         AuthPolicySysUafAllAccounts = false,
         AuthPolicySysUafIdentifiers = false,
         AuthPolicySysUafWasdIdentifiers = false,
         AuthPromiscuous = false,
         AuthSysUafEnabled = false,
         AuthSysUafPromiscuous = false,
         AuthVmsUserSecProfileEnabled = false;

int  AuthDisplayRecordCount,
     AuthFreeRecordCount;

unsigned long  AuthHttpsOnlyVmsIdentifier,
               AuthWasdPwdVmsIdentifier,
               AuthWasdHttpsVmsIdentifier,
               AuthWasdReadVmsIdentifier,
               AuthWasdWriteVmsIdentifier;

unsigned long  AuthCacheTreeBinTime [2];

char  *AuthPromiscuousPwdPtr;

struct AuthCacheRecordStruct  *AuthCacheTreeHead,
                              *AuthCacheTreeTailPtr;

struct AuthConfigStruct  AuthServerConfig;

unsigned short  HttpdUserProfileLength;
unsigned char  *HttpdUserProfilePtr;
$DESCRIPTOR (HttpdUserProfileDsc, "");

char  ErrorAuthSysFaoBufferOvf [] = "*ERROR* sys$fao() BUFFEROVF",
      ErrorAuthSysFao [] = "*ERROR* sys$fao()";

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern int  AdminWatchEnabled;
extern int  ServerPort;
extern int  ServiceCount;
extern unsigned short  ControlMbxChannel;
extern char  ErrorSanityCheck[];
extern char  HttpdUserName[];
extern char  ServerHostName[];
extern char  ServerHostPort[];
extern char  SoftwareID[];
extern char  Utility[];
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;
extern struct MsgStruct  Msgs;
extern struct ServiceStruct  *ServiceListHead;

/*****************************************************************************/
/*
Initialize the authentication/authorization environment.  Is used at server
startup and when the authorization configuration is reloaded.
*/ 

AuthInit ()

{
   static $DESCRIPTOR (AuthHttpsOnlyVmsIdentifierDsc, AUTH_HTTPS_ONLY_VMS_ID);
   static $DESCRIPTOR (AuthWasdHttpsVmsIdentifierDsc, AUTH_WASD_HTTPS_VMS_ID);
   static $DESCRIPTOR (AuthWasdPwdVmsIdentifierDsc, AUTH_WASD_PWD_VMS_ID);
   static $DESCRIPTOR (AuthWasdReadVmsIdentifierDsc, AUTH_WASD_READ_VMS_ID);
   static $DESCRIPTOR (AuthWasdWriteVmsIdentifierDsc, AUTH_WASD_WRITE_VMS_ID);

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthInit()\n");

   /* it's implied! */
   if (AuthPolicySysUafWasdIdentifiers) AuthPolicySysUafIdentifiers = true;

   /* initialize the configuration database */
   AuthPathConfig (NULL);

   if (AuthSysUafEnabled)
      AuthConfigReport (&AuthServerConfig,
         "%HTTPD-I-AUTH, SYSUAF authentication enabled\n");

   if (AuthPolicyAuthorizedOnly)
      AuthConfigReport (&AuthServerConfig,
         "%HTTPD-I-AUTH, all request paths must be authorized\n");

   if (AuthPolicySslOnly)
      AuthConfigReport (&AuthServerConfig,
         "%HTTPD-I-AUTH, only SSL requests (https:) will be authorized\n");

   if (AuthPolicySysUafIdentifiers)
      AuthConfigReport (&AuthServerConfig,
         "%HTTPD-I-AUTH, SYSUAF authentication only with identifier\n");

   if (AuthPolicySysUafWasdIdentifiers)
      AuthConfigReport (&AuthServerConfig,
"%HTTPD-I-AUTH, SYSUAF authentication only with WASD identifier (deprecated)\n");

   if (AuthVmsUserSecProfileEnabled)
      AuthConfigReport (&AuthServerConfig,
         "%HTTPD-I-AUTH, VMS security profile enabled\n");

   if (AuthSysUafPromiscuous)
      AuthConfigReport (&AuthServerConfig,
         "%HTTPD-W-AUTH, SYSUAF !AZ (testing only)\n",
         AUTH_REALM_PROMISCUOUS);

   if (AuthPromiscuous)
      AuthConfigReport (&AuthServerConfig,
          "%HTTPD-W-AUTH, !AZ authenticating any username/password!!\n",
          AUTH_REALM_PROMISCUOUS);

   /* load the authorization paths */
   AuthPathConfig (&AuthServerConfig);

   if (AuthorizationEnabled)
   {
      if (!(Config.AuthBasicEnabled || Config.AuthDigestEnabled))
         AuthConfigReport (&AuthServerConfig,
            "%HTTPD-W-AUTH, neither BASIC or DIGEST authentication enabled\n");
   }
   else
      AuthConfigReport (&AuthServerConfig,
            "%HTTPD-W-AUTH, none configured\n");

   if (AuthPolicySysUafIdentifiers)
   {
      status = sys$asctoid (&AuthHttpsOnlyVmsIdentifierDsc,
                            &AuthHttpsOnlyVmsIdentifier, 0);
      if (VMSnok (status))
         AuthConfigReport (&AuthServerConfig,
            "%HTTPD-W-AUTH, required identifier !AZ\n-!AZ\n",
            AUTH_HTTPS_ONLY_VMS_ID, SysGetMsg(status)+1);
   }

   if (AuthPolicySysUafWasdIdentifiers)
   {
      status = sys$asctoid (&AuthWasdWriteVmsIdentifierDsc,
                            &AuthWasdWriteVmsIdentifier, 0);
      if (VMSnok (status))
         AuthConfigReport (&AuthServerConfig,
            "%HTTPD-W-AUTH, required identifier !AZ\n-!AZ\n",
            AUTH_WASD_WRITE_VMS_ID, SysGetMsg(status)+1);

      status = sys$asctoid (&AuthWasdReadVmsIdentifierDsc,
                            &AuthWasdReadVmsIdentifier, 0);
      if (VMSnok (status))
         AuthConfigReport (&AuthServerConfig,
            "%HTTPD-W-AUTH, required identifier !AZ\n!-!AZ\n",
            AUTH_WASD_READ_VMS_ID, SysGetMsg(status)+1);

      status = sys$asctoid (&AuthWasdHttpsVmsIdentifierDsc,
                            &AuthWasdHttpsVmsIdentifier, 0);
      if (VMSnok (status))
         AuthConfigReport (&AuthServerConfig,
            "%HTTPD-W-AUTH, required identifier !AZ\n-!AZ\n",
            AUTH_WASD_HTTPS_VMS_ID, SysGetMsg(status)+1);

      status = sys$asctoid (&AuthWasdPwdVmsIdentifierDsc,
                            &AuthWasdPwdVmsIdentifier, 0);
      if (VMSnok (status))
         AuthConfigReport (&AuthServerConfig,
            "%HTTPD-W-AUTH, required identifier !AZ\n-!AZ\n",
            AUTH_WASD_PWD_VMS_ID, SysGetMsg(status)+1);
   }
}

/*****************************************************************************/
/*
This function is pivotal to the HTTPd's authorization and authentication.

All requests should call this function before proceding.  This function should
allow or deny access to all requests.

Other functions called by this function are expected to return one of three
status codes.  SS$_NORMAL (or other success status), to indicate authentication
success, SS$_INVLOGIN to indicate authentication failure, or a VMS error status
(error messages must be generated where and when they occur).

If a non-success status is returned by this function the calling function
should immediately conclude processing.  It is expected that error will have
been reported (and thence a message generated) where and when they occur.
*/ 

int Authorize (struct RequestStruct *rqptr)

/*
{
int  status;

Debug = 1;
status = Authorize_ (rqptr);
Debug = 0;
return status;
}

int Authorize_ (struct RequestStruct *rqptr)
*/

{
   static $DESCRIPTOR (StringFaoDsc, "Forbidden to !AZ <TT>!AZ</TT>");
   static $DESCRIPTOR (StringDsc, "");

   boolean  RestrictedBy,
            WatchThis;
   int  status;
   unsigned short  Length;
   char  String [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "Authorize() |%s|\n", rqptr->PathInfoPtr);

   if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
      WatchThis = true;
   else
      WatchThis = false;

   if (WatchThis)
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                      "PATH !AZ", rqptr->PathInfoPtr);

   if (!AuthorizationEnabled &&
       !AuthPolicyAuthorizedOnly)
   {
      /********************************************/
      /* no authorization is mapped, use defaults */
      /********************************************/

      if (WatchThis)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH, "DISABLED");

      AUTH_RESET_REQUEST

      rqptr->AuthGroupCan =
         rqptr->AuthRequestCan =
         rqptr->AuthUserCan = AuthDefaultCan;

      return (SS$_NORMAL);
   }

   /****************************/
   /* authorization is enabled */
   /****************************/

   if (AuthPolicySslOnly && rqptr->RequestScheme != SCHEME_HTTPS)
   {
      /***********************************************/
      /* policy ... authorization only when "https:" */
      /***********************************************/

      if (WatchThis)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "POLICY https: (SSL) ONLY");

      rqptr->ResponseStatusCode = 403;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_ACCESS_DENIED), FI_NOLI);

      AUTH_RESET_REQUEST

      return (STS$K_ERROR);
   }

   /**************/
   /* check path */
   /**************/

   status = AuthPathCheck (rqptr, NULL);

   if (VMSnok (status))
   {
      /*************************/
      /* not a controlled path */
      /*************************/

      if (Debug) fprintf (stdout, "not controlled!\n");

      if (AuthPolicyAuthorizedOnly)
      {
         /******************************************/
         /* policy: no authorization ... no access */
         /******************************************/

         if (WatchThis)
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "POLICY authorized paths ONLY");

         rqptr->ResponseStatusCode = 403;
         ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_ACCESS_DENIED), FI_NOLI);

         AUTH_RESET_REQUEST

         return (STS$K_ERROR);
      }

      /**********************/
      /* use default access */
      /**********************/

      if (WatchThis)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH, "NONE applies");

      /*
         If there is already a source realm set then this has been from
         authorization applied to the script portion of a script request.
         Do not reset this as in the absence of authorization on the
         "file" portion it defines the level of access for that request.
      */
      if (!rqptr->AuthSourceRealm)
      {
         AUTH_RESET_REQUEST

         rqptr->AuthGroupCan =
            rqptr->AuthRequestCan =
            rqptr->AuthUserCan = AuthDefaultCan;
      }

      return (SS$_NORMAL);
   }

   /*********************************/
   /* controlled path, authorize it */
   /*********************************/

   if (Debug)
   {
      fprintf (stdout,
"|%s|%d|%s|%d|%s|%d|\n\
GroupCan: %08.08X WorldCan: %08.08X\n\
GroupRestriction |%s|\n\
WorldRestriction |%s|\n",
      rqptr->AuthRealmPtr,
      rqptr->AuthSourceRealm,
      rqptr->AuthGroupWritePtr,
      rqptr->AuthSourceGroupWrite,
      rqptr->AuthGroupReadPtr,
      rqptr->AuthSourceGroupRead,
      rqptr->AuthGroupCan, rqptr->AuthWorldCan,
      rqptr->AuthGroupRestrictListPtr,
      rqptr->AuthWorldRestrictListPtr);
   }

   if (WatchThis)
   {
      boolean ShowRealmDescription = rqptr->AuthRealmDescriptionPtr[0] &&
                                     rqptr->AuthRealmDescriptionPtr !=
                                     rqptr->AuthRealmPtr; 

      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
         "[!AZ!AZ!AZ!AZ!AZ;!AZ!AZ;!AZ!AZ] !AZ !AZ ; !AZ !AZ",
         ShowRealmDescription ? "\"" : "",
         ShowRealmDescription ? (char*)rqptr->AuthRealmDescriptionPtr : "",
         ShowRealmDescription ? "\"=" : "",
         rqptr->AuthRealmPtr[0] ? (char*)rqptr->AuthRealmPtr : "-",
         AuthSourceString (rqptr->AuthRealmPtr, rqptr->AuthSourceRealm),
         rqptr->AuthGroupWritePtr[0] ? (char*)rqptr->AuthGroupWritePtr : "-",
         AuthSourceString (rqptr->AuthGroupWritePtr,
                           rqptr->AuthSourceGroupWrite),
         rqptr->AuthGroupReadPtr[0] ? (char*)rqptr->AuthGroupReadPtr : "-",
         AuthSourceString (rqptr->AuthGroupReadPtr,
                           rqptr->AuthSourceGroupRead),
         rqptr->AuthGroupRestrictListPtr[0] ?
            (char*)rqptr->AuthGroupRestrictListPtr : "-",
         AuthCanString (rqptr->AuthGroupCan, AUTH_CAN_FORMAT_SHORT),
         rqptr->AuthWorldRestrictListPtr[0] ?
            (char*)rqptr->AuthWorldRestrictListPtr : "-",
         AuthCanString (rqptr->AuthWorldCan, AUTH_CAN_FORMAT_SHORT));
   }

   if (rqptr->AuthSourceRealm == AUTH_SOURCE_NONE)
   {
      /******************/
      /* the NONE realm */
      /******************/

      /*
         The purpose of the "NONE" realm is to set the '->AuthRealmSource'
         so that the server knows it's passed through authorization, without
         actually requiring any authorization, etc.  Supports the use of
         'AuthPolicyAuthorizedOnly' for reducing the possibility of
         inadvertant resource access.
      */

      AUTH_RESET_REQUEST

      return (SS$_NORMAL);
   }

   if (rqptr->AuthSourceRealm == AUTH_SOURCE_FAIL)
   {
      /*************************************************/
      /* problem with the configuration, default fail! */
      /*************************************************/

      rqptr->ResponseStatusCode = 403;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_ACCESS_DENIED), FI_NOLI);

      AUTH_RESET_REQUEST

      return (STS$K_ERROR);
   }

   if (rqptr->HttpMethod & rqptr->AuthWorldCan)
   {
      /**********************/
      /* world capabilities */
      /**********************/

      if (rqptr->AuthWorldRestrictListPtr[0])
      {
         /***********************************************/
         /* check access against world restriction list */
         /***********************************************/

         if (AuthRestrictList (rqptr, rqptr->AuthWorldRestrictListPtr))
         {
            rqptr->ResponseStatusCode = 403;
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_ACCESS_DENIED), FI_NOLI);

            AUTH_RESET_REQUEST

            return (STS$K_ERROR);
         }
      }

      if (WatchThis)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
            "ACCESS world:!AZ",
            AuthCanString (rqptr->AuthWorldCan, AUTH_CAN_FORMAT_SHORT));

      rqptr->AuthRequestCan = rqptr->AuthWorldCan;
      return (SS$_NORMAL);
   }

   if (rqptr->AuthSourceRealm == AUTH_SOURCE_EXTERNAL)
   {
      /**********************************/
      /* externally (script) authorized */
      /**********************************/

      if (WatchThis)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
            "ACCESS external:!AZ",
            AuthCanString (rqptr->AuthGroupCan, AUTH_CAN_FORMAT_SHORT));

      status = AuthParseAuthorization (rqptr);

      rqptr->AuthRequestCan = rqptr->AuthGroupCan;
      return (SS$_NORMAL);
   }

   if (rqptr->AuthSourceRealm == AUTH_SOURCE_VMS ||
       rqptr->AuthSourceRealm == AUTH_SOURCE_ID ||
       rqptr->AuthSourceRealm == AUTH_SOURCE_WASD_ID)
   {
      if (!AuthSysUafEnabled ||
          (AuthPolicySysUafSslOnly &&
           rqptr->RequestScheme != SCHEME_HTTPS))
      {
         /**********************/
         /* SYSUAF not allowed */
         /**********************/

         /* SYSUAF-authentication is disabled, or disabled for non-"https:" */ 

         if (WatchThis)
         {
            if (AuthPolicySysUafSslOnly)
               AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                               "SYSUAF SSL only");
            else
               AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                               "SYSUAF disabled");
         }

         rqptr->ResponseStatusCode = 403;
         ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_ACCESS_DENIED), FI_NOLI);

         AUTH_RESET_REQUEST

         return (STS$K_ERROR);
      }
   }

   if (rqptr->AuthGroupRestrictListPtr[0])
   {
      /**********************************************/
      /* check HOST access against restriction list */
      /**********************************************/

      if (RestrictedBy =
          AuthRestrictList (rqptr, rqptr->AuthGroupRestrictListPtr))
      {
         if (RestrictedBy == AUTH_RESTRICTED_BY_FAIL ||
             RestrictedBy == AUTH_RESTRICTED_BY_HOSTNAME ||
             RestrictedBy == AUTH_RESTRICTED_BY_PROTOCOL)
         {
            rqptr->ResponseStatusCode = 403;
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_ACCESS_DENIED), FI_NOLI);

            AUTH_RESET_REQUEST

            return (STS$K_ERROR);
         }
      }
   }

   if (rqptr->AuthSourceRealm == AUTH_SOURCE_WORLD)
   {
      /********************************************/
      /* the WORLD realm (i.e. no authentication) */
      /********************************************/

      rqptr->AuthRequestCan = rqptr->AuthUserCan = rqptr->AuthGroupCan;
      strcpy (rqptr->RemoteUser, AUTH_REALM_WORLD);
   }
   else
   {
      /*************************/
      /* authenticate the user */
      /*************************/

      status = AuthParseAuthorization (rqptr);

      if (VMSok (status))
      {
         if (WatchThis)
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "AUTHENTICATE user:!AZ realm:!AZ!AZ",
                            rqptr->RemoteUser, rqptr->AuthRealmPtr,
                            AuthSourceString (rqptr->AuthRealmPtr,
                                              rqptr->AuthSourceRealm));

         status = AuthDoAuthorization (rqptr);
      }

      if (VMSnok (status))
      {
         if (status == SS$_INVLOGIN)
         {
            /**************************/
            /* authentication failure */
            /**************************/

            Accounting.AuthNotAuthenticatedCount++;
            rqptr->ResponseStatusCode = 401;
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_FAILED), FI_NOLI);
         }

         if (status == AUTH_RESTRICTED_BY_FAIL ||
             status == AUTH_RESTRICTED_BY_HOUR ||
             status == AUTH_RESTRICTED_BY_OTHER ||
             status == AUTH_RESTRICTED_BY_PROTOCOL)
         {
            /********************/
            /* access forbidden */
            /********************/

            rqptr->ResponseStatusCode = 403;
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_ACCESS_DENIED), FI_NOLI);
            return (STS$K_ERROR);
         }

         if (status == AUTH_LOGOUT)
         {
            /******************************/
            /* authentication user cancel */
            /******************************/

            rqptr->ResponseStatusCode = 401;
/***
            ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_NO_LOGOUT), FI_NOLI);
***/
            ErrorGeneral (rqptr,
"<B>Logout has probably not occured!</B>\n\
<P>Reload/refresh this page, clear the fields, resubmit, then cancel the \
new username/password dialog.\n\
A failure report indicates successful logout!\n\
Then navigate backwards.",
FI_NOLI);
            return (SS$_INVLOGIN);
         }

         return (status);
      }
   }

   if (rqptr->AuthGroupRestrictListPtr[0])
   {
      /*****************************************/
      /* check access against restriction list */
      /*****************************************/

      if (RestrictedBy =
          AuthRestrictList (rqptr, rqptr->AuthGroupRestrictListPtr))
      {
         if (rqptr->AuthSourceRealm == AUTH_SOURCE_WORLD ||
             RestrictedBy == AUTH_RESTRICTED_BY_FAIL ||
             RestrictedBy == AUTH_RESTRICTED_BY_HOSTNAME ||
             RestrictedBy == AUTH_RESTRICTED_BY_PROTOCOL)
            rqptr->ResponseStatusCode = 403;
         else
            rqptr->ResponseStatusCode = 401;
         ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_ACCESS_DENIED), FI_NOLI);
         return (STS$K_ERROR);
      }
   }

   /**************************************/
   /* set and check request capabilities */
   /**************************************/

   /* bit-wise AND of capabilities ensures minimum apply */
   rqptr->AuthRequestCan = rqptr->AuthGroupCan & rqptr->AuthUserCan;

   if (Debug)
   {
      fprintf (stdout,
      "|%s|%s|%s|\nGroupCan: %08.08X UserCan: %08.08X RequestCan: %08.08X\n",
      rqptr->AuthRealmPtr, rqptr->RemoteUser, rqptr->RemoteUserPassword,
      rqptr->AuthGroupCan, rqptr->AuthUserCan, rqptr->AuthRequestCan);
   }

   if (WatchThis)
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
         "CAN group:!AZ user:!AZ request:!AZ",
         AuthCanString (rqptr->AuthGroupCan, AUTH_CAN_FORMAT_SHORT),
         AuthCanString (rqptr->AuthUserCan, AUTH_CAN_FORMAT_SHORT),
         AuthCanString (rqptr->AuthRequestCan, AUTH_CAN_FORMAT_SHORT));

   if (rqptr->HttpMethod & rqptr->AuthRequestCan)
   {
      /*************************/
      /* authorized (finally!) */
      /*************************/

      Accounting.AuthAuthorizedCount++;
      return (SS$_NORMAL);
   }

   /*****************/
   /* 403 forbidden */
   /*****************/

   Accounting.AuthNotAuthorizedCount++;

   StringDsc.dsc$a_pointer = String;
   StringDsc.dsc$w_length = sizeof(String)-1;

   sys$fao (&StringFaoDsc, &Length, &StringDsc,
            rqptr->HttpMethodName, rqptr->PathInfoPtr);
   String[Length] = '\0';

   rqptr->ResponseStatusCode = 403;
   ErrorGeneral (rqptr, String, FI_NOLI);
   return (STS$K_ERROR);
}

/*****************************************************************************/
/*
Parse the "Authorization:" HTTP request header line.
*/ 

int AuthParseAuthorization (struct RequestStruct *rqptr)

{
   register char  *cptr, *sptr, *zptr;

   int  status;
   char  EncodedString [256],
         UserNamePassword [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthParseAuthorization()\n");

   if ((cptr = rqptr->HttpAuthorizationPtr) == NULL)
   {
      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "FAIL no \"Authorization:\"");
      return (SS$_INVLOGIN);
   }

   /***********************************/
   /* get the HTTP authorization line */
   /***********************************/

   if (Debug) fprintf (stdout, "HttpAuthorizationPtr |%s|\n", cptr);

   zptr = (sptr = rqptr->AuthType) + sizeof(rqptr->AuthType);
   while (*cptr && !ISLWS(*cptr) && sptr < zptr) *sptr++ = toupper(*cptr++);
   if (sptr >= zptr)
   {
      rqptr->ResponseStatusCode = 401;
      ErrorGeneralOverflow (rqptr, FI_LI);
      return (STS$K_ERROR);
   }
   *sptr = '\0';

   if (strsame (rqptr->AuthType, "BASIC"))
      return (BasicAuthorization (rqptr));
   else
   if (strsame (rqptr->AuthType, "DIGEST"))
      return (DigestAuthorization (rqptr));
   else
   {
      rqptr->ResponseStatusCode = 401;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_SCHEME), FI_LI);
      return (STS$K_ERROR);
   }
}

/*****************************************************************************/
/*
The list is a plain-text string of comma-separated elements, the presence of
which restricts the path to that element.  The element can be an IP host
address (numeric or alphabetic, with or without asterisk wildcard) or an
authenticated username.  IP addresses are recognised by either a leading digit
(for numeric host addresses), a leading asterisk (where hosts are wildcarded
across a domain), a leading "#" which forces an element to be compared to a
host name or address, "http:", "https:", or a leading "~" which indicates a
username.
*/ 

int AuthRestrictList
(
struct RequestStruct *rqptr,
char *RestrictList
)
{
   register char  *lptr, *cptr;

   boolean  CheckedHost,
            CheckedProtocol,
            CheckedUser,
            FoundHost,
            FoundProtocol,
            FoundUser,
            RestrictedOnFail,
            RestrictedOnHost,
            RestrictedOnProtocol,
            RestrictedOnUser;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthRestrictList() |%s|\n", RestrictList);

   RestrictedOnFail = false;
   RestrictedOnHost = RestrictedOnProtocol = RestrictedOnUser = true;
   FoundHost = FoundProtocol = FoundUser = false;
   lptr = RestrictList;
   while (*lptr)
   {
      CheckedUser = CheckedHost = false;
      while (*lptr && *lptr == ',') lptr++;
      if (*lptr == AUTH_RESTRICTED_FAIL[0] &&
          strsame (lptr, AUTH_RESTRICTED_FAIL, 6))
      {
         /* there was a problem with the restriction list, automatic failure */
         RestrictedOnFail = true;
         break;
      }
      if (isdigit(*lptr) || (lptr[0] == '#' && isdigit(lptr[1])))
      {
         /* numeric IP address */
         if (*lptr == '#') lptr++;
         cptr = rqptr->ClientIpAddressString;
         CheckedHost = FoundHost = true;
         /* drop through to the comparison routine */
      }
      else
      if (*lptr == '~')
      {
         /* authenticated user name */
         lptr++;
         cptr = rqptr->RemoteUser;
         CheckedUser = FoundUser = true;
         /* drop through to the comparison routine */
      }
      else
      if (tolower(*lptr) == 'h' && strsame (lptr, "https:", 6))
      {
         CheckedProtocol = FoundProtocol = true;
         if (rqptr->RequestScheme == SCHEME_HTTPS)
            RestrictedOnProtocol = false;
         while (*lptr && *lptr != ',') lptr++;
         continue;
      }
      else
      if (tolower(*lptr) == 'h' && strsame (lptr, "http:", 5))
      {
         CheckedProtocol = FoundProtocol = true;
         if (rqptr->RequestScheme == SCHEME_HTTP)
            RestrictedOnProtocol = false;
         while (*lptr && *lptr != ',') lptr++;
         continue;
      }
      else
      if (tolower(*lptr) == 'l' &&
          (strsame (lptr, "localhost", -1) ||
           strsame (lptr, "localhost,", 10)))
      {
         /* check reserved-word, server against client host's IP address */
         CheckedHost = FoundHost = true;
         if (strsame (rqptr->ServicePtr->ServerIpAddressString,
                      rqptr->ClientIpAddressString, -1))
            RestrictedOnHost = false;
         while (*lptr && *lptr != ',') lptr++;
         continue;
      }
      else
      {
         /* by default, alpha-numeric IP host name */
         if (*lptr == '#') lptr++;
         cptr = rqptr->ClientHostName;
         CheckedHost = FoundHost = true;
         /* drop through to the comparison routine */
      }
      if (Debug) fprintf (stdout, "|%s|%s|\n", lptr, cptr);

      while (*cptr && *lptr && *lptr != ',')
      {
         if (*lptr == '*')
         {
            while (*lptr == '*') lptr++;
            while (*cptr && tolower(*cptr) != tolower(*lptr)) cptr++;
         }
         while (tolower(*cptr) == tolower(*lptr) && *cptr && *lptr)
         {
            lptr++;
            cptr++;
         }
         if (*lptr != '*') break;
      }
      if (!*cptr && (!*lptr || *lptr == ','))
      {
         if (CheckedUser) RestrictedOnUser = false;
         if (CheckedHost) RestrictedOnHost = false;
      }
      while (*lptr && *lptr != ',') lptr++;
   }

   if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
   {
      char  Scratch [32];

      *(cptr = Scratch) = '\0';
      if (FoundProtocol && RestrictedOnProtocol)
      {
         strcpy (cptr, "PROTOCOL");
         cptr += 8;
      }
      if (FoundHost && RestrictedOnHost)
      {
         if (Scratch[0]) *cptr++ = '+';
         strcpy (cptr, "HOST");
         cptr += 4;
      }
      if (FoundUser && RestrictedOnUser)
      {
         if (Scratch[0]) *cptr++ = '+';
         strcpy (cptr, "USER");
      }
      if (RestrictedOnFail) strcpy (Scratch, AUTH_RESTRICTED_FAIL);
      if (Scratch[0])
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "RESTRICT \"!AZ\" on:!AZ", RestrictList, Scratch);
   }

   if (RestrictedOnFail) return (AUTH_RESTRICTED_BY_FAIL);
   if (FoundProtocol && RestrictedOnProtocol)
      return (AUTH_RESTRICTED_BY_PROTOCOL);
   if (FoundHost && RestrictedOnHost) return (AUTH_RESTRICTED_BY_HOSTNAME);
   if (FoundUser && RestrictedOnUser) return (AUTH_RESTRICTED_BY_USERNAME);

   if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                      "RESTRICT \"!AZ\" none applies", RestrictList);
   return (0);
}

/*****************************************************************************/
/*
Look for an authentication record with the realm and username in the linked-
list, binary tree.  If one doesn't exists create a new one.  Check the supplied
password against the one in the record.  If it matches then authentication has
succeeded.  If not, the check the supplied password against the database.  If
it matches then copy the supplied password into the authentication record and
the authentication succeeds.  If it doesn't match then the authentication
fails.  Return SS$_INVLOGIN to indicate authentication failure, SS$_NORMAL
indicating authentication success, or any other error status indicating any
other error.
*/ 

int AuthDoAuthorization (struct RequestStruct *rqptr)

{
   static int  NoDuplicatesFlag = 0;

   struct AuthCacheRecordStruct  *acrptr;
   register char  *cptr, *sptr, *zptr;

   boolean  AccessDetermined,
            Authenticated,
            HttpdLogout;
   int  status,
        Minutes;
   struct AuthCacheRecordStruct  AuthCacheRecord;
   
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthDoAuthorization()\n");

   if (strsame (rqptr->QueryStringPtr, "httpd=logout", 12))
      HttpdLogout = true;
   else
      HttpdLogout = false;

   rqptr->AuthUserCan = 0;

   if (rqptr->AuthScheme == AUTH_SCHEME_DIGEST)
   {
      Accounting.AuthDigestCount++;
      if (!rqptr->RemoteUser[0] || !rqptr->AuthDigestResponsePtr[0])
      {
         if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "FAIL authentication incomplete");
         return (SS$_INVLOGIN);
      }
   }
   else
   if (rqptr->AuthScheme == AUTH_SCHEME_BASIC)
   {
      Accounting.AuthBasicCount++;
      if (!rqptr->RemoteUser[0] || !rqptr->RemoteUserPassword[0])
      {
         if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "FAIL authentication incomplete");
         return (SS$_INVLOGIN);
      }
   }
   else     
      ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);

   /*********************/
   /* lookup the record */
   /*********************/

   cptr = rqptr->AuthRealmPtr;
   zptr = (sptr = AuthCacheRecord.Realm) + sizeof(AuthCacheRecord.Realm);
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      ErrorGeneralOverflow (rqptr, FI_LI);
      return (STS$K_ERROR); 
   }
   *sptr = '\0';
   AuthCacheRecord.RealmLength = sptr - AuthCacheRecord.Realm;

   cptr = rqptr->AuthGroupWritePtr;
   zptr = (sptr = AuthCacheRecord.GroupWrite) +
          sizeof(AuthCacheRecord.GroupWrite);
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      ErrorGeneralOverflow (rqptr, FI_LI);
      return (STS$K_ERROR); 
   }
   *sptr = '\0';
   AuthCacheRecord.GroupWriteLength = sptr - AuthCacheRecord.GroupWrite;

   cptr = rqptr->AuthGroupReadPtr;
   zptr = (sptr = AuthCacheRecord.GroupRead) +
          sizeof(AuthCacheRecord.GroupRead);
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      ErrorGeneralOverflow (rqptr, FI_LI);
      return (STS$K_ERROR); 
   }
   *sptr = '\0';
   AuthCacheRecord.GroupReadLength = sptr - AuthCacheRecord.GroupRead;

   cptr = rqptr->RemoteUser;
   zptr = (sptr = AuthCacheRecord.UserName) + sizeof(AuthCacheRecord.UserName);
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      rqptr->ResponseStatusCode = 403;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_USERNAME_LENGTH), FI_LI);
      return (STS$K_ERROR); 
   }
   *sptr = '\0';
   AuthCacheRecord.UserNameLength = sptr - AuthCacheRecord.UserName;

   cptr = rqptr->RemoteUserPassword;
   zptr = (sptr = AuthCacheRecord.Password) + sizeof(AuthCacheRecord.Password);
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      rqptr->ResponseStatusCode = 403;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_PASSWORD_LENGTH), FI_LI);
      return (STS$K_ERROR); 
   }
   *sptr = '\0';

   AuthCacheRecord.SourceRealm = rqptr->AuthSourceRealm;
   AuthCacheRecord.SourceGroupRead = rqptr->AuthSourceGroupRead;
   AuthCacheRecord.SourceGroupWrite = rqptr->AuthSourceGroupWrite;

   status = lib$lookup_tree (&AuthCacheTreeHead,
                             &AuthCacheRecord,
                             &AuthCacheTreeCompareRecord,
                             &acrptr);
   if (Debug) fprintf (stdout, "lib$lookup_tree() %%X%08.08X\n", status);

   if (VMSok (status))
   {
      /*******************/
      /* existing record */
      /*******************/

      if (Debug)
         fprintf (stdout,
"AuthCacheRecord\n\
|%s|%d|\n\
|%s|%d|\n\
|%s|%d|\n\
|%s|%08.08X|%d|%d|%d|%d|%d\n",
            acrptr->Realm, acrptr->SourceRealm,
            acrptr->GroupWrite, acrptr->SourceGroupWrite,
            acrptr->GroupRead, acrptr->SourceGroupRead, acrptr->UserName,
            acrptr->AuthUserCan, acrptr->CacheMinutes,
            acrptr->AccessCount, acrptr->FailureCount,
            acrptr->VmsUserProfileLength, acrptr->VmsUserProfilePtr);

      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
      {
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
"CACHE [!AZ!AZ;!AZ!AZ;!AZ!AZ] !AZ failures:!UL accesses:!UL revalidate:!UL",
            acrptr->Realm[0] ? (char*)acrptr->Realm : "-",
            AuthSourceString (acrptr->Realm, acrptr->SourceRealm),
            acrptr->GroupWrite[0] ? (char*)acrptr->GroupWrite : "-",
            AuthSourceString (acrptr->GroupWrite, acrptr->SourceGroupWrite),
            acrptr->GroupRead[0] ? (char*)acrptr->GroupRead : "-",
            AuthSourceString (acrptr->GroupRead, acrptr->SourceGroupRead),
            acrptr->UserName, acrptr->FailureCount, acrptr->AccessCount,
            acrptr->CacheMinutes);
      }

      if (VMSnok (status = AuthRestrictAny (rqptr, acrptr)))
      {
         /****************/
         /* restrictions */
         /****************/

         return (status);
      }

      acrptr->LastAccessBinTime[0] = rqptr->BinaryTime[0];
      acrptr->LastAccessBinTime[1] = rqptr->BinaryTime[1];

      if (acrptr->FailureCount >= Config.AuthFailureLimit)
      {
         /*********************/
         /* TOO MANY FAILURES */
         /*********************/

         acrptr->FailureCount++;

         memset (acrptr->Password, 0, sizeof(acrptr->Password));
         memset (acrptr->DigestResponse, 0, sizeof(acrptr->DigestResponse));

         fprintf (stdout, "%%%s-W-AUTHFAILIM, %s, %d %s %s %s\n",
                  Utility, DateTime(&rqptr->BinaryTime, 20),
                  acrptr->FailureCount, acrptr->Realm,
                  acrptr->UserName, rqptr->ClientHostName);

         if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "FAIL failures exceed limit");

         return (SS$_INVLOGIN); 
      }

      if (Config.AuthRevalidateUserMinutes)
      {
         /********************/
         /* revalidate user? */
         /********************/

         /*
            User validation only lasts for a limited period (< 24 hours).
            Check how long ago was the last authorization attempt, if that is
            exceeded force the user to reenter via browser dialog.
         */

         Minutes = rqptr->NumericTime[3] * 60 + rqptr->NumericTime[4];
         if (Minutes < acrptr->RevalidateUserMinutes)
             acrptr->RevalidateUserMinutes =
                1440 - acrptr->RevalidateUserMinutes;
         if (Debug)
            fprintf (stdout, "Minutes %d + %d > %d\n",
               acrptr->RevalidateUserMinutes,
               Config.AuthRevalidateUserMinutes, Minutes);
         if (Minutes >= acrptr->RevalidateUserMinutes +
                        Config.AuthRevalidateUserMinutes)
         {
            acrptr->Password[0]  = '\0';
            acrptr->RevalidateUserMinutes = Minutes;

            if (rqptr->AdminWatchItem &&
                (AdminWatchEnabled & ADMIN_WATCH_AUTH))
               AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                               "FAIL revalidate timeout");

            return (SS$_INVLOGIN); 
         }
         /* update the time stamp with this last access */
         acrptr->RevalidateUserMinutes = Minutes;
      }

      /*********************/
      /* revalidate cache? */
      /*********************/

      /*
         Cache authentication only lasts for a limited period (< 24 hours).
         Check how long ago was the last authorization attempt.
      */
      Minutes = rqptr->NumericTime[3] * 60 + rqptr->NumericTime[4];
      if (Minutes < acrptr->CacheMinutes)
         acrptr->CacheMinutes = 1440 - acrptr->CacheMinutes;
      if (Debug)
         fprintf (stdout, "Minutes %d + %d > %d\n",
            acrptr->CacheMinutes, Config.AuthCacheMinutes, Minutes);
      if (Minutes >= acrptr->CacheMinutes + Config.AuthCacheMinutes)
      {
         /********/
         /* yes! */
         /********/

         acrptr->Password[0]  = '\0';
         acrptr->CacheMinutes = Minutes;
         /* drop through to recheck authentication */
      }

      /************************/
      /* check authentication */
      /************************/

      if (rqptr->AuthScheme == AUTH_SCHEME_DIGEST)
      {
         if (acrptr->DigestResponse[0] &&
             !strcmp (rqptr->AuthDigestResponsePtr, acrptr->DigestResponse))
         {
            Authenticated = true;
            acrptr->DigestCount++;
         }
         else
            Authenticated = false;
      }
      else
      if (rqptr->AuthScheme == AUTH_SCHEME_BASIC)
      {
         if (acrptr->Password[0] &&
             !strcmp (rqptr->RemoteUserPassword, acrptr->Password))
         {
            Authenticated = true;
            acrptr->BasicCount++;
         }
         else
            Authenticated = false;
      }

      if (HttpdLogout) Authenticated = false;

      if (Authenticated)
      {
         /*****************/
         /* authenticated */
         /*****************/

         acrptr->FailureCount = 0;
         acrptr->AccessCount++;
         rqptr->AuthCanChangeSysUafPwd = acrptr->CanChangeSysUafPwd;
         rqptr->AuthSysUafAuthenticated = acrptr->SysUafAuthenticated;
         rqptr->AuthSourceRealm = acrptr->SourceRealm;
         rqptr->AuthSourceGroupRead = acrptr->SourceGroupRead;
         rqptr->AuthSourceGroupWrite = acrptr->SourceGroupWrite;
         rqptr->AuthUserCan = acrptr->AuthUserCan;
         rqptr->AuthVmsUserProfilePtr = acrptr->VmsUserProfilePtr;
         rqptr->AuthVmsUserProfileLength = acrptr->VmsUserProfileLength;

         if (Debug) fprintf (stdout, "authenticated (in tree)!\n");
         return (SS$_NORMAL); 
      }

      /*********************/
      /* not authenticated */
      /*********************/

      /* any VMS profile is no longer valid */
      if (acrptr->VmsUserProfilePtr != NULL) VmFree (acrptr->VmsUserProfilePtr);
      acrptr->VmsUserProfileLength = 0;
      acrptr->VmsUserProfilePtr = NULL;
   }
   else
   if (status == LIB$_KEYNOTFOU)
   {
      /****************************************/
      /* record not found in tree, create one */
      /****************************************/

      acrptr = VmGet (sizeof(struct AuthCacheRecordStruct));
      
      /* we know these'll fit, sizes have been checked during authentication */
      strcpy (acrptr->Realm, AuthCacheRecord.Realm);
      acrptr->RealmLength = AuthCacheRecord.RealmLength;
      strcpy (acrptr->GroupWrite, AuthCacheRecord.GroupWrite);
      acrptr->GroupWriteLength = AuthCacheRecord.GroupWriteLength;
      strcpy (acrptr->GroupRead, AuthCacheRecord.GroupRead);
      acrptr->GroupReadLength = AuthCacheRecord.GroupReadLength;
      strcpy (acrptr->UserName, AuthCacheRecord.UserName);
      acrptr->UserNameLength = AuthCacheRecord.UserNameLength;
      acrptr->SourceRealm = AuthCacheRecord.SourceRealm;
      acrptr->SourceGroupRead = AuthCacheRecord.SourceGroupRead;
      acrptr->SourceGroupWrite = AuthCacheRecord.SourceGroupWrite;
      Minutes = rqptr->NumericTime[3] * 60 + rqptr->NumericTime[4];
      acrptr->CacheMinutes = acrptr->RevalidateUserMinutes = Minutes;
      acrptr->LastAccessBinTime[0] = rqptr->BinaryTime[0];
      acrptr->LastAccessBinTime[1] = rqptr->BinaryTime[1];
      /* all other fields have zeroes from the memory allocation! */

      status = lib$insert_tree (&AuthCacheTreeHead, acrptr,
                  &NoDuplicatesFlag, &AuthCacheTreeCompareRecord,
                  &AuthCacheTreeAllocateRecord, &acrptr, 0);
      if (Debug) fprintf (stdout, "lib$insert_tree() %%X%08.08X\n", status);
      if (VMSnok (status)) VmFree (acrptr);
   }

   if (VMSnok (status))
   {
      rqptr->ErrorTextPtr = MsgFor(rqptr,MSG_AUTH_USER);
      ErrorVmsStatus (rqptr, status, FI_LI);
      return (status);
   }

   /**********************************************/
   /* get authentication from appropriate source */
   /**********************************************/

   AccessDetermined = Authenticated = false;
   acrptr->DataBaseCount++;

   if (HttpdLogout)
   {
      AccessDetermined = true;
      status = AUTH_LOGOUT;
   }
   else
   if (AuthPromiscuous)
   {
      AccessDetermined = true;
      status = SS$_NORMAL;
      if (AuthPromiscuousPwdPtr != NULL)
         if (!strsame (rqptr->RemoteUserPassword, AuthPromiscuousPwdPtr, -1))
            status = SS$_INVLOGIN;

      /* allow the user to do any/everything */
      if (VMSok (status)) rqptr->AuthUserCan = AUTH_READWRITE_ACCESS;
   }
   else
   {
      switch (rqptr->AuthSourceRealm)
      {
         case AUTH_SOURCE_VMS :
         case AUTH_SOURCE_ID :
         case AUTH_SOURCE_WASD_ID :

            /* check against SYSUAF */
            Accounting.AuthVmsCount++;
            acrptr->SysUafAuthenticated = rqptr->AuthSysUafAuthenticated = true;

            status = AuthVerifyVmsPassword (rqptr, acrptr);

            /*
               SYSUAF authentication can only be performed using the basic
               scheme (it needs a clear-text password to hash and then compare)
            */
            rqptr->AuthChallengeScheme = AUTH_SCHEME_BASIC;

            break;

         case AUTH_SOURCE_LIST :

            /* check against simple list */
            /*** Accounting.AuthSimpleListCount++; ***/
            status = AuthReadSimpleList (rqptr, rqptr->AuthRealmPtr, true);

            /* if the list has the user name and corrrect password */
            if (VMSok (status)) rqptr->AuthUserCan = AUTH_READWRITE_ACCESS;

            break;

         case AUTH_SOURCE_HTA :

            /* check against HTA database */
            Accounting.AuthHtDatabaseCount++;
            status = AuthReadHtDatabase (rqptr, rqptr->AuthRealmPtr, true);

            break;

         default :

            ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
      }
   }

   if (VMSok (status)) Authenticated = true;

   acrptr->HttpsOnly = rqptr->AuthHttpsOnly;

   /*********************/
   /* any restrictions? */
   /*********************/

   if (VMSok (status))
      status = AuthRestrictAny (rqptr, acrptr);

   /*******************************************/
   /* check for WASD "hard-wired" identifiers */
   /*******************************************/

   if (VMSok (status) &&
       !AccessDetermined &&
       rqptr->AuthSourceRealm == AUTH_SOURCE_WASD_ID)
   {
      AccessDetermined = true;
      rqptr->AuthUserCan = 0;

      status = AuthHoldsVmsIdentifier (rqptr, AUTH_WASD_WRITE_VMS_ID,
                                       AuthWasdWriteVmsIdentifier);
      /* if it has the full-access identifier */
      if (VMSok (status))
         rqptr->AuthUserCan = AUTH_READWRITE_ACCESS;
      else
      if (status == 0)
      {
         /* did not hold the full-access identifier */
         status = AuthHoldsVmsIdentifier (rqptr, AUTH_WASD_READ_VMS_ID,
                                          AuthWasdReadVmsIdentifier);
         /* if it has the read-only identifier */
         if (VMSok (status))
            rqptr->AuthUserCan = AUTH_READONLY_ACCESS;
         else
         if (status == 0)
         {
            /* one or other of these MUST have allowed authentication! */
            ErrorGeneral (rqptr, ErrorSanityCheck, FI_NOLI);
            return (status);
         }
      }

      if (rqptr->AuthSourceGroupWrite)
      {
         /*
            Group membership specified.  Check for possession of
            a "WASD_VMS__groupname" identifier.
         */

         if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "CHECK group !AZ!AZ!AZ",
                            AUTH_WASD__GROUP_VMS_ID,
                            rqptr->AuthGroupWritePtr,
                            AuthSourceString (rqptr->AuthGroupWritePtr,
                                              rqptr->AuthSourceGroupWrite));
   
         status = AuthHoldsWasdGroupIdentifier (rqptr,
                                                rqptr->AuthGroupWritePtr);
 
         if (VMSnok (status))
         {
            /* doesn't hold the group identifier, no access */
            rqptr->AuthUserCan = 0;
            status = AUTH_RESTRICTED_BY_OTHER;
         }
      }
   }

   /********************************/
   /* check any group databases(s) */
   /********************************/

   if (VMSok (status) &&
       !AccessDetermined &&
       rqptr->AuthSourceGroupWrite)
   {
      /* check for full-access (read and write) group membership */
      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "CHECK r+w group !AZ!AZ",
                         rqptr->AuthGroupWritePtr,
                         AuthSourceString (rqptr->AuthGroupWritePtr,
                                           rqptr->AuthSourceGroupWrite));

      AccessDetermined = true;
      rqptr->AuthUserCan = 0;

      switch (rqptr->AuthSourceGroupWrite)
      {
         case AUTH_SOURCE_ID :

            status = AuthHoldsVmsIdentifier (rqptr,
                        rqptr->AuthGroupWritePtr,
                        rqptr->AuthGroupWriteVmsIdentifier);
            break;

         case AUTH_SOURCE_LIST :

            status = AuthReadSimpleList (rqptr,
                        rqptr->AuthGroupWritePtr, false);
            break;

         case AUTH_SOURCE_HTA :

            status = AuthReadHtDatabase (rqptr,
                        rqptr->AuthGroupWritePtr, false);
            break;

         default :

            ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
      }

      /* if it was found in the full-access group */
      if (VMSok (status))
         rqptr->AuthUserCan = AUTH_READWRITE_ACCESS;
      else
         status = AUTH_RESTRICTED_BY_OTHER;

      if (rqptr->AuthSourceGroupRead &&
          !rqptr->AuthUserCan)
      {
         /* check for read-only access group membership */
         if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "CHECK r-only group !AZ!AZ",
                            rqptr->AuthGroupReadPtr,
                            AuthSourceString (rqptr->AuthGroupReadPtr,
                                              rqptr->AuthSourceGroupRead));

         rqptr->AuthUserCan = 0;

         switch (rqptr->AuthSourceGroupWrite)
         {
            case AUTH_SOURCE_ID :

               status = AuthHoldsVmsIdentifier (rqptr,
                           rqptr->AuthGroupReadPtr,
                           rqptr->AuthGroupReadVmsIdentifier);
               break;

            case AUTH_SOURCE_LIST :

               status = AuthReadSimpleList (rqptr,
                           rqptr->AuthGroupReadPtr, false);
               break;

            case AUTH_SOURCE_HTA :

               status = AuthReadHtDatabase (rqptr,
                           rqptr->AuthGroupReadPtr, false);
            break;

            default :

               ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
         }

         /* if found in the read-only access group */
         if (VMSok (status))
            rqptr->AuthUserCan = AUTH_READONLY_ACCESS;
         else
            status = AUTH_RESTRICTED_BY_OTHER;
      }
   }

   if (Debug)
      fprintf (stdout,
         "AuthCacheRecord |%s|%s|%s|%s| %08.08X %d %d\n",
         acrptr->Realm, acrptr->GroupWrite, acrptr->GroupRead,
         acrptr->UserName, acrptr->AuthUserCan,
         acrptr->AccessCount, acrptr->FailureCount);

   if (Authenticated)
   {
      /*****************/
      /* authenticated */
      /*****************/

      if (rqptr->AuthScheme == AUTH_SCHEME_BASIC)
      {
         strcpy (acrptr->Password, rqptr->RemoteUserPassword);
         acrptr->BasicCount++;
      }
      else
      if (rqptr->AuthScheme == AUTH_SCHEME_DIGEST)
      {
         strcpy (acrptr->DigestResponse, rqptr->AuthDigestResponsePtr);
         acrptr->DigestCount++;
      }

      if (acrptr->FailureCount)
      {
         /* just let the log know the access finally succeeded */
         fprintf (stdout, "%%%s-I-AUTHFAILOK, %s, %d %s %s %s\n",
                  Utility, DateTime(&rqptr->BinaryTime, 20),
                  acrptr->FailureCount, acrptr->Realm,
                  acrptr->UserName, rqptr->ClientHostName);
      }

      acrptr->FailureCount = 0;
      acrptr->AccessCount++;
      acrptr->AuthUserCan = rqptr->AuthUserCan;
      acrptr->CanChangeSysUafPwd = rqptr->AuthCanChangeSysUafPwd;
      acrptr->VmsUserProfileLength = 0;
      acrptr->VmsUserProfilePtr = NULL;

      if (rqptr->AuthSysUafAuthenticated && AuthVmsUserSecProfileEnabled)
      {
         status = AuthCreateVmsUserProfile (rqptr);
         acrptr->VmsUserProfilePtr = rqptr->AuthVmsUserProfilePtr;
         acrptr->VmsUserProfileLength = rqptr->AuthVmsUserProfileLength;
      }

      if (Debug) fprintf (stdout, "authenticated!\n");

      return (status); 
   }
   else
   {
      /*********************/
      /* not authenticated */
      /*********************/

      acrptr->FailureCount++;

      if (status == AUTH_LOGOUT)
         fprintf (stdout, "%%%s-W-AUTHLOGOUT, %s, %d %s %s %s\n",
                  Utility, DateTime(&rqptr->BinaryTime, 20),
                  acrptr->FailureCount, acrptr->Realm,
                  acrptr->UserName, rqptr->ClientHostName);
      else
         fprintf (stdout, "%%%s-W-AUTHFAIL, %s, %d %s %s %s\n",
                  Utility, DateTime(&rqptr->BinaryTime, 20),
                  acrptr->FailureCount, acrptr->Realm,
                  acrptr->UserName, rqptr->ClientHostName);

      return (status);
   }
}

/*****************************************************************************/
/*
Check for any access restrictions.  SSL only.  SYSUAF restrictions on hours.
*/ 

int AuthRestrictAny
(
struct RequestStruct *rqptr,
struct AuthCacheRecordStruct *acrptr
)
{
   static unsigned long  DayOfWeek,
                         DayOfWeekBit,
                         PreviousDateDay;

   int  status;
   unsigned long  HourOfDayBit;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthRestrictAny()\n");

   /* can only use this authentication with "https:" (SSL)? */
   if (acrptr->HttpsOnly &&
       rqptr->RequestScheme != SCHEME_HTTPS)
   {
      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "FAIL restricted to SSL");
      return (AUTH_RESTRICTED_BY_PROTOCOL);
   }

   /* limited access? */
   if (acrptr->SysUafAuthenticated)
   {
      if (rqptr->NumericTime[2] != PreviousDateDay)
      {
         /* day in date has changed, ergo ... so has day-of-week! */
         PreviousDateDay = rqptr->NumericTime[2];
         if (VMSnok (status = lib$day_of_week (0, &DayOfWeek)))
            ErrorExitVmsStatus (status, "lib$day_of_week()", FI_LI);
         DayOfWeekBit = 1 << (DayOfWeek - 1);
      }

      HourOfDayBit = 1 << rqptr->NumericTime[3];

      if (Debug)
         fprintf (stdout, "DayOfWeek: %d %08.08X %08.08X\n",
                  DayOfWeek, DayOfWeekBit, acrptr->UaiPrimeDays);

      if (acrptr->UaiPrimeDays & DayOfWeekBit)
      {
         /* secondary day */
         if (Debug)
            fprintf (stdout, "secondary: %08.08X %08.08X %08.08X\n",
                     HourOfDayBit, acrptr->UaiNetworkAccessS,
                     acrptr->UaiRemoteAccessS);

         if (HourOfDayBit & acrptr->UaiNetworkAccessS ||
             HourOfDayBit & acrptr->UaiRemoteAccessS)
         {
            if (rqptr->AdminWatchItem &&
                (AdminWatchEnabled & ADMIN_WATCH_AUTH))
               AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                              "FAIL restricted SYSUAF secondary hours");
            return (AUTH_RESTRICTED_BY_HOUR);
         }
      }
      else
      {
         /* prime day */
         if (Debug)
            fprintf (stdout, "prime: %08.08X %08.08X %08.08X\n",
                     HourOfDayBit, acrptr->UaiNetworkAccessP,
                     acrptr->UaiRemoteAccessP);

         if (HourOfDayBit & acrptr->UaiNetworkAccessP ||
             HourOfDayBit & acrptr->UaiRemoteAccessP)
         {
            if (rqptr->AdminWatchItem &&
                (AdminWatchEnabled & ADMIN_WATCH_AUTH))
               AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                              "FAIL restricted SYSUAF primary hours");
            return (AUTH_RESTRICTED_BY_HOUR);
         }
      }
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Called by lib$insert_tree() and lib$lookup_tree() in a variety of functions.
Returns negative number if user node is lexicographically less than the tree
node, positive if greater, or zero if exactly the same.
*/ 

int AuthCacheTreeCompareRecord
(
struct AuthCacheRecordStruct *acr1ptr,
struct AuthCacheRecordStruct *acr2ptr,
unsigned long Unused
)
{
   register int  cmp;
   
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCacheTreeCompareRecord()\n");

   /* lengths of all strings in the records */
   if (cmp = acr1ptr->RealmLength - acr2ptr->RealmLength)
      return (cmp);
   if (cmp = acr1ptr->GroupWriteLength - acr2ptr->GroupWriteLength)
      return (cmp);
   if (cmp = acr1ptr->GroupReadLength - acr2ptr->GroupReadLength)
      return (cmp);
   if (cmp = acr1ptr->UserNameLength - acr2ptr->UserNameLength)
      return (cmp);

   /* values of source indicators */
   if (cmp = acr1ptr->SourceRealm - acr2ptr->SourceRealm)
      return (cmp);
   if (cmp = acr1ptr->SourceGroupWrite - acr2ptr->SourceGroupWrite)
      return (cmp);
   if (cmp = acr1ptr->SourceGroupRead - acr2ptr->SourceGroupRead)
      return (cmp);

   /* strings */
   if (cmp = strcmp (acr1ptr->Realm, acr2ptr->Realm))
      return (cmp);
   if (cmp = strcmp (acr1ptr->GroupWrite, acr2ptr->GroupWrite))
      return (cmp);
   if (cmp = strcmp (acr1ptr->GroupRead, acr2ptr->GroupRead))
      return (cmp);
   return (strcmp (acr1ptr->UserName, acr2ptr->UserName));
}

/*****************************************************************************/
/*
Called by lib$insert_tree() in AuthDoAuthorization().  Insert the address
in 'acrptr' into the location pointed at by 'acrptrptr' (inside the binary tree
structure presumably!)
*/ 

AuthCacheTreeAllocateRecord
(
struct AuthCacheRecordStruct *acrptr,
struct AuthCacheRecordStruct **acrptrptr,
unsigned long Unused
)
{
   if (Debug) fprintf (stdout, "AuthCacheTreeAllocateRecord()\n");

   *acrptrptr = acrptr;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Clear all of the authentication records from the tree.  Will result in all
subsequent requests being re-authenticated from their respective on-disk
databases.  Called from the Admin.c and Control.c modules.
*/

int AuthCacheTreeFree ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCacheTreeFree()\n");

   sys$gettim (&AuthCacheTreeBinTime);
   if (!AuthCacheTreeHead) return (SS$_NORMAL);
   AuthFreeRecordCount = 0;

   status = lib$traverse_tree (&AuthCacheTreeHead,
                               &AuthCacheTreeFreeRecord, 0);
   if (Debug) fprintf (stdout, "lib$traverse_tree() %%X%08.08X\n", status);

   if (AuthCacheTreeTailPtr) VmFree (AuthCacheTreeTailPtr);
   AuthCacheTreeHead = AuthCacheTreeTailPtr = 0;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Free the dynamically allocated memory associated with the user tree record.
Called from AuthCacheTreeFree().
*/

AuthCacheTreeFreeRecord
(                      
struct AuthCacheRecordStruct *acrptr,
unsigned long  Unused
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCacheTreeFreeRecord()\n");

   if (Debug)
      fprintf (stdout, "AuthCacheRecord |%s|%s|\n",
               acrptr->Realm, acrptr->UserName);

   AuthFreeRecordCount++;

   if (acrptr->LeftLink)
   {
      /* any VMS profile is no longer valid either */
      if (acrptr->LeftLink->VmsUserProfilePtr != NULL)
         VmFree (acrptr->LeftLink->VmsUserProfilePtr);

      VmFree (acrptr->LeftLink);
   }
   else
      AuthCacheTreeTailPtr = acrptr;
}

/****************************************************************************/
/*
Verify the request username/password and/or read capabilities from the on-disk
HTA (HTTPd) authroization database.

Returns a success status if user password authenticated, SS$_INVLOGIN if not
authenticated, or other error status if a genuine VMS error occurs (which
should be reported where and when it occurs).

The digest is stored in both upper and lower case versions because clients
will have case-lock either on or off producing a digest of either upper or
lower case username and password.  Keep digests of both so that provided one
or other case is used exclusively the digest thereof can still be matched
with one or other stored here :^)
*/ 

int AuthReadHtDatabase
(
struct RequestStruct* rqptr,
char *DatabaseName,
boolean AuthenticatePassword
)
{
   static $DESCRIPTOR (PasswordDsc, "");
   static $DESCRIPTOR (UserNameDsc, "");

   boolean  PasswordAuthenticated,
            UserNameEnabled;
   int  status;
   unsigned long  HashedPwd [2];
   char  A1HexDigest [33],
         HexDigest [33];
   struct AuthHtRecordStruct AuthHtRecord;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "AuthReadHtDatabase() |%s|%s|\n",
               DatabaseName, rqptr->RemoteUser);

   /* set the capabilities to zero before we do anything else! */
   rqptr->AuthUserCan = 0;

   /* look for the record, leave the database file open if found */
   status = AuthAccessHtDatabase (true, DatabaseName, rqptr->RemoteUser,
                                  &AuthHtRecord, NULL, NULL);
   if (status == SS$_INVLOGIN)
   {
      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "FAIL !AZ!AZ username", DatabaseName,
                         AuthSourceString (DatabaseName, AUTH_SOURCE_HTA));

      AuthAccessHtDatabase (false, NULL, NULL, NULL, NULL, NULL);
      return (status);
   }

   if (VMSnok (status))
   {
      rqptr->ErrorTextPtr = MsgFor(rqptr,MSG_AUTH_DATABASE);
      rqptr->ErrorOtherTextPtr = DatabaseName;
      ErrorVmsStatus (rqptr, status, FI_LI);
      return (status);
   }

   /****************/
   /* record found */
   /****************/

   PasswordAuthenticated = false;

   if (AuthHtRecord.Flags & AUTH_FLAG_ENABLED)
   {
      UserNameEnabled = true;
      if (AuthenticatePassword &&
          AuthHtRecord.HashedPwd[0] && AuthHtRecord.HashedPwd[1])
      {
         if (rqptr->AuthScheme == AUTH_SCHEME_BASIC)
         {
            if (Debug) fprintf (stdout, "BASIC\n");

            UserNameDsc.dsc$w_length = strlen(rqptr->RemoteUser);
            UserNameDsc.dsc$a_pointer = rqptr->RemoteUser;
            PasswordDsc.dsc$w_length = strlen(rqptr->RemoteUserPassword);
            PasswordDsc.dsc$a_pointer = rqptr->RemoteUserPassword;

            status = sys$hash_password (&PasswordDsc, UAI$C_PURDY_S, 0,
                                        &UserNameDsc, &HashedPwd);
            if (Debug)
               fprintf (stdout, "sys$hash_password() %%X%08.08X\n", status);
            if (VMSnok (status))
            {
               /* ensure the currently open database is closed */
               AuthAccessHtDatabase (false, NULL, NULL, NULL, NULL, NULL);
               rqptr->ErrorTextPtr = MsgFor(rqptr,MSG_AUTH_USER);
               ErrorVmsStatus (rqptr, status, FI_LI);
               return (status);
            }

            if (Debug)
               fprintf (stdout, "|%08.08X%08.08X|%08.08X%08.08X|\n",
                        HashedPwd[1], HashedPwd[0],
                        AuthHtRecord.HashedPwd[1], AuthHtRecord.HashedPwd[0]);

            if (HashedPwd[0] == AuthHtRecord.HashedPwd[0] &&
                HashedPwd[1] == AuthHtRecord.HashedPwd[1])
               PasswordAuthenticated = true;
         }
         else
         if (rqptr->AuthScheme == AUTH_SCHEME_DIGEST)
         {
            if (Debug)
               fprintf (stdout, "DIGEST |%s|\n",
                        rqptr->AuthDigestResponsePtr);

            DigestHexString (AuthHtRecord.A1DigestLoCase, 16, A1HexDigest);
            DigestResponse (A1HexDigest,
                            rqptr->AuthDigestNoncePtr,
                            rqptr->HttpMethodName,
                            rqptr->AuthDigestUriPtr,
                            HexDigest);
            if (!strcmp (rqptr->AuthDigestResponsePtr, HexDigest))
               PasswordAuthenticated = true;
            else
            {
               DigestHexString (AuthHtRecord.A1DigestUpCase, 16, A1HexDigest);
               DigestResponse (A1HexDigest,
                               rqptr->AuthDigestNoncePtr,
                               rqptr->HttpMethodName,
                               rqptr->AuthDigestUriPtr,
                               HexDigest);
               if (!strcmp (rqptr->AuthDigestResponsePtr, HexDigest))
                  PasswordAuthenticated = true;
            }
         }

         if (!PasswordAuthenticated)
            if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
               AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                              "FAIL !AZ!AZ password", DatabaseName,
                              AuthSourceString (DatabaseName, AUTH_SOURCE_HTA));

         if (Debug)
            fprintf (stdout, "PasswordAuthenticated: %d\n",
                     PasswordAuthenticated);
      }

      if (!AuthenticatePassword || PasswordAuthenticated)
      {
         /* set the appropriate authorization flags */
         if (AuthHtRecord.Flags & AUTH_FLAG_DELETE)
            rqptr->AuthUserCan |= HTTP_METHOD_DELETE;
         if (AuthHtRecord.Flags & AUTH_FLAG_GET)
            rqptr->AuthUserCan |= HTTP_METHOD_GET | HTTP_METHOD_HEAD;
         if (AuthHtRecord.Flags & AUTH_FLAG_HEAD)
            rqptr->AuthUserCan |= HTTP_METHOD_HEAD;
         if (AuthHtRecord.Flags & AUTH_FLAG_POST)
            rqptr->AuthUserCan |= HTTP_METHOD_POST;
         if (AuthHtRecord.Flags & AUTH_FLAG_PUT)
            rqptr->AuthUserCan |= HTTP_METHOD_PUT;

         if (AuthHtRecord.Flags & AUTH_FLAG_HTTPS_ONLY)
            rqptr->AuthHttpsOnly = true;
         else
            rqptr->AuthHttpsOnly = false;
      }
   }

   if (UserNameEnabled && (!AuthenticatePassword || PasswordAuthenticated))
   {
      AuthHtRecord.AccessCount++;
      memcpy (&AuthHtRecord.LastAccessBinTime, &rqptr->BinaryTime, 8);
   }
   else
   {
      AuthHtRecord.FailureCount++;
      memcpy (&AuthHtRecord.LastFailureBinTime, &rqptr->BinaryTime, 8);
   }

   /* update the record, close the database file */
   status = AuthAccessHtDatabase (false, NULL, NULL, NULL, NULL, &AuthHtRecord);
   if (VMSnok (status))
   {
      rqptr->ErrorTextPtr = MsgFor(rqptr,MSG_AUTH_DATABASE);
      rqptr->ErrorOtherTextPtr = DatabaseName;
      ErrorVmsStatus (rqptr, status, FI_LI);
      return (status);
   }

   if (!AuthenticatePassword || PasswordAuthenticated)
      return (SS$_NORMAL);
   else
      return (SS$_INVLOGIN);
}

/****************************************************************************/
/*
Locate a specified user record in a specified on-disk database.  Either return
that record if found or if an update record supplied, update it! Can return any
RMS or VMS error status code.  Returns SS$_NORMAL is operation (read, add or
update) completed successfully.  Returns SS$_INVLOGIN if the record could not
be found.  Returns SS$_INCOMPAT if the database file version is incorrect.

THE DATABASE FILE IS ALWAYS CLOSED ON AN ERROR CONDITION.

'LeaveFileOpen' will leave the database file open with the current record
context ready for update, or for locating another user record.  This function
is not reentrant and therefore this context is valid ONLY within the one AST.

Call with 'DatabaseName' and 'UserName' non-NULL and all other pointers
set to NULL to check whether the user name has a record in the database.
*/ 

int AuthAccessHtDatabase
(
boolean LeaveFileOpen,
char *DatabaseName,
char *UserName,
struct AuthHtRecordStruct *AuthHtRecordReadPtr,
struct AuthHtRecordStruct *AuthHtRecordAddPtr,
struct AuthHtRecordStruct *AuthHtRecordUpdatePtr
)
{
   static char  AuthFileName [64],
                PrevDatabaseName [64];
   static unsigned short  AuthFileNameLength;
   static $DESCRIPTOR (AuthFileNameFaoDsc, "!AZ!AZ!AZ");
   static $DESCRIPTOR (AuthFileNameDsc, AuthFileName);
   static struct FAB  AuthFileFab;
   static struct NAM  AuthFileNam;
   static struct RAB  AuthFileRab;

   register char  *cptr, *sptr;

   int  status,
        UserNameLength;
   struct AuthHtRecordStruct *AuthHtRecordPtr;
   struct AuthHtRecordStruct AuthHtRecord;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "AuthAccessHtDatabase() %d |%s|%s| %d %d %d\n",
               LeaveFileOpen, DatabaseName, UserName,
               AuthHtRecordReadPtr, AuthHtRecordAddPtr, AuthHtRecordUpdatePtr);

   if (AuthHtRecordReadPtr == NULL &&
       AuthHtRecordAddPtr == NULL &&
       AuthHtRecordUpdatePtr == NULL)
   {
      /************************/
      /* force database close */
      /************************/

      sys$close (&AuthFileFab, 0, 0);
      return (SS$_NORMAL);
   }

   if (UserName == NULL && AuthHtRecordUpdatePtr != NULL)
   {
      /*********************************/
      /* update previously read record */
      /*********************************/

      AuthFileRab.rab$l_rbf = AuthHtRecordUpdatePtr;
      AuthFileRab.rab$w_rsz = sizeof(struct AuthHtRecordStruct);
      status = sys$update (&AuthFileRab, 0, 0);
      if (Debug) fprintf (stdout, "sys$update() %%X%08.08X\n", status);
      if (VMSok (status))
      {
         /* update completed successfully */
         if (!LeaveFileOpen) sys$close (&AuthFileFab, 0, 0);
         return (SS$_NORMAL);
      }
      /* update error */
      sys$close (&AuthFileFab, 0, 0);
      return (status);
   }

   if (DatabaseName == NULL)
   {
      /*****************************************/
      /* locate within currently open database */
      /*****************************************/

      if (VMSnok (status = sys$rewind (&AuthFileRab, 0, 0)))
      {
         sys$close (&AuthFileFab, 0, 0);
         return (status);
      }
      if (Debug) fprintf (stdout, "sys$rewind() %%X%08.08X\n", status);
   }
   else
   {
      /**********************************/
      /* open currently closed database */
      /**********************************/

      if (strcmp (DatabaseName, PrevDatabaseName))
      {
         sys$fao (&AuthFileNameFaoDsc, &AuthFileNameLength, &AuthFileNameDsc,
                  HTA_DIRECTORY, DatabaseName, HTA_FILE_TYPE);
         AuthFileName[AuthFileNameLength] = '\0';
      }
      if (Debug) fprintf (stdout, "AuthFileName |%s|\n", AuthFileName);

      AuthFileFab = cc$rms_fab;
      /* if a read-only operation without file left open then read-only! */
      if (!LeaveFileOpen &&
          AuthHtRecordAddPtr == NULL &&
          AuthHtRecordUpdatePtr == NULL)
         AuthFileFab.fab$b_fac = FAB$M_GET;
      else
         AuthFileFab.fab$b_fac = FAB$M_GET | FAB$M_PUT | FAB$M_UPD;
      AuthFileFab.fab$l_fna = AuthFileName;  
      AuthFileFab.fab$b_fns = AuthFileNameLength;
      AuthFileFab.fab$b_shr = FAB$M_SHRGET | FAB$M_SHRPUT | FAB$M_SHRUPD;

      /* turn on SYSPRV to allow access to authentication database file */
      EnableSysPrv ();
      status = sys$open (&AuthFileFab, 0, 0);
      DisableSysPrv ();

      /* status from sys$open() */
      if (VMSnok (status)) return (status);

      AuthFileRab = cc$rms_rab;
      AuthFileRab.rab$l_fab = &AuthFileFab;
      AuthFileRab.rab$b_mbf = 2;
      AuthFileRab.rab$l_rop = RAB$M_RAH;

      if (VMSnok (status = sys$connect (&AuthFileRab, 0, 0)))
      {
         sys$close (&AuthFileFab, 0, 0);
         return (status);
      }
   }

   /*****************/
   /* locate record */
   /*****************/

   /* for add and update read into the scratch record */
   if (AuthHtRecordReadPtr == NULL)
      AuthFileRab.rab$l_ubf = AuthHtRecordPtr = &AuthHtRecord;
   else
      AuthFileRab.rab$l_ubf = AuthHtRecordPtr = AuthHtRecordReadPtr;
   AuthFileRab.rab$w_usz = sizeof(struct AuthHtRecordStruct);

   if (UserName != NULL) UserNameLength = strlen(UserName);

   while (VMSok (status = sys$get (&AuthFileRab, 0, 0)))
   {
      /* check the version of the authorization database */
      if (AuthHtRecordPtr->DatabaseVersion &&
          AuthHtRecordPtr->DatabaseVersion != AuthCurrentDatabaseVersion)
      {
         status = SS$_INCOMPAT & 0xfffffffe;
         break;
      }

      /* if deleted record (all set to zeroes) continue */
      if (AuthHtRecordAddPtr == NULL && !AuthHtRecordPtr->UserNameLength)
         continue;

      /* if adding a record then use the first deleted one found */
      if (AuthHtRecordAddPtr != NULL && !AuthHtRecordPtr->UserNameLength)
         break;
/**
      if (Debug)
         fprintf (stdout, "AuthHtRecordPtr->UserName |%s|\n",
                  AuthHtRecordPtr->UserName);
**/
      if (UserNameLength != AuthHtRecordPtr->UserNameLength) continue;
      cptr = AuthHtRecordPtr->UserName;
      sptr = UserName;
      while (*cptr && *sptr && tolower(*cptr) == tolower(*sptr))
         { cptr++; sptr++; }
      if (*cptr || *sptr) continue;
      break;
   }
   if (Debug) fprintf (stdout, "sys$get() %%X%08.08X\n", status);

   if (AuthHtRecordUpdatePtr == NULL && AuthHtRecordAddPtr == NULL)
   {
      /***************/
      /* read record */
      /***************/

      if (VMSok (status))
      {
         /* record found */
         if (!LeaveFileOpen) sys$close (&AuthFileFab, 0, 0);
         return (SS$_NORMAL);
      }

      if (status == RMS$_EOF)
      {
         /* user record not found */
         if (!LeaveFileOpen) sys$close (&AuthFileFab, 0, 0);
         return (SS$_INVLOGIN);
      }
      /* RMS error */
      sys$close (&AuthFileFab);
      return (status);
   }

   if (AuthHtRecordUpdatePtr != NULL)
   {
      /*****************/
      /* update record */
      /*****************/

      if (VMSnok (status))
      {
         if (status == RMS$_EOF)
         {
            /* user record not found */
            if (!LeaveFileOpen) sys$close (&AuthFileFab, 0, 0);
            return (SS$_INVLOGIN);
         }
         /* RMS error */
         sys$close (&AuthFileFab, 0, 0);
         return (status);
      }

      AuthFileRab.rab$l_rbf = AuthHtRecordUpdatePtr;
      AuthFileRab.rab$w_rsz = sizeof(struct AuthHtRecordStruct);

      status = sys$update (&AuthFileRab, 0, 0);
      if (Debug) fprintf (stdout, "sys$update() %%X%08.08X\n", status);
      if (VMSok (status))
      {
         /* update completed successfully */
         if (!LeaveFileOpen) sys$close (&AuthFileFab, 0, 0);
         return (SS$_NORMAL);
      }
      /* RMS error */
      sys$close (&AuthFileFab, 0, 0);
      return (status);
   }

   if (AuthHtRecordAddPtr != NULL)
   {
      /****************/
      /* add a record */
      /****************/

      if (VMSnok (status) && status != RMS$_EOF)
      {
         /* RMS error */
         sys$close (&AuthFileFab, 0, 0);
         return (status);
      }

      AuthFileRab.rab$l_rbf = AuthHtRecordAddPtr;
      AuthFileRab.rab$w_rsz = sizeof(struct AuthHtRecordStruct);

      /* if reached end-of-file then add a record, else undate zeroed one */
      if (status == RMS$_EOF)
         status = sys$put (&AuthFileRab, 0, 0);
      else
         status = sys$update (&AuthFileRab, 0, 0);
      if (Debug) fprintf (stdout, "sys$put/update() %%X%08.08X\n", status);
      if (VMSok (status))
      {
         /* put or update completed successfully */
         if (!LeaveFileOpen) sys$close (&AuthFileFab, 0, 0);
         return (SS$_NORMAL);
      }
      /* RMS error */
      sys$close (&AuthFileFab, 0, 0);
      return (status);
   }

   /* sanity check error, should never get here */
   return (SS$_BUGCHECK);
}

/****************************************************************************/
/*
*/ 

int AuthReadSimpleList
(
struct RequestStruct* rqptr,
char *DatabaseName,
boolean AuthenticatePassword
)
{
   static char  AuthFileName [64];
   static $DESCRIPTOR (AuthFileNameFaoDsc, "!AZ!AZ!AZ");
   static $DESCRIPTOR (AuthFileNameDsc, AuthFileName);

   register char  *cptr, *sptr;

   int  status;
   unsigned short  AuthFileNameLength;
   char  Line [1024];
   struct FAB  AuthFileFab;
   struct NAM  AuthFileNam;
   struct RAB  AuthFileRab;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "AuthReadSimpleList() |%s|%s| %d\n",
               DatabaseName, rqptr->RemoteUser, AuthenticatePassword);

   /* must supply a user name and database name */
   if (!rqptr->RemoteUser[0] || !DatabaseName[0]) return (SS$_INVLOGIN);

   sys$fao (&AuthFileNameFaoDsc, &AuthFileNameLength, &AuthFileNameDsc,
            HTA_DIRECTORY, DatabaseName, HTL_FILE_TYPE);
   AuthFileName[AuthFileNameLength] = '\0';
   if (Debug) fprintf (stdout, "AuthFileName |%s|\n", AuthFileName);

   AuthFileFab = cc$rms_fab;
   AuthFileFab.fab$b_fac = FAB$M_GET;
   AuthFileFab.fab$l_fna = AuthFileName;  
   AuthFileFab.fab$b_fns = AuthFileNameLength;
   AuthFileFab.fab$b_shr = FAB$M_SHRGET;

   /* turn on SYSPRV to allow access to authentication database file */
   EnableSysPrv ();
   status = sys$open (&AuthFileFab, 0, 0);
   DisableSysPrv ();

   /* status from sys$open() */
   if (VMSnok (status))
   {
      rqptr->ErrorTextPtr = MsgFor(rqptr,MSG_AUTH_DATABASE);
      rqptr->ErrorOtherTextPtr = DatabaseName;
      ErrorVmsStatus (rqptr, status, FI_LI);
      return (status);
   }

   AuthFileRab = cc$rms_rab;
   AuthFileRab.rab$l_fab = &AuthFileFab;
   AuthFileRab.rab$b_mbf = 2;
   AuthFileRab.rab$l_rop = RAB$M_RAH;

   if (VMSnok (status = sys$connect (&AuthFileRab, 0, 0)))
   {
      sys$close (&AuthFileFab, 0, 0);
      return (status);
   }

   /*****************/
   /* locate record */
   /*****************/

   AuthFileRab.rab$l_ubf = Line;
   AuthFileRab.rab$w_usz = sizeof(Line)-1;

   while (VMSok (status = sys$get (&AuthFileRab, 0, 0)))
   {
      AuthFileRab.rab$l_rbf[AuthFileRab.rab$w_rsz] = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", AuthFileRab.rab$l_rbf);

      if (AuthFileRab.rab$w_rsz)
      {
         if (AuthFileRab.rab$l_ubf[AuthFileRab.rab$w_rsz-1] == '\\')
         {
            /* directive is continued on next line */
            AuthFileRab.rab$l_ubf[AuthFileRab.rab$w_rsz-1] = ' ';
            AuthFileRab.rab$l_ubf += AuthFileRab.rab$w_rsz;
            AuthFileRab.rab$w_usz -= AuthFileRab.rab$w_rsz;
            continue;
         }
      }

      if (Debug) fprintf (stdout, "|%s|\n", Line);
      for (cptr = Line; *cptr && ISLWS(*cptr); cptr++)
      if (*cptr == '#' || *cptr == '!') continue;

      sptr = rqptr->RemoteUser;
      while (*cptr && !ISLWS(*cptr) && *cptr != '=' && *sptr &&
             tolower(*cptr) == tolower(*sptr))
         { cptr++; sptr++; }
      if ((*cptr && !ISLWS(*cptr) && *cptr != '=') || *sptr) continue;
      break;
   }
   if (Debug) fprintf (stdout, "sys$get() %%X%08.08X\n", status);

   sys$close (&AuthFileFab);

   if (status == RMS$_EOF)
   {
      /* user record not found */
      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "FAIL !AZ!AZ username", DatabaseName,
                         AuthSourceString (DatabaseName, AUTH_SOURCE_LIST));
      return (SS$_INVLOGIN);
   }

   /* if just checking if the user name is an entry in the list file */
   if (!AuthenticatePassword) return (status);

   /* simple, clear-text password comparison */
   if (*cptr != '=')
   {
      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "FAIL !AZ!AZ no password", DatabaseName,
                         AuthSourceString (DatabaseName, AUTH_SOURCE_LIST));
      return (SS$_INVLOGIN);
   }
   while (*cptr && *cptr == '=') cptr++;
   sptr = rqptr->RemoteUserPassword;
   while (*cptr && !ISLWS(*cptr) && *sptr && tolower(*cptr) == tolower(*sptr))
      { cptr++; sptr++; }
   if ((*cptr && !ISLWS(*cptr)) || *sptr)
   {
      /* passwords do not match */
      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "FAIL !AZ!AZ password", DatabaseName,
                         AuthSourceString (DatabaseName, AUTH_SOURCE_LIST));
      return (SS$_INVLOGIN);
   }

   /* passwords match */
   return (status);
}

/*****************************************************************************/
/*
Verify the request username/password from the on-disk SYSUAF database.

Get the specified user's flags, authorized privileges, quadword password, hash
salt and encryption algorithm from SYSUAF.  If any of specified bits in the
flags are set (e.g. "disusered") then fail the password authentication.
Using the salt and encryption algorithm hash the supplied password and compare
it to the UAF hashed password.  A SYSPRV privileged account is always failed.

Returns a success status if user password authenticated, SS$_INVLOGIN if not
authenticated, or other error status if a genuine VMS error occurs (which
should be reported where and when it occurs).
*/ 

int AuthVerifyVmsPassword
(
struct RequestStruct* rqptr,
struct AuthCacheRecordStruct *acrptr
)
{
   /* UAI flags that disallow an identifier-controlled account access */
   static unsigned long  DisallowIdFlags =
          UAI$M_DISACNT | UAI$M_PWD_EXPIRED | UAI$M_PWD2_EXPIRED;

   /* UAI flags that disallow SYSUAF-controlled account access */
   static unsigned long  DisallowVmsFlags =
          UAI$M_DISACNT | UAI$M_PWD_EXPIRED | UAI$M_PWD2_EXPIRED |
          UAI$M_CAPTIVE | UAI$M_RESTRICTED;

   static unsigned long  Context = -1;

   static unsigned long  UaiFlags,
                         UaiUic;
   static unsigned long  UaiPriv [2],
                         HashedPwd [2],
                         UaiPwd [2];
   static unsigned short  UaiSalt;
   static unsigned char  UaiEncrypt;

   static struct {
      short BufferLength;
      short ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } UaiItems [] = 
   {
      { sizeof(UaiUic), UAI$_UIC, &UaiUic, 0 },
      { sizeof(UaiFlags), UAI$_FLAGS, &UaiFlags, 0 },
      { sizeof(UaiPriv), UAI$_PRIV, &UaiPriv, 0 },
      { sizeof(UaiPwd), UAI$_PWD, &UaiPwd, 0 },
      { sizeof(UaiEncrypt), UAI$_ENCRYPT, &UaiEncrypt, 0 },
      { sizeof(UaiSalt), UAI$_SALT, &UaiSalt, 0 },
      { sizeof(unsigned long), UAI$_PRIMEDAYS, 0, 0 },
      { 3, UAI$_NETWORK_ACCESS_P, 0, 0 },
      { 3, UAI$_NETWORK_ACCESS_S, 0, 0 },
      { 3, UAI$_REMOTE_ACCESS_P, 0, 0 },
      { 3, UAI$_REMOTE_ACCESS_S, 0, 0 },
      { 0, 0, 0, 0 }
   };

   int  status;
   $DESCRIPTOR (UserNameDsc, rqptr->RemoteUser);
   $DESCRIPTOR (PasswordDsc, rqptr->RemoteUserPassword);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthVerifyVmsPassword()\n");

   /* double-check! */
   if (!AuthSysUafEnabled) return (SS$_INVLOGIN);

   acrptr->UaiPrimeDays =
     acrptr->UaiNetworkAccessP = acrptr->UaiNetworkAccessS =
     acrptr->UaiRemoteAccessP = acrptr->UaiRemoteAccessS = 0;

   UaiItems[6].BufferPtr = &acrptr->UaiPrimeDays;
   UaiItems[7].BufferPtr = &acrptr->UaiNetworkAccessP;
   UaiItems[8].BufferPtr = &acrptr->UaiNetworkAccessS;
   UaiItems[9].BufferPtr = &acrptr->UaiRemoteAccessP;
   UaiItems[10].BufferPtr = &acrptr->UaiRemoteAccessS;

   /* turn on SYSPRV to allow access to SYSUAF records */
   EnableSysPrv ();

   UserNameDsc.dsc$w_length = strlen(rqptr->RemoteUser);
   status = sys$getuai (0, &Context, &UserNameDsc, &UaiItems, 0, 0, 0);
   if (Debug) fprintf (stdout, "sys$getuai() %%X%08.08X\n", status);

   /* turn off SYSPRV */
   DisableSysPrv ();

   if (VMSnok (status)) 
   {
      if (status == RMS$_RNF)
      {
         if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "FAIL SYSUAF username");
         return (SS$_INVLOGIN);
      }
      rqptr->ErrorTextPtr = MsgFor(rqptr,MSG_AUTH_DATABASE_VMS);
      ErrorVmsStatus (rqptr, status, FI_LI);
      return (status);
   }

   /* automatically disallow if any of these flags are set! */
   if ((AuthPolicySysUafIdentifiers && (UaiFlags & DisallowIdFlags)) ||
       (AuthPolicySysUafWasdIdentifiers && (UaiFlags & DisallowIdFlags)) ||
       (!AuthPolicySysUafIdentifiers && (UaiFlags & DisallowVmsFlags)))
   {
      if (Debug) fprintf (stdout, "failed on flags\n");
      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "FAIL SYSUAF flags");
      return (SS$_INVLOGIN);
   }

   PasswordDsc.dsc$w_length = strlen(rqptr->RemoteUserPassword);
   status = sys$hash_password (&PasswordDsc, UaiEncrypt,
                               UaiSalt, &UserNameDsc, &HashedPwd);
   if (Debug) fprintf (stdout, "sys$hash_password() %%X%08.08X\n", status);
   if (VMSnok (status))
   {
      rqptr->ErrorTextPtr = MsgFor(rqptr,MSG_AUTH_USER);
      ErrorVmsStatus (rqptr, status, FI_LI);
      return (status);
   }

   if (HashedPwd[0] != UaiPwd[0] || HashedPwd[1] != UaiPwd[1])
   {
      if (Debug) fprintf (stdout, "failed on password match\n");
      if (!AuthSysUafPromiscuous)
      {
         if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "FAIL SYSUAF password");
         return (SS$_INVLOGIN);
      }
   }

   /* identifier checking overrides privilege checking! */
   if (AuthPolicySysUafIdentifiers)
   {
      /* check SYSUAF-authenticated account holds the correct identifier */
      if (VMSnok (status = AuthLoadVmsIdentifiers (rqptr, UaiUic)))
         return (status);
      return (AuthSysUafCheckIdentifier (rqptr));
   }

   if (!AuthPolicySysUafAllAccounts && (UaiPriv[0] & PRV$M_SYSPRV))
   {
      /* not allowing all accounts, exclude those with extended privileges */
      if (Debug) fprintf (stdout, "failed on privileges\n");
      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "FAIL SYSUAF privileges");
      return (SS$_INVLOGIN);
   }

   /* authenticated ... user can do anything (the path allows!) */
   rqptr->AuthUserCan = AUTH_READWRITE_ACCESS;

   if (Debug) fprintf (stdout, "SYSUAF verified!\n");
   return (SS$_NORMAL);
}

/****************************************************************************/
/*
SYSUAF authentication is being controlled by possession of an identifier ...
got the identifier you can be SYSUAF-authenticated, haven't then you cant! See
description at beginning of module.  Returns normal status if account possesses
suitable identifier, AUTH_RESTRICTED_BY_OTHER if it doesn't, or error status.
*/ 

AuthSysUafCheckIdentifier (struct RequestStruct *rqptr)

{
   boolean  HoldsRealmId,
            HoldsWasdReadId,
            HoldsWasdWriteId;
   unsigned long  ThisId;
   unsigned long  *idptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthSysUafCheckIdentifier()\n");

   /* can't do anything until allowed to! */
   rqptr->AuthUserCan = 0;

   if (rqptr->AuthVmsIdentifiersPtr == NULL)
      ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);

   HoldsRealmId = HoldsWasdReadId = HoldsWasdWriteId = false;

   for (idptr = rqptr->AuthVmsIdentifiersPtr; ThisId = *idptr; idptr++)
   {
      if (Debug) fprintf (stdout, "%08.08X\n", ThisId);

      if (rqptr->AuthSourceRealm == AUTH_SOURCE_ID &&
          ThisId == rqptr->AuthRealmVmsIdentifier)
      {                     
         HoldsRealmId = true;

         if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "HOLDS identifier !AZ", rqptr->AuthRealmPtr);

         continue;
      }

      if (rqptr->AuthSourceRealm == AUTH_SOURCE_WASD_ID)
      {
         if (ThisId == AuthWasdReadVmsIdentifier)
         {
            HoldsWasdReadId = true;

            if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
               AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                               "HOLDS identifier !AZ", AUTH_WASD_READ_VMS_ID);

            continue;
         }

         if (ThisId == AuthWasdWriteVmsIdentifier)
         {                     
            HoldsWasdWriteId = true;

            if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
               AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                               "HOLDS identifier !AZ", AUTH_WASD_WRITE_VMS_ID);

            continue;
         }
      }

      if (ThisId == AuthHttpsOnlyVmsIdentifier)
      {
         /* account is only allowed to authenticate via SSL */
         rqptr->AuthHttpsOnly = true;

         if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "HOLDS identifier !AZ", AUTH_HTTPS_ONLY_VMS_ID);

         continue;
      }

      if (ThisId == AuthWasdHttpsVmsIdentifier)
      {
         /* account is only allowed to authenticate via SSL */
         rqptr->AuthHttpsOnly = true;

         if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "HOLDS identifier !AZ", AUTH_WASD_HTTPS_VMS_ID);

         continue;
      }

      if (ThisId == AuthWasdPwdVmsIdentifier)
      {
         /* account is allowed to change it's password via the server */
         rqptr->AuthCanChangeSysUafPwd = true;

         if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
            AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                            "HOLDS identifier !AZ", AUTH_WASD_PWD_VMS_ID);

         continue;
      }
   }

   /************************/
   /* check authentication */
   /************************/

   if (rqptr->AuthSourceRealm == AUTH_SOURCE_ID &&
       !HoldsRealmId)
   {
      /* authentication is by local identifier ... and doesn't have it */
      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "FAIL SYSUAF access identifier");
      return (AUTH_RESTRICTED_BY_OTHER);
   }

   if (rqptr->AuthSourceRealm == AUTH_SOURCE_WASD_ID &&
       !HoldsWasdWriteId &&
       !HoldsWasdReadId)
   {
      /* authentication only with WASD identifier .. and doesn't have it */
      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "FAIL SYSUAF access identifier");
      return (AUTH_RESTRICTED_BY_OTHER);
   }

   /* authenticated ... user can do anything (the path allows!) */
   rqptr->AuthUserCan = AUTH_READWRITE_ACCESS;

   if (Debug) fprintf (stdout, "identifier verified!\n");
   return (SS$_NORMAL);
}

/****************************************************************************/
/*
*/ 

AuthLoadVmsIdentifiers
(
struct RequestStruct *rqptr,
unsigned long  UaiUic
)
{
   static struct {
      unsigned long  Uic;
      unsigned long  Unused;
   } Holder = { 0, 0 };

   int  status,
        idcnt;
   unsigned long  Ctx,
                  ThisId;
   unsigned long  *idptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthLoadVmsIdentifiers()\n");

   /* if already loaded then just continue */
   if (rqptr->AuthVmsIdentifiersPtr != NULL) return (SS$_NORMAL);

   /* plus one longword as a sentinal (will contain zero) */
   rqptr->AuthVmsIdentifiersPtr = idptr =
      (unsigned long*)VmGetHeap (rqptr, (AUTH_MAX_VMS_IDENTIFIERS + 1) *
                                         sizeof(unsigned long));

   Holder.Uic = UaiUic;

   idcnt = AUTH_MAX_VMS_IDENTIFIERS;

   /* turn on SYSPRV to allow this to be done */
   EnableSysPrv ();

   Ctx = 0;
   while (VMSok (status = sys$find_held (&Holder, &ThisId, 0, &Ctx)))
   {
      if (Debug) fprintf (stdout, "%08.08X\n", ThisId);
      if (!--idcnt)
      {
         status = SS$_BUFFEROVF;
         break;
      }
      *idptr++ = ThisId;
   }
   /* sentinal (not really necessary, but let's be over-cautious) */
   *idptr = 0;

   /* turn off SYSPRV */
   DisableSysPrv ();

   if (Debug) fprintf (stdout, "sys$find_held() %%X%08.08X\n", status);

   /* if we got to the end of the identifiers without mishap */
   if (status == SS$_NOSUCHID) return (SS$_NORMAL);

   /*********/
   /* error */
   /*********/

   sys$finish_rdb (&Ctx);
   rqptr->ErrorTextPtr = MsgFor(rqptr,MSG_AUTH_DATABASE_VMS);
   ErrorVmsStatus (rqptr, status, FI_LI);
   return (status);
}

/****************************************************************************/
/*
Look through the request's array of a VMS authenticated user's identifiers for
the specified one.  Return zero to indicate the identifier was not held, normal
to indicate it was, or any other error status.
*/ 

int AuthHoldsVmsIdentifier
(
struct RequestStruct *rqptr,
char *IdentifierName,
unsigned long ThisId
)
{
   register unsigned long  *idptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthHoldsVmsIdentifier() %08.08X\n", ThisId);

   /* if not already loaded via SYSUAF authentication then really an error */
   if (rqptr->AuthVmsIdentifiersPtr == NULL)
   {
      /* has NOT been authenticated via the SYSUAF (shouldn't happen!) */

      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "IDENTIFIER request not SYSUAF authenticated");

      rqptr->ResponseStatusCode = 403;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_AUTH_ACCESS_DENIED), FI_NOLI);

      return (SS$_ABORT);
   }

   if (!ThisId) return (false);

   for (idptr = rqptr->AuthVmsIdentifiersPtr;
        *idptr && *idptr != ThisId;
        idptr++);

   if (*idptr)
   {
      if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_AUTH,
                         "HOLDS identifier !AZ", IdentifierName);

      return (SS$_NORMAL);
   }

   return (0);
}

/****************************************************************************/
/*
*/ 

AuthHoldsWasdGroupIdentifier
(
struct RequestStruct *rqptr,
char *GroupName
)
{
   static char GroupIdName [AUTH_MAX_REALM_GROUP_LENGTH+1] =
                           AUTH_WASD__GROUP_VMS_ID;
   static $DESCRIPTOR (GroupIdNameDsc, GroupIdName);

   register char  *cptr, *sptr, *zptr;

   int  status;
   unsigned long  ThisId;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "AuthHoldsWasdGroupIdentifier() |%s|\n", GroupName);

   zptr = (sptr = GroupIdName) + sizeof(GroupIdName);
   for (cptr = GroupName; *cptr && sptr < zptr; *sptr++ = *cptr++);
   if (sptr >= zptr)
   {
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_GENERAL_OVERFLOW), FI_LI);
      return (SS$_ABORT);
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", GroupIdName);
   GroupIdNameDsc.dsc$w_length = sptr - GroupIdName;
   if (VMSnok (status = sys$asctoid (&GroupIdNameDsc, &ThisId, 0)))
   {
      rqptr->ErrorTextPtr = GroupIdName;
      ErrorVmsStatus (rqptr, status, FI_LI);
      return (status);
   }

   return (AuthHoldsVmsIdentifier (rqptr, GroupIdName, ThisId));
}

/****************************************************************************/
/*
Place the VMS-hashed password into the pointed to quadword.  This CANNOT be
used to hash the UAF password as it does not use the UAF entry salt, etc.
*/ 

AuthGenerateHashPassword
(
char *UserName,
char *Password,
unsigned long *HashedPwdPtr
)
{
   static $DESCRIPTOR (PasswordDsc, "");
   static $DESCRIPTOR (UserNameDsc, "");

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "AuthGenerateHashPassword() |%s|%s|\n",
               UserName, Password);

   UserNameDsc.dsc$w_length = strlen(UserName);
   UserNameDsc.dsc$a_pointer = UserName;

   PasswordDsc.dsc$w_length = strlen(Password);
   PasswordDsc.dsc$a_pointer = Password;

   status = sys$hash_password (&PasswordDsc, UAI$C_PURDY_S, 0,
                               &UserNameDsc, HashedPwdPtr);
   if (Debug)
      fprintf (stdout, "sys$hash_password() %%X%08.08X|%08.08X%08.08X|\n",
               status, HashedPwdPtr[1], HashedPwdPtr[0]);
   return (status);
}

/****************************************************************************/
/*
Place the MD5 Digest upper and lower case passwords into to pointed to
two 16-byte arrays.  This is the 128 bit, binary digest, NOT the hexadecimal
version.  See note on upper and lower case versions in program comments.
*/ 

AuthGenerateDigestPassword
(
char *DatabaseName,
char *UserName,
char *Password,
unsigned char *A1DigestLoCasePtr,
unsigned char *A1DigestUpCasePtr
)
{
   register char  *cptr, *sptr;

   int  status;
   char PasswordUpCase [AUTH_MAX_PASSWORD_LENGTH+1],
        PasswordLoCase [AUTH_MAX_PASSWORD_LENGTH+1],
        UserNameUpCase [AUTH_MAX_PASSWORD_LENGTH+1],
        UserNameLoCase [AUTH_MAX_PASSWORD_LENGTH+1];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "AuthGenerateDigestPassword() |%s|%s|%s|\n",
               DatabaseName, UserName, Password);

   sptr = PasswordUpCase;
   for (cptr = Password; *cptr; *sptr++ = toupper(*cptr++));
   *sptr = '\0';
   sptr = PasswordLoCase;
   for (cptr = Password; *cptr; *sptr++ = tolower(*cptr++));
   *sptr = '\0';
   sptr = UserNameUpCase;
   for (cptr = UserName; *cptr; *sptr++ = toupper(*cptr++));
   *sptr = '\0';
   sptr = UserNameLoCase;
   for (cptr = UserName; *cptr; *sptr++ = tolower(*cptr++));
   *sptr = '\0';
   if (VMSok (status =
       DigestA1 (UserNameUpCase, DatabaseName, PasswordUpCase,
                 A1DigestUpCasePtr)))
       status = DigestA1 (UserNameLoCase, DatabaseName, PasswordLoCase,
                          A1DigestLoCasePtr);

   return (status);
}

/*****************************************************************************/
/*
Using sys$create_user_profile() create a permanent security profile of the
SYSUAF-authenticated user name.  This will be stored in the cache and attached
to the request structure each time the user is re-authenticated from it.  This
profile will be used to check file/directory (see AuthCheckVmsUserAccess())
access permission against the authenticated user not the server account.
*/

int AuthCreateVmsUserProfile (struct RequestStruct* rqptr)

{
   static unsigned long  Flags = 0;
   static unsigned long  Context = -1;

   int  status;
   unsigned short  Length;
   unsigned char  UserProfile [2048];
   $DESCRIPTOR (UserNameDsc, rqptr->RemoteUser);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "AuthCreateVmsUserProfile() |%s|\n", rqptr->RemoteUser);

   rqptr->AuthVmsUserProfilePtr = NULL;
   rqptr->AuthVmsUserProfileLength = 0;

   /* turn on SYSPRV to allow this to be done */
   EnableSysPrv ();

   UserNameDsc.dsc$w_length = strlen(rqptr->RemoteUser);
   Length = sizeof(UserProfile);
   status = sys$create_user_profile (&UserNameDsc, 0, Flags,
                                     UserProfile, &Length, &Context);
   if (Debug)
      fprintf (stdout, "sys$create_user_profile() %%X%08.08X\n", status);

   /* turn off SYSPRV */
   DisableSysPrv ();

   if (VMSnok (status)) 
   {
      rqptr->ErrorTextPtr = "sys$create_user_profile()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return (status);
   }

   if (Debug) fprintf (stdout, "Length %d\n", Length);

   rqptr->AuthVmsUserProfilePtr =
      VmGet (rqptr->AuthVmsUserProfileLength = Length);
   if (Debug) fprintf (stdout, "ptr: %d\n", rqptr->AuthVmsUserProfilePtr);

   memcpy (rqptr->AuthVmsUserProfilePtr, UserProfile, Length);
      
   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Using sys$create_user_profile() create a permanent security profile of the
HTTPd server account.  Returns no status, exits server if error encountered.
*/ 
 
AuthCreateHttpdProfile ()

{
   static unsigned long  Flags = 0;

   int  status;
   char  HttpdUserProfile [512];
   $DESCRIPTOR (HttpdUserNameDsc, "");

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCreateHttpdProfile()\n");

   if (HttpdUserProfileLength) return;

   HttpdUserNameDsc.dsc$a_pointer = HttpdUserName;
   HttpdUserNameDsc.dsc$w_length = strlen(HttpdUserName);

   status = sys$create_user_profile (&HttpdUserNameDsc, 0, Flags,
                                     HttpdUserProfile,
                                     &HttpdUserProfileLength, 0);
   if (Debug)
      fprintf (stdout, "sys$create_user_profile() %%X%08.08X\n", status);

   if (VMSnok (status))
      ErrorExitVmsStatus (status, "sys$create_user_profile()", FI_LI);

   HttpdUserProfilePtr = VmGet (HttpdUserProfileLength);
   if (Debug) fprintf (stdout, "ptr: %d\n", HttpdUserProfilePtr);

   memcpy (HttpdUserProfilePtr, HttpdUserProfile, HttpdUserProfileLength);
}

/*****************************************************************************/
/*
Using sys$check_access() check file read access permission for the supplied
file name against the SYSUAF-authenticated user's security profile (created by
AuthCreateVmsUserProfile()).  File or directory specifications can be supplied.
Returns SS$_NORMAL if access allowed, RMS$_PRV if denied, and other RMS status
codes. A VMS error other than from RMS (e.g. INSFMEM, which would disrupt the
system service results) will cause the server to exit.
*/

int AuthCheckVmsUserAccess
(
struct RequestStruct* rqptr,
char *FileName,
int FileNameLength
)
{
   static unsigned long  Flags = 0,
                         ArmReadAccess = ARM$M_READ;
   static $DESCRIPTOR (ClassNameDsc, "FILE");
   static $DESCRIPTOR (UserProfileDsc, "");
   static struct
   {
      short BufferLength;
      short ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   ReadItem [] = 
   {
      { sizeof(ArmReadAccess), CHP$_ACCESS, &ArmReadAccess, 0 },
      { 0, 0, 0, 0 }
   };

   int  status,
        DirFileNameLength;
   char  DirFileName [256];
   $DESCRIPTOR (FileNameDsc, FileName);

   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      if (rqptr == NULL)
         fprintf (stdout, "AuthCheckVmsUserAccess() |%s|\n", HttpdUserName);
      else
         fprintf (stdout, "AuthCheckVmsUserAccess() |%s|%s| %d %d\n",
                  rqptr->RemoteUser, FileName,
                  rqptr->AuthVmsUserProfileLength,
                  rqptr->AuthVmsUserProfilePtr);
   }

   if (FileNameLength <= 0) FileNameLength = strlen(FileName);

   if (FileNameLength && FileName[FileNameLength-1] == ']')
   {
      /* a directory has been specified, get the directory file name */
      NameOfDirectoryFile (FileName, FileNameLength,
                           DirFileName, &DirFileNameLength);
      FileNameDsc.dsc$a_pointer = DirFileName;
      FileNameDsc.dsc$w_length = DirFileNameLength;
   }
   else
   {
      FileNameDsc.dsc$a_pointer = FileName;
      FileNameDsc.dsc$w_length = FileNameLength;
   }

   /****************/
   /* check access */
   /****************/

   if (rqptr == NULL)
   {
      /* against the HTTPd server account */
      if (!HttpdUserProfileLength) AuthCreateHttpdProfile ();
      UserProfileDsc.dsc$a_pointer = HttpdUserProfilePtr;
      UserProfileDsc.dsc$w_length = HttpdUserProfileLength;
   }
   else
   {
      /* against a SYSUAF-authenticated account */
      UserProfileDsc.dsc$a_pointer = rqptr->AuthVmsUserProfilePtr;
      UserProfileDsc.dsc$w_length = rqptr->AuthVmsUserProfileLength;
   }

   /* turn on SYSPRV to allow this to be done */
   EnableSysPrv ();

   status = sys$check_access (0, &FileNameDsc, 0, &ArmReadAccess,
                              0, &ClassNameDsc, 0, &UserProfileDsc);
   if (Debug) fprintf (stdout, "sys$check_access() %%X%08.08X\n", status);

   /* turn off SYSPRV */
   DisableSysPrv ();

   /* we are only dealing with files ... turn it into an RMS message */
   if (status == SS$_NOPRIV) return (RMS$_PRV);

   /* if access is permitted */
   if (VMSok (status)) return (status);

   /* if an RMS error return normal letting caller continue and report it */
   if (((status & STS$M_FAC_NO) >> STS$V_FAC_NO) == RMS$_FACILITY)
      return (SS$_NORMAL);

   /* cannot generate error message, do it at the calling routine */
   if (rqptr == NULL) return (status);

   rqptr->ErrorTextPtr = "sys$check_access()";
   rqptr->ErrorOtherTextPtr = FileName;
   ErrorVmsStatus (rqptr, status, FI_LI);
   return (status);
}

/*****************************************************************************/
/*
Check that the HTTPd account has CONTROL or WRITE access to the parent
directory of the file or directory specification supplied in 'FileName'. If
the server has CONTROL access then there is no general access to the
directory, only an authenticated VMS user can write into it. If there is WRITE
access then any authenticated user can write into it. Returns SS$_NORMAL if
access allowed, RMS$_PRV if denied, and other status codes.  Error report
should be generated by the calling routine.
*/ 
 
int AuthCheckWriteAccess
(
struct RequestStruct *rqptr,
char *FileName,
int FileNameLength
)
{
   static unsigned long  Flags = 0;
   static unsigned long  ArmWriteAccess = ARM$M_WRITE,
                         ArmControlAccess = ARM$M_CONTROL;

   static unsigned short  HttpdUserProfileLength = 0;
   static char  HttpdUserProfile [512];

   static struct {
      unsigned short  BufferLength;
      unsigned short  ItemCode;
      unsigned long  BufferAddress;
      unsigned long  ReturnLengthAddress;
   } WriteAccessItem [] =
       { { sizeof(ArmWriteAccess), CHP$_ACCESS, &ArmWriteAccess, 0 },
         { 0, 0, 0, 0 } },
     ControlAccessItem [] =
       { { sizeof(ArmControlAccess), CHP$_ACCESS, &ArmControlAccess, 0 },
         { 0, 0, 0, 0 } };

   static $DESCRIPTOR (ClassNameDsc, "FILE");
   static $DESCRIPTOR (HttpdUserProfileDsc, HttpdUserProfile);
   static $DESCRIPTOR (ObjectNameDsc, "");
   static $DESCRIPTOR (UserProfileDsc, "");

   int  status,
        DirectoryFileLength;
   char  DirectoryFile [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCheckWriteAccess() |%s|\n", FileName);

   if (!HttpdUserProfileLength)
   {
      /*********************************************************/
      /* create a permanent user profile for the HTTPd account */
      /*********************************************************/

      $DESCRIPTOR (HttpdUserNameDsc, "");

      HttpdUserNameDsc.dsc$a_pointer = HttpdUserName;
      HttpdUserNameDsc.dsc$w_length = strlen(HttpdUserName);

      status = sys$create_user_profile (&HttpdUserNameDsc, 0, Flags,
                                        HttpdUserProfile,
                                        &HttpdUserProfileLength, 0);
      if (Debug)
         fprintf (stdout, "sys$create_user_profile() %%X%08.08X\n", status);
      if (VMSnok (status)) return (status);

      HttpdUserProfileDsc.dsc$w_length = HttpdUserProfileLength;
   }

   if (FileNameLength <= 0) FileNameLength = strlen(FileName);

   NameOfDirectoryFile (FileName, FileNameLength,
                        DirectoryFile, &DirectoryFileLength);

   ObjectNameDsc.dsc$a_pointer = DirectoryFile;
   ObjectNameDsc.dsc$w_length = DirectoryFileLength;

   /* turn on SYSPRV to allow this to be done */
   EnableSysPrv ();

   if (AuthVmsUserSecProfileEnabled)
   {
      /****************************/
      /* check for control access */
      /****************************/

      status = sys$check_access (0, &ObjectNameDsc, 0, &ControlAccessItem,
                                 0, &ClassNameDsc, 0, &HttpdUserProfileDsc);
      if (Debug) fprintf (stdout, "1 sys$check_access() %%X%08.08X\n", status);

      if (VMSok (status))
      {
         /* control access for HTTPd, must be authenticated VMS user! */
         if (!rqptr->AuthVmsUserProfileLength)
         {
            /* turn off SYSPRV! */
            DisableSysPrv ();
            /* not a SYSUAF-authenticated user therefore no privilege */
            return (SS$_NOPRIV);
         }

         /* now check if the authenticated VMS user has write access */
         UserProfileDsc.dsc$a_pointer = rqptr->AuthVmsUserProfilePtr;
         UserProfileDsc.dsc$w_length = rqptr->AuthVmsUserProfileLength;

         status = sys$check_access (0, &ObjectNameDsc, 0, &WriteAccessItem,
                                    0, &ClassNameDsc, 0, &UserProfileDsc);
         if (Debug)
            fprintf (stdout, "2 sys$check_access() %%X%08.08X\n", status);

         /* turn off SYSPRV */
         DisableSysPrv ();

         /* no privilege to write */
         if (status == SS$_NOPRIV) return (status);

         /* write access OK */
         if (VMSok (status)) return (status);

         /* if the directory file not exist then directory does not exist */
         if (status == RMS$_FNF) status = RMS$_DNF;

         return (status);
      }

      /* no control access, drop through to check for write access */
   }

   /**************************/
   /* check for write access */
   /**************************/

   status = sys$check_access (0, &ObjectNameDsc, 0, &WriteAccessItem,
                              0, &ClassNameDsc, 0, &HttpdUserProfileDsc);
   if (Debug) fprintf (stdout, "3 sys$check_access() %%X%08.08X\n", status);

   /* turn off SYSPRV */
   DisableSysPrv ();

   /* no privilege to write */
   if (status == SS$_NOPRIV) return (status);

   /* write access OK */
   if (VMSok (status)) return (status);

   /* if the directory file did not exist then directory does not exist */
   if (status == RMS$_FNF) status = RMS$_DNF;

   return (status);
}

/*****************************************************************************/
/*
Open the authorization configuration file defined by the logical name
HTTPD$AUTH.  Read the directives applying them to the authorization path tree. 
Close the file.  Can also be called from the HTTPd.c or Control.c modules.
*/

int AuthPathConfig (struct AuthConfigStruct *acptr)

{
   boolean  DebugBuffer;
   int  status;
   char  ExpandedFileName [256],
         Line [512],
         LineBuffer [512];
   struct FAB  AuthFileFab;
   struct NAM  AuthFileNam;
   struct RAB  AuthFileRab;
   struct XABDAT  AuthFileXabDat;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthPathConfig()\n");

   if (acptr == NULL)
   {
      /**************/
      /* initialize */
      /**************/

      acptr = &AuthServerConfig;
      AuthCacheTreeFree ();
      AuthPathListFree (acptr);
      if (acptr->ProblemReportPtr != NULL) free (acptr->ProblemReportPtr);
      memset (acptr, 0, sizeof(struct AuthConfigStruct));

      return (SS$_NORMAL);
   }

#ifdef DBUG
   /* not interested anymore in seeing debug information for path load! */
   DebugBuffer = Debug;
   Debug = false;
#endif

   /* initialize */
   AuthPathLine (acptr, NULL);
   sys$gettim (&acptr->LoadBinTime);

   /************************/
   /* open the config file */
   /************************/

   AuthFileFab = cc$rms_fab;
   AuthFileFab.fab$b_fac = FAB$M_GET;
   AuthFileFab.fab$l_fna = CONFIG_AUTH_FILE_NAME;
   AuthFileFab.fab$b_fns = strlen(CONFIG_AUTH_FILE_NAME);
   AuthFileFab.fab$l_nam = &AuthFileNam;
   AuthFileFab.fab$b_shr = FAB$M_SHRGET;
   AuthFileFab.fab$l_xab = &AuthFileXabDat;

   AuthFileNam = cc$rms_nam;
   AuthFileNam.nam$l_esa = ExpandedFileName;
   AuthFileNam.nam$b_ess = sizeof(ExpandedFileName)-1;
   AuthFileNam.nam$l_rsa = acptr->LoadFileName;
   AuthFileNam.nam$b_rss = sizeof(acptr->LoadFileName)-1;

   /* initialize the date extended attribute block */
   AuthFileXabDat = cc$rms_xabdat;

   /* turn on SYSPRV to allow access to possibly protected file */
   EnableSysPrv ();
   status = sys$open (&AuthFileFab, 0, 0);
   DisableSysPrv ();

   /* status from sys$open() */
   if (VMSnok (status))
   {
      if (Debug) fprintf (stdout, "sys$open() %%X%08.08X\n", status);
      AuthConfigReport (acptr,
         "%HTTPD-W-AUTH, opening configuration file\n!-!AZ\n",
         SysGetMsg(status)+1);
      /* loading at startup, report any errors */
      if (acptr->ProblemCount && !acptr->RequestPtr)
         fputs (acptr->ProblemReportPtr, stdout);
#ifdef DBUG
      Debug = DebugBuffer;
#endif
      return (status);
   }

   acptr->LoadFileName[AuthFileNam.nam$b_rsl] = '\0';
   if (Debug)
      fprintf (stdout, "LoadFileName |%s|\n", acptr->LoadFileName);

   memcpy (acptr->RevBinTime, &AuthFileXabDat.xab$q_rdt, 8);

   /* record access block */
   AuthFileRab = cc$rms_rab;
   AuthFileRab.rab$l_fab = &AuthFileFab;
   /* 2 buffers */
   AuthFileRab.rab$b_mbf = 2;
   /* read ahead performance option */
   AuthFileRab.rab$l_rop = RAB$M_RAH;

   if (VMSnok (status = sys$connect (&AuthFileRab, 0, 0)))
   {
      sys$close (&AuthFileFab, 0, 0);
      if (Debug) fprintf (stdout, "sys$connect() %%X%08.08X\n", status);
      AuthConfigReport (acptr,
         "%HTTPD-W-AUTH, opening configuration file\n!-!AZ\n",
         SysGetMsg(status)+1);
      /* loading at startup, report any errors */
      if (acptr->ProblemCount && !acptr->RequestPtr)
         fputs (acptr->ProblemReportPtr, stdout);
#ifdef DBUG
      Debug = DebugBuffer;
#endif
      return (status);
   }

   /* guarantee access to the administration menu */
   if (AuthPromiscuous)
   {
      /* these may generate duplicate rule errors at startup! */
      AuthPathLine (acptr, "[PROMISCUOUS]");
      /* access to the server administration menu */
      strcpy (Line, "/httpd/-/admin/* r+w");
      AuthPathLine (acptr, Line);
      /* access to write into the "normal" location for configuration files */
      strcpy (Line, "/ht_root/local/* r+w");
      AuthPathLine (acptr, Line);
   }

   /************************/
   /* read the config file */
   /************************/

   AuthFileRab.rab$l_ubf = Line;
   AuthFileRab.rab$w_usz = sizeof(Line)-1;

   while (VMSok (status = sys$get (&AuthFileRab, 0, 0)))
   {
      acptr->LineNumber++;
      AuthFileRab.rab$l_ubf[AuthFileRab.rab$w_rsz] = '\0';
      memcpy (acptr->LinePtr = LineBuffer, Line, AuthFileRab.rab$w_rsz+1);
      if (Debug)
         fprintf (stdout, "line %d |%s|\n",
                  acptr->LineNumber, AuthFileRab.rab$l_ubf);

      if (AuthFileRab.rab$w_rsz)
      {
         if (AuthFileRab.rab$l_ubf[AuthFileRab.rab$w_rsz-1] == '\\')
         {
            /* directive is continued on next line */
            AuthFileRab.rab$l_ubf[AuthFileRab.rab$w_rsz-1] = ' ';
            AuthFileRab.rab$l_ubf += AuthFileRab.rab$w_rsz;
            AuthFileRab.rab$w_usz -= AuthFileRab.rab$w_rsz;
            continue;
         }
      }

      AuthPathLine (acptr, Line);

      AuthFileRab.rab$l_ubf = Line;
      AuthFileRab.rab$w_usz = sizeof(Line)-1;
   }

   if (status == RMS$_EOF) status = SS$_NORMAL;

   /******************/
   /* close the file */
   /******************/

   sys$close (&AuthFileFab, 0, 0); 

   if (VMSnok (status))
      ErrorExitVmsStatus (status, "reading HTTPD$AUTH file", FI_LI);

   if (Debug)
   {
      register struct AuthPathRecordStruct  *plptr;

      fprintf (stdout, "-\n");
      for (plptr = acptr->PathListHead;
           plptr != NULL;
           plptr = plptr->NextPtr)
         fprintf (stdout, "plptr |%s|%s|%s|%s|\n",
                  plptr->PathPtr, plptr->RealmPtr,
                  plptr->GroupWritePtr, plptr->GroupReadPtr);
      fprintf (stdout, "-\n");
   }

   /* let this function do its initialization */
   AuthPathCheck (NULL, NULL);

#ifdef DBUG
   Debug = DebugBuffer;
#endif
   return (status);
}

/*****************************************************************************/
/*
Process a line of authorization configuration.
*/

#define ACCESS_RESTRICTION_LIST_SIZE 256

AuthPathLine
(
struct AuthConfigStruct *acptr,
char *Line
)
{
   static boolean  VirtualServerUnknown;
   static int  SourceGroupRead,
               SourceGroupWrite,
               SourceRealm;
   static char  GroupRead [AUTH_MAX_REALM_GROUP_LENGTH+1],
                GroupWrite [AUTH_MAX_REALM_GROUP_LENGTH+1],
                Realm [AUTH_MAX_REALM_GROUP_LENGTH+1],
                RealmDescription [AUTH_MAX_REALM_GROUP_LENGTH+1],
                RealmCanString [256],
                VirtualServerHostPort [256];

   register char  *lptr, *sptr, *zptr;

   int  status;
   unsigned long  *CanFlagsPtr;
   unsigned long  GroupCanFlags,
                  GroupReadVmsIdentifier,
                  GroupWriteVmsIdentifier,
                  RealmCanFlags,
                  RealmVmsIdentifier,
                  WorldCanFlags;
   char  *PathTemplatePtr,
         *RestrictListPtr,
         *VirtualServerHostPortPtr;
   char  GroupRestrictList [ACCESS_RESTRICTION_LIST_SIZE],
         WorldRestrictList [ACCESS_RESTRICTION_LIST_SIZE];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "AuthPathLine() |%s|%s|%s|%s|%s|\n",
               Realm, RealmDescription, GroupRead, GroupWrite, Line);

   if (Line == NULL)
   {
      /* initialize */
      strcpy (Realm, AUTH_REALM_FAIL);
      VirtualServerUnknown = false;
      strcpy (VirtualServerHostPort, "*");
      return;
   }

   if (strsame (Realm, AUTH_REALM_FAIL, -1))
   {
      /* for the "always failed realm" make sure these are always reset */
      GroupReadVmsIdentifier = GroupWriteVmsIdentifier =
         RealmVmsIdentifier = GroupCanFlags = RealmCanFlags =
         SourceGroupRead = SourceGroupWrite = 0;
      GroupRead[0] = GroupWrite[0] = RealmCanString[0] = '\0';
      SourceRealm = AUTH_SOURCE_FAIL;
   }

   lptr = Line;
   while (ISLWS(*lptr)) lptr++;
   if (!*lptr || *lptr == '#') return;

   GroupCanFlags = WorldCanFlags = 0;
   GroupRestrictList[0] = WorldRestrictList[0] = '\0';
   CanFlagsPtr = &GroupCanFlags;
   RestrictListPtr = &GroupRestrictList;

   if (*lptr == '/')
   {
      /********/
      /* path */
      /********/

      /* ignore all paths associated with an unknown virtual server */
      if (VirtualServerUnknown) return;

      /* unless in promiscuous mode completely ignore PROMISCUOUS realm! */
      if (!AuthPromiscuous && strsame (Realm, AUTH_REALM_PROMISCUOUS, -1))
         return;

      /* note start of path template */
      PathTemplatePtr = lptr;
      while (*lptr && !ISLWS(*lptr)) lptr++;
      /* terminate at the end of the path template */
      if (*lptr) *lptr++ = '\0';

      if (SourceRealm == AUTH_SOURCE_NONE)
         while (*lptr) lptr++;
      else
      { 
         /* find start of access flags */
         while (*lptr && ISLWS(*lptr)) lptr++;
         if (!*lptr)
         {
            if (RealmCanString[0])
               lptr = RealmCanString;
            else
            {
               AuthConfigReport (acptr,
"%HTTPD-W-AUTH, no access specified \
(no realm defaults to apply) at line !UL\n",
                  acptr->LineNumber);
               lptr = "";
            }
         }
      }

      while (*lptr)
      {
         /********************************/
         /* set flags controlling access */
         /********************************/

         /* find the start of the next element */
         while (*lptr && (ISLWS(*lptr) || *lptr == ',' || *lptr == ';'))
         {
            if (*lptr++ != ';') continue;
            /* semicolon separates realm/group and optional world access */
            CanFlagsPtr = &WorldCanFlags;
            RestrictListPtr = &WorldRestrictList;
         }
         if (!*lptr) break;

         if (*lptr == '*' ||
             *lptr == '#' ||
             isdigit(*lptr) ||
             *lptr == '~' ||
             strsame (lptr, "http:", 5) ||
             strsame (lptr, "https:", 6) ||
             strsame (lptr, "localhost", 9))
         {
            /* access restriction list */
            for (sptr = RestrictListPtr; *sptr; sptr++);
            zptr = RestrictListPtr + ACCESS_RESTRICTION_LIST_SIZE;
            if (sptr > RestrictListPtr && sptr < zptr) *sptr++ = ',';
            while (*lptr && !ISLWS(*lptr) &&
                   *lptr != ',' && *lptr != ';' &&
                   sptr < zptr)
               *sptr++ = *lptr++;
            if (sptr >= zptr)
            {
               AuthConfigReport (acptr,
                  "%HTTPD-W-AUTH, restriction list too long at line !UL\n",
                  acptr->LineNumber);
               strcpy (RestrictListPtr, AUTH_RESTRICTED_FAIL);
               break;
            }
            else
               *sptr = '\0';
         }
         else
         if (strsame (lptr, "none", 4) && !isalpha(lptr[4]))
            *CanFlagsPtr = 0;
         else
         /* must come before "read" for obvious reasons! */
         if ((strsame (lptr, "READ+WRITE", 10) && !isalpha(lptr[10])) ||
             (strsame (lptr, "R+W", 3) && !isalpha(lptr[3])))
            *CanFlagsPtr |= (HTTP_METHOD_GET | HTTP_METHOD_HEAD |
                             HTTP_METHOD_DELETE | HTTP_METHOD_POST |
                             HTTP_METHOD_PUT);
         else
         if ((strsame (lptr, "READ", 4) && !isalpha(lptr[4])) ||
             (strsame (lptr, "R", 1) && !isalpha(lptr[1])))
            *CanFlagsPtr |= (HTTP_METHOD_GET | HTTP_METHOD_HEAD);
         else
         if ((strsame (lptr, "WRITE", 5) && !isalpha(lptr[5])) ||
             (strsame (lptr, "W", 1) && !isalpha(lptr[1])))
            *CanFlagsPtr |= (HTTP_METHOD_DELETE | HTTP_METHOD_POST |
                             HTTP_METHOD_PUT);
         else
         if (strsame (lptr, "DELETE", 6) && !isalpha(lptr[6]))
            *CanFlagsPtr |= HTTP_METHOD_DELETE;
         else
         if (strsame (lptr, "GET", 3) && !isalpha(lptr[3]))
            *CanFlagsPtr |= HTTP_METHOD_GET | HTTP_METHOD_HEAD;
         else
         if (strsame (lptr, "HEAD", 4) && !isalpha(lptr[4]))
            *CanFlagsPtr |= HTTP_METHOD_HEAD;
         else
         if (strsame (lptr, "POST", 4) && !isalpha(lptr[4]))
            *CanFlagsPtr |= HTTP_METHOD_POST;
         else
         if (strsame (lptr, "PUT", 3) && !isalpha(lptr[3]))
            *CanFlagsPtr |= HTTP_METHOD_PUT;
         else
         {
            /* otherwise assume it's an alpha-numeric host name */
            for (sptr = RestrictListPtr; *sptr; sptr++);
            zptr = RestrictListPtr + ACCESS_RESTRICTION_LIST_SIZE;
            if (sptr > RestrictListPtr && sptr < zptr) *sptr++ = ',';
            while (*lptr && !ISLWS(*lptr) &&
                   *lptr != ',' && *lptr != ';' &&
                   sptr < zptr)
               *sptr++ = *lptr++;
            if (sptr >= zptr)
            {
               AuthConfigReport (acptr,
                 "%HTTPD-W-AUTH, restriction list too long at line !UL\n",
                  acptr->LineNumber);
               strcpy (RestrictListPtr, AUTH_RESTRICTED_FAIL);
               break;
            }
            else
               *sptr = '\0';
         }
         while (*lptr && !ISLWS(*lptr) && *lptr != ',' && *lptr != ';') lptr++;
      }

      if (VMSnok (status =
          AuthPathAdd (acptr,
                       VirtualServerHostPort,
                       Realm, RealmDescription, SourceRealm, RealmVmsIdentifier,
                       GroupWrite, SourceGroupWrite, GroupWriteVmsIdentifier,
                       GroupRead, SourceGroupRead, GroupReadVmsIdentifier,
                       PathTemplatePtr,
                       GroupRestrictList,
                       WorldRestrictList,
                       GroupCanFlags, WorldCanFlags)))
      if (status == RMS$_DUP)
         AuthConfigReport (acptr,
"%HTTPD-W-AUTH, duplicate path (ignored) at line !UL\n",
            acptr->LineNumber);
      else
         ErrorExitVmsStatus (status, "AuthPathAdd()", FI_LI);

      /* increment this boolean if authorization paths are loaded */
      AuthorizationEnabled++;

      return;
   }

   if (*(unsigned short*)lptr == '[[')
   {
      /*******************/
      /* virtual service */
      /*******************/

      VirtualServerHostPortPtr = lptr += 2;
      for (sptr = lptr; *sptr && *sptr != ']'; sptr++);
      if (*(unsigned short*)sptr != ']]')
      {
         AuthConfigReport (acptr,
"%HTTPD-W-AUTH, virtual service problem at line !UL\n",
            acptr->LineNumber);
         VirtualServerUnknown = true;
         return;
      }
      *sptr = '\0';

      if (*VirtualServerHostPortPtr != '*')
      {
         if (!NetServiceConfigured (VirtualServerHostPortPtr))
         {
            AuthConfigReport (acptr,
"%HTTPD-W-AUTH, virtual service not configured at line !UL\n",
               acptr->LineNumber);
            VirtualServerUnknown = true;
            return;
         }
      }
      strcpy (VirtualServerHostPort, VirtualServerHostPortPtr);
      VirtualServerUnknown = false;
      return;
   }

   if (*lptr == '[' || strsame (lptr, "REALM", 5))
   {
      /*********/
      /* realm */
      /*********/

      /* ignore all realms associated with an unknown virtual server */
      if (VirtualServerUnknown) return;

      if (*lptr == '[')
      {
         lptr++;
         while (*lptr && ISLWS(*lptr)) lptr++;
      }
      else
      {
         /* skip over keyword and find start of realm name */
         while (*lptr && !ISLWS(*lptr)) lptr++;
         while (*lptr && ISLWS(*lptr)) lptr++;
      }

      GroupReadVmsIdentifier = GroupWriteVmsIdentifier =
         RealmVmsIdentifier = GroupCanFlags = RealmCanFlags =
         SourceGroupRead = SourceGroupWrite = SourceRealm = 0;
      GroupRead[0] = GroupWrite[0] =
         Realm[0] = RealmDescription[0] = RealmCanString[0] = '\0';

      if (*lptr == '\"')
      {
         lptr++;
         zptr = (sptr = RealmDescription) + sizeof(RealmDescription);
         while (*lptr &&
                *lptr != '\"' &&
                sptr < zptr)
            *sptr++ = *lptr++;
         if (sptr >= zptr)
         {
            AuthConfigReport (acptr,
               "%HTTPD-W-AUTH, confused at line !UL\n",
               acptr->LineNumber);
            strcpy (Realm, AUTH_REALM_FAIL);
            return;
         }
         else
            *sptr = '\0';
         if (*lptr == '\"') lptr++;
         if (*lptr == '=') lptr++;
      }

      SourceRealm = AUTH_SOURCE_HTA;
      zptr = (sptr = Realm) + sizeof(Realm);
      while (*lptr &&
             *lptr != ';' &&
             *lptr != '=' &&
             *lptr != ']' &&
             !ISLWS(*lptr) &&
             sptr < zptr)
         *sptr++ = toupper(*lptr++);
      if (sptr >= zptr)
      {
         AuthConfigReport (acptr,
            "%HTTPD-W-AUTH, confused at line !UL\n",
            acptr->LineNumber);
         strcpy (Realm, AUTH_REALM_FAIL);
         return;
      }
      else
      {
         *sptr = '\0';
         if (!Realm[0])
         {
            AuthConfigReport (acptr,
               "%HTTPD-W-AUTH, realm not specified at line !UL\n",
               acptr->LineNumber);
            strcpy (Realm, AUTH_REALM_FAIL);
            return;
         }
      }

      if (*lptr == '=')
      {
         if (strsame (lptr, "=HTA", 4))
         {
            lptr += 4;
            SourceRealm = AUTH_SOURCE_HTA;
         }
         else
         if (strsame (lptr, "=ID", 3))
         {
            lptr += 3;
            SourceRealm = AUTH_SOURCE_ID;
            if (VMSnok (status =
                AuthConfigIdentifier (Realm, &RealmVmsIdentifier)))
            {
               AuthConfigReport (&AuthServerConfig,
"%HTTPD-W-AUTH, realm !AZ identifier at line !UL\n-!AZ\n",
                  Realm, acptr->LineNumber, SysGetMsg(status)+1);
               strcpy (Realm, AUTH_REALM_FAIL);
               return;
            }
         }
         else
         if (strsame (lptr, "=LIST", 5))
         {
            lptr += 5;
            SourceRealm = AUTH_SOURCE_LIST;
         }
         else
         if (strsame (lptr, "=VMS", 4))
         {
            lptr += 4;
            SourceRealm = AUTH_SOURCE_VMS;
         }
         else
         {
            AuthConfigReport (acptr,
               "%HTTPD-W-AUTH, confused at line !UL\n",
               acptr->LineNumber);
            strcpy (Realm, AUTH_REALM_FAIL);
            return;
         }
      }

      /* if was "VMS" then it's SYSUAF no matter what was made equal to!! */
      if (strsame (Realm, AUTH_REALM_VMS, -1))
         SourceRealm = AUTH_SOURCE_VMS;
      else
      /* same for "WORLD" */
      if (strsame (Realm, AUTH_REALM_WORLD, -1))
         SourceRealm = AUTH_SOURCE_WORLD;
      else
      /* and for "EXTERNAL" */
      if (strsame (Realm, AUTH_REALM_EXTERNAL, -1))
         SourceRealm = AUTH_SOURCE_EXTERNAL;
      else
      /* and "NONE" */
      if (strsame (Realm, AUTH_REALM_NONE, -1))
         SourceRealm = AUTH_SOURCE_NONE;
      else
      /* and last but by no means least "PROMISCUOUS" */
      if (strsame (Realm, AUTH_REALM_PROMISCUOUS, -1))
         SourceRealm = AUTH_SOURCE_PROMISCUOUS;

      if (AuthPolicySysUafWasdIdentifiers)
      {
         /*
            If this is the VMS realm and identifiers are mandatory for
            SYSUAF authentication but the realm is not a specific identifier
            then it must be through the "hard-wired" WASD identifiers.
         */
         if (SourceRealm == AUTH_SOURCE_VMS && AuthPolicySysUafIdentifiers)
            SourceRealm = AUTH_SOURCE_WASD_ID;
      }

      while (*lptr && ISLWS(*lptr)) lptr++;

      /* semicolon separating realm from optional full-access (write) group */
      if (*lptr == ';')
      {
         SourceGroupWrite = AUTH_SOURCE_HTA;

         lptr++;
         while (*lptr && ISLWS(*lptr)) lptr++;
         zptr = (sptr = GroupWrite) + sizeof(GroupWrite);
         while (*lptr &&
                *lptr != ';' &&
                *lptr != '=' &&
                *lptr != ']' &&
                !ISLWS(*lptr) &&
                sptr < zptr)
            *sptr++ = toupper(*lptr++);
         if (sptr >= zptr)
         {
            AuthConfigReport (acptr,
               "%HTTPD-W-AUTH, confused at line !UL\n",
               acptr->LineNumber);
            strcpy (Realm, AUTH_REALM_FAIL);
            return;
         }
         *sptr = '\0';

         if (*lptr == '=')
         {
            if (strsame (lptr, "=HTA", 4))
            {
               lptr += 4;
               SourceGroupWrite = AUTH_SOURCE_HTA;
            }
            else
            if (strsame (lptr, "=ID", 3))
            {
               lptr += 3;
               SourceGroupWrite = AUTH_SOURCE_ID;
               if (VMSnok (status =
                   AuthConfigIdentifier (GroupWrite, &GroupWriteVmsIdentifier)))
               {
                  AuthConfigReport (&AuthServerConfig,
"%HTTPD-W-AUTH, group !AZ identifier at line !UL\n-!AZ\n",
                     GroupWrite, acptr->LineNumber, SysGetMsg(status)+1);
                  strcpy (Realm, AUTH_REALM_FAIL);
                  return;
               }
            }
            else
            if (strsame (lptr, "=LIST", 5))
            {
               lptr += 5;
               SourceGroupWrite = AUTH_SOURCE_LIST;
            }
            else
            {
               AuthConfigReport (acptr,
                  "%HTTPD-W-AUTH, confused at line !UL\n",
                  acptr->LineNumber);
               strcpy (Realm, AUTH_REALM_FAIL);
               return;
            }
         }
      }

      /* semicolon separating realm from optional read-only-access group */
      if (*lptr == ';')
      {
         SourceGroupRead = AUTH_SOURCE_HTA;

         lptr++;
         while (*lptr && ISLWS(*lptr)) lptr++;
         zptr = (sptr = GroupRead) + sizeof(GroupRead);
         while (*lptr &&
                *lptr != '=' &&
                *lptr != ']' &&
                !ISLWS(*lptr) &&
                sptr < zptr)
            *sptr++ = toupper(*lptr++);
         if (sptr >= zptr)
         {
            AuthConfigReport (acptr,
               "%HTTPD-W-AUTH, confused at line !UL\n",
               acptr->LineNumber);
            strcpy (Realm, AUTH_REALM_FAIL);
            return;
         }
         *sptr = '\0';

         if (*lptr == '=')
         {
            if (strsame (lptr, "=HTA", 4))
            {
               lptr += 4;
               SourceGroupRead = AUTH_SOURCE_HTA;
            }
            if (strsame (lptr, "=ID", 3))
            {
               lptr += 3;
               SourceGroupRead = AUTH_SOURCE_ID;
               if (VMSnok (status =
                   AuthConfigIdentifier (GroupRead, &GroupReadVmsIdentifier)))
               {
                  AuthConfigReport (&AuthServerConfig,
"%HTTPD-W-AUTH, group !AZ identifier at line !UL\n-!AZ\n",
                     GroupWrite, acptr->LineNumber, SysGetMsg(status)+1);
                  strcpy (Realm, AUTH_REALM_FAIL);
                  return;
               }
            }
            else
            if (strsame (lptr, "=LIST", 5))
            {
               lptr += 5;
               SourceGroupRead = AUTH_SOURCE_LIST;
            }
            else
            {
               AuthConfigReport (acptr,
                  "%HTTPD-W-AUTH, confused at line !UL\n",
                  acptr->LineNumber);
               strcpy (Realm, AUTH_REALM_FAIL);
               return;
            }
         }
      }

      while (*lptr && ISLWS(*lptr)) lptr++;
      if (*lptr == ']')
         lptr++;
      else
      {
         AuthConfigReport (acptr,
            "%HTTPD-W-AUTH, confused at line !UL\n",
            acptr->LineNumber);
         strcpy (Realm, AUTH_REALM_FAIL);
         return;
      }

      /* find start of any realm-context access flags */
      while (*lptr && ISLWS(*lptr)) lptr++;
      if (*lptr)
         strcpy (RealmCanString, lptr);
      else
         RealmCanString[0] = '\0';

      /***************************/
      /* bit more usage checking */
      /***************************/

      if ((SourceGroupWrite == AUTH_SOURCE_ID ||
           SourceGroupRead == AUTH_SOURCE_ID) &&
          SourceRealm != AUTH_SOURCE_VMS &&
          SourceRealm != AUTH_SOURCE_ID)
      {
         AuthConfigReport (acptr,
"%HTTPD-W-AUTH, identifier used without SYSUAF authentication at line !UL\n",
            acptr->LineNumber);
         strcpy (Realm, AUTH_REALM_FAIL);
         return;
      }

      if (AuthPolicySysUafIdentifiers &&
          SourceRealm == AUTH_SOURCE_VMS)
      {
         AuthConfigReport (acptr,
"%HTTPD-W-AUTH, realm indicates VMS but /SYSUAF=ID active at line !UL\n",
            acptr->LineNumber);
         strcpy (Realm, AUTH_REALM_FAIL);
         return;
      }

      if (AuthPolicySysUafWasdIdentifiers &&
          SourceRealm == AUTH_SOURCE_WASD_ID &&
          SourceGroupWrite != AUTH_SOURCE_WASD_ID &&
          (GroupWrite[0] || GroupRead[0]))
      {
         AuthConfigReport (acptr,
"%HTTPD-W-AUTH, \"hard-wired\" WASD identifier usage problem at line !UL\n",
            acptr->LineNumber);
         strcpy (Realm, AUTH_REALM_FAIL);
         return;
      }

      return;
   }

   AuthConfigReport (acptr, "%HTTPD-W-AUTH, confused at line !UL\n",
                     acptr->LineNumber);
}

/****************************************************************************/
/*
VMS sys$fao()-formatted print statement. This function formats an error report. 
All lines are concatenated onto a single string of dynamically allocated memory
that (obviously) grows as reports are added to it.  This string is then output
if loading the server configuration or is available for inclusion in an HTML
page.
*/

AuthConfigReport
(
struct AuthConfigStruct *acptr,
char *FormatString,
...
)
{
   int  status,
        argcnt;
   unsigned short  Length;
   unsigned long  *vecptr;
   unsigned long  FaoVector [32];
   char  Buffer [512],
         HtmlBuffer [512];
   $DESCRIPTOR (BufferDsc, Buffer);
   $DESCRIPTOR (FormatFaoDsc, FormatString);
   va_list  argptr;

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug)
      fprintf (stdout, "AuthConfigReport() |%s| %d\n", FormatString, argcnt);

   vecptr = FaoVector;
   va_start (argptr, FormatString);
   for (argcnt -= 2; argcnt && argcnt < 32; argcnt--)
      *vecptr++ = va_arg (argptr, unsigned long);
   va_end (argptr);
   if (argcnt >= 32)
      ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);

   FormatFaoDsc.dsc$w_length = strlen(FormatString);

   status = sys$faol (&FormatFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (Debug) fprintf (stdout, "sys$faol() %%X%08.08X\n", status);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
      ErrorExitVmsStatus (status, "sys$faol()", FI_LI);
   Buffer[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", Buffer);

   if (acptr->RequestPtr == NULL) fputs (Buffer, stdout);

   if ((Length = CopyToHtml (HtmlBuffer, sizeof(HtmlBuffer), Buffer, -1)) < 0)
      ErrorExitVmsStatus (SS$_BUFFEROVF, "CopyToHtml()", FI_LI);

   acptr->ProblemReportPtr =
      VmRealloc (acptr->ProblemReportPtr, acptr->ProblemReportLength+Length+1);

   /* include the terminating null */
   memcpy (acptr->ProblemReportPtr+acptr->ProblemReportLength,
           HtmlBuffer, Length+1);
   acptr->ProblemReportLength += Length;

   acptr->ProblemCount++;
}

/****************************************************************************/
/*
Get the identifier value from the supplied name.
*/ 

AuthConfigIdentifier
(
char *NamePtr,
unsigned long  *ValuePtr
)
{
   static $DESCRIPTOR (NameDsc, "");

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthConfigIdentifier() |%s|\n", NamePtr);

   *ValuePtr = 0;
   NameDsc.dsc$a_pointer = NamePtr;
   NameDsc.dsc$w_length = strlen(NamePtr);
   status = sys$asctoid (&NameDsc, ValuePtr, 0);
   if (Debug)
      fprintf (stdout, "sys$asctoid() %%X%08.08X %08.08X\n", status, *ValuePtr);
   return (status);
}

/*****************************************************************************/
/*
Scan through the path list until the first record GREATER than the new path is
encountered.  Insert the new path immediately BEFORE that record.  The data
structure has strings stored immediately following the formally declared
structure and pointed to by character pointers within the formal structure.
*/ 

int AuthPathAdd
(
struct AuthConfigStruct *acptr,
char *VirtualServerHostPort,
char *Realm,
char *RealmDescription,
unsigned long SourceRealm,
unsigned long RealmVmsIdentifier,
char *GroupWrite,
unsigned long SourceGroupWrite,
unsigned long GroupWriteVmsIdentifier,
char *GroupRead,
unsigned long SourceGroupRead,
unsigned long GroupReadVmsIdentifier,
char *Path,
char *GroupRestrictList,
char *WorldRestrictList,
unsigned long  GroupCanFlags,
unsigned long  WorldCanFlags
)
{
   register char  *cptr, *pptr, *sptr, *tptr;
   register struct AuthPathRecordStruct  *plptr, *nlptr;

   int  GroupReadLength,
        GroupWriteLength,
        GroupRestrictListLength,
        PathLength,
        RealmLength,
        RealmDescriptionLength,
        StringSpace,
        VirtualServerHostPortLength,
        WorldRestrictListLength;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
"AuthPathAdd() |%s|\n\
|%s|%s|%d|%08.08X|\n\
|%s|%d|%08.08X|\n\
|%s|%d|%08.08X|\n\
|%s|\n\
|%s|%s|%08.08X|%08.08X|\n",
         VirtualServerHostPort,
         Realm, RealmDescription, SourceRealm, RealmVmsIdentifier,
         GroupWrite, SourceGroupWrite, GroupWriteVmsIdentifier,
         GroupRead, SourceGroupRead, GroupReadVmsIdentifier,
         Path, GroupRestrictList,
         WorldRestrictList, GroupCanFlags, WorldCanFlags);

   if (acptr == NULL) acptr = &AuthServerConfig;

   GroupReadLength = strlen(GroupRead);
   GroupWriteLength = strlen(GroupWrite);
   PathLength = strlen(Path);
   RealmLength = strlen(Realm);
   RealmDescriptionLength = strlen(RealmDescription);
   GroupRestrictListLength = strlen(GroupRestrictList);
   VirtualServerHostPortLength = strlen(VirtualServerHostPort);
   WorldRestrictListLength = strlen(WorldRestrictList);
   cptr = Path;

   for (plptr = acptr->PathListHead; plptr != NULL; plptr = plptr->NextPtr)
   {
/*
      if (Debug)
         fprintf (stdout, "plptr |%s|%s|%08.08X|\n",
                  plptr->PathPtr, plptr->RealmPtr, plptr->AuthGroupCan);
*/

      pptr = cptr;
      tptr = plptr->PathPtr;
      while (*tptr && *pptr && tolower(*tptr) == tolower(*pptr))
      {
         tptr++;
         pptr++;
      }

      if (!*tptr && !*pptr)
      {
         /* paths are the same, same virtual server? */
         if (*plptr->VirtualServerHostPortPtr != '*')
         {
            /* rule is for a specific virtual server */
            if (strsame (VirtualServerHostPort,
                         plptr->VirtualServerHostPortPtr,
                         -1))
            {
               /* server name and port match, duplicate record! */
               return (RMS$_DUP);
            }
            /* otherwise, just drop through to continue processing */
         }
         else
         {
            /* no specific virtual server, duplicate record! */
            return (RMS$_DUP);
         }
      }

      /* got to end of template and new path longer (sub-path), goes first */
      if (*tptr == '*' && PathLength > plptr->PathLength) break;

      /* got to end of new path, must be a "parent"-path */
      if (*pptr == '*') continue;

      /* if not at end of template and new path is less then add now */
      if (*tptr && (tolower(*tptr) >= tolower(*pptr))) break;
   }

   StringSpace = VirtualServerHostPortLength + 1 +
                 GroupReadLength + 1 +
                 GroupWriteLength + 1 +
                 RealmLength + 1 +
                 RealmDescriptionLength + 1 +
                 PathLength + 1 +
                 GroupRestrictListLength + 1 +
                 WorldRestrictListLength + 1;

   nlptr = VmGet (sizeof(struct AuthPathRecordStruct) + StringSpace);
   StringSpace = sizeof(struct AuthPathRecordStruct);

   cptr = VirtualServerHostPort;
   nlptr->VirtualServerHostPortPtr = sptr = (char*)nlptr + StringSpace;
   while (*cptr) *sptr++ = tolower(*cptr++);      
   *sptr = '\0';
   nlptr->VirtualServerHostPortLength = sptr - nlptr->VirtualServerHostPortPtr;
   StringSpace += VirtualServerHostPortLength + 1;
   nlptr->VirtualServerHostNameHashValue =
      NetHostNameHash (nlptr->VirtualServerHostPortPtr);

   cptr = Path;
   nlptr->PathPtr = sptr = (char*)nlptr + StringSpace;
   while (*cptr) *sptr++ = tolower(*cptr++);      
   *sptr = '\0';
   StringSpace += (nlptr->PathLength = PathLength) + 1;

   cptr = Realm;
   nlptr->RealmPtr = sptr = (char*)nlptr + StringSpace;
   while (*cptr) *sptr++ = *cptr++;
   *sptr = '\0';
   StringSpace += (nlptr->RealmLength = RealmLength) + 1;
   nlptr->SourceRealm = SourceRealm;
   nlptr->RealmVmsIdentifier = RealmVmsIdentifier;

   cptr = RealmDescription;
   nlptr->RealmDescriptionPtr = sptr = (char*)nlptr + StringSpace;
   while (*cptr) *sptr++ = *cptr++;
   *sptr = '\0';
   StringSpace += (nlptr->RealmDescriptionLength = RealmDescriptionLength) + 1;

   cptr = GroupWrite;
   nlptr->GroupWritePtr = sptr = (char*)nlptr + StringSpace;
   while (*cptr) *sptr++ = *cptr++;
   *sptr = '\0';
   StringSpace += (nlptr->GroupWriteLength = GroupWriteLength) + 1;
   nlptr->SourceGroupWrite = SourceGroupWrite;
   nlptr->GroupWriteVmsIdentifier = GroupWriteVmsIdentifier;

   cptr = GroupRead;
   nlptr->GroupReadPtr = sptr = (char*)nlptr + StringSpace;
   while (*cptr) *sptr++ = *cptr++;
   *sptr = '\0';
   StringSpace += (nlptr->GroupReadLength = GroupReadLength) + 1;
   nlptr->SourceGroupRead = SourceGroupRead;
   nlptr->GroupReadVmsIdentifier = GroupReadVmsIdentifier;

   cptr = GroupRestrictList;
   nlptr->GroupRestrictListPtr = sptr = (char*)nlptr + StringSpace;
   while (*cptr) *sptr++ = tolower(*cptr++);      
   *sptr = '\0';
   StringSpace += (nlptr->GroupRestrictListLength =
                  GroupRestrictListLength) + 1;

   cptr = WorldRestrictList;
   nlptr->WorldRestrictListPtr = sptr = (char*)nlptr + StringSpace;
   while (*cptr) *sptr++ = tolower(*cptr++);      
   *sptr = '\0';

   nlptr->AuthGroupCan = GroupCanFlags;
   nlptr->AuthWorldCan = WorldCanFlags;

   if (plptr == NULL)
   {
      /* no entry greater than the new one */
      if (acptr->PathListHead == NULL)
      {
         /* empty list */
         acptr->PathListHead = acptr->PathListTail = nlptr;
         nlptr->PrevPtr = nlptr->NextPtr = NULL;
      }
      else
      {
         /* add to end of list */
         nlptr->PrevPtr = acptr->PathListTail;
         acptr->PathListTail->NextPtr = nlptr;
         nlptr->NextPtr = NULL;
         acptr->PathListTail = nlptr;
      }
   }
   else
   {
      /* insert before the entry found */
      nlptr->PrevPtr = plptr->PrevPtr;
      if (nlptr->PrevPtr != NULL) nlptr->PrevPtr->NextPtr = nlptr;
      nlptr->NextPtr = plptr;
      plptr->PrevPtr = nlptr;
      if (acptr->PathListHead == plptr)
          acptr->PathListHead = nlptr;
   }

   if (Debug)
      fprintf (stdout, "nlptr |%s|%s| group: %08.08X world: %08.08X\n",
               nlptr->PathPtr, nlptr->RealmPtr,
               nlptr->AuthGroupCan, nlptr->AuthWorldCan);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Remove the entry for the specified path from the path authorization list.
Return true if found and removed, false if not.
*/

boolean AuthPathRemove
(
struct AuthConfigStruct *acptr,
char *Path
)
{
   register int  EntryCount;
   register char  *pptr, *tptr;
   register struct AuthPathRecordStruct  *plptr, *nlptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthPathRemove() |%s|\n", Path);

   if (acptr == NULL) acptr = &AuthServerConfig;
   if (acptr->PathListHead == NULL) return (false);

   for (plptr = acptr->PathListHead; plptr != NULL; plptr = plptr->NextPtr)
   {
      if (Debug) fprintf (stdout, "plptr |%s|\n", plptr->PathPtr);

      pptr = Path;
      tptr = plptr->PathPtr;
      while (*tptr && *pptr && tolower(*tptr) == tolower(*pptr))
      {
         tptr++;
         pptr++;
      }

      /* try next entry in list if not the same */
      if (*tptr && *pptr) continue;

      /* found entry for path */
      if (acptr->PathListHead == plptr && acptr->PathListTail == plptr)
      {
         /* only entry in list */
         acptr->PathListHead = acptr->PathListTail = NULL;
      }
      else
      if (acptr->PathListHead == plptr)
      {
         /* first entry in list */
         acptr->PathListHead = plptr->NextPtr;
         plptr->NextPtr->PrevPtr = NULL;
      }
      else
      if (acptr->PathListTail == plptr)
      {
         /* last entry in list */
         acptr->PathListTail = plptr->PrevPtr;
         plptr->PrevPtr->NextPtr = NULL;
      }
      else
      {
         /* somewhere in the middle! */
         plptr->NextPtr->PrevPtr = plptr->PrevPtr;
         plptr->PrevPtr->NextPtr = plptr->NextPtr;
      }

      VmFree (plptr);
      return (true);
   }

   return (false);
}

/*****************************************************************************/
/*
Frees all dynamic structures (memory) in the authorization path list.
*/

AuthPathListFree (struct AuthConfigStruct *acptr)

{
   register int  EntryCount;
   register struct AuthPathRecordStruct  *plptr, *nlptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthPathListFree()\n");

   if (acptr == NULL) acptr = &AuthServerConfig;

   if (acptr->PathListHead == NULL) return;

   for (plptr = acptr->PathListHead; plptr != NULL; plptr = nlptr)
   {
      if (Debug)
         fprintf (stdout, "plptr |%s|%s|%s|%s|\n",
                  plptr->PathPtr, plptr->RealmPtr,
                  plptr->GroupWritePtr, plptr->GroupReadPtr);

      nlptr = plptr->NextPtr;
      VmFree (plptr);
   }

   acptr->PathListHead = acptr->PathListTail = NULL;
}

/*****************************************************************************/
/*
Look for a path template matching the request path in the path authorization
database.  If one can't be found return with '->AuthRealmPtr',
'rqptr->AuthGroupRestrictListPtr' and 'rqptr->AuthWorldRestrictListPtr'
pointing to empty strings and the path capabilities '->AuthGroupCan' and
'->AuthWorldCan' set to zero.  If found return with the realm name and path
capabilities set to what's in the database.  The 'QuickIndex' attempts to
reduce search times in larger authorization databases by allowing the search to
begin in the vicinity of a likely hit.  This function must be called with a
parameter of NULL to initialize the quick index after the paths are loaded.
*/ 

int AuthPathCheck
(
struct RequestStruct *rqptr,
struct AuthConfigStruct *acptr
)
{
   static struct AuthPathRecordStruct  *QuickIndex [26];

   register char  c;
   register char  *cptr, *pptr, *tptr;
   register struct AuthPathRecordStruct  *plptr;

   boolean  WatchThis;
   int  TotalLength;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthPathCheck()\n");

   if (acptr == NULL) acptr = &AuthServerConfig;

   if (rqptr == NULL)
   {
      /***************************/
      /* function initialization */
      /***************************/

      /* note the first instance of a path beginning with 'A' thru to 'Z' */
      for (c = 0; c < 26; c++) QuickIndex[c] = NULL;
      for (plptr = acptr->PathListHead;
           plptr != NULL;
           plptr = plptr->NextPtr)
      {
         c = plptr->PathPtr[1];
         if (c < 'A' || c > 'Z') continue;
         c -= 'A';
         if (QuickIndex[c]) continue;
         QuickIndex[c] = plptr;
         if (Debug) fprintf (stdout, "%d |%s|\n", c, QuickIndex[c]->PathPtr);
      }
      return (SS$_NORMAL);
   }

   /**********************/
   /* function execution */
   /**********************/

   if (Debug) fprintf (stdout, "|%s|\n", rqptr->PathInfoPtr);

   if ((plptr = acptr->PathListHead) == NULL) return (STS$K_ERROR);

   /* no need to compare the leading slash */
   c = tolower(*(cptr = rqptr->PathInfoPtr + 1));

   /* if there is an entry for this character that start with that one */
   if (c >= 'A' && c <= 'Z')
   {
      c -= 'A';
      if (QuickIndex[c])
         plptr = QuickIndex[c];
      else
      {
         /* there can't be a controlled path beginning with that character */
         return (STS$K_ERROR);
      }
   }

   if (rqptr->AdminWatchItem && (AdminWatchEnabled & ADMIN_WATCH_AUTH))
      WatchThis = true;
   else
      WatchThis = false;

   if (Debug) fprintf (stdout, "|%s|\n", plptr->PathPtr);

   for (/* from above */; plptr != NULL; plptr = plptr->NextPtr)
   {
      if (Debug)
         fprintf (stdout,
"plptr |%s|\n\
|%s|\n\
|%s|%s|%d|%08.08X|\n\
|%s|%d|%08.08X|\n\
|%s|%d|%08.08X|\n\
|%s|%s|\n\
GroupCan: %08.08X WorldCan: %08.08X\n",
            plptr->VirtualServerHostPortPtr,
            plptr->PathPtr,
            plptr->RealmPtr, plptr->RealmDescriptionPtr,
            plptr->SourceRealm, plptr->RealmVmsIdentifier,
            plptr->GroupWritePtr, plptr->SourceGroupWrite,
            plptr->GroupWriteVmsIdentifier,
            plptr->GroupReadPtr, plptr->SourceGroupRead,
            plptr->GroupReadVmsIdentifier,
            plptr->GroupRestrictListPtr,
            plptr->WorldRestrictListPtr,
            plptr->AuthGroupCan, plptr->AuthWorldCan);

      if (WatchThis)
      {
         /****************/
         /* WATCH report */
         /****************/

         static $DESCRIPTOR (WatchFaoDsc, "!#AZ !AZ\n");
         static $DESCRIPTOR (BufferDsc, "");

         int  status;
         unsigned long  *vecptr;
         unsigned short  Length;
         unsigned long  FaoVector [16];
         char  Buffer [256];

         BufferDsc.dsc$a_pointer = Buffer;
         BufferDsc.dsc$w_length = sizeof(Buffer)-1;

         vecptr = FaoVector;
         if (plptr->PathLength > 31)
            *vecptr++ = plptr->PathLength;
         else
            *vecptr++ = 31;
         *vecptr++ = plptr->PathPtr;
         *vecptr++ = plptr->VirtualServerHostPortPtr;

         status = sys$faol (&WatchFaoDsc, &Length, &BufferDsc, &FaoVector);
         if (status == SS$_BUFFEROVF)
         {
            strcpy (Buffer, ErrorAuthSysFaoBufferOvf);
            Length = sizeof(ErrorAuthSysFaoBufferOvf)-1;
         }
         else
         if (VMSnok (status))
         {
            strcpy (Buffer, ErrorAuthSysFao);
            Length = sizeof(ErrorAuthSysFao)-1;
         }
         else
            Buffer[Length] = '\0';

         AdminWatchData (Buffer, Length);
      }

      pptr = cptr;
      tptr = plptr->PathPtr + 1;

      /* no need to continue on if past the lexicographic cutoff point */
      if (isalpha(*pptr) && isalpha(*tptr) && tolower(*pptr) < tolower(*tptr))
         return (STS$K_ERROR);

      while (*tptr && *pptr &&
             *tptr != '*' &&
             (tolower(*tptr) == tolower(*pptr) ||
             *pptr == '*' || *pptr == '%' || *pptr == '.'))
      {
         /** if (Debug) fprintf (stdout, "|%s|%s|\n", tptr, pptr); **/
         if (*pptr == '*' || *pptr == '%' || *pptr == '.')
            while (*pptr == '*' || *pptr == '%' || *pptr == '.') pptr++;
         else
         {
            tptr++;
            pptr++;
         }
      }

      if (*tptr == '*' || !(*tptr || *pptr))
      {
         /****************/
         /* path matched */
         /****************/

         /* path template ended in a wildcard or exact match */
         if (Debug)
            fprintf (stdout, "|%s|%s|\n",
                     plptr->VirtualServerHostPortPtr,
                     rqptr->ServicePtr->ServerHostPort);

         /* if the following rules apply to all services then break */
         if (*plptr->VirtualServerHostPortPtr == '*')
            break;

         /* if the request service matches the rule service then break */
         if (rqptr->ServicePtr->ServerHostNameHashValue ==
             plptr->VirtualServerHostNameHashValue &&
             NetThisVirtualService (rqptr->ServicePtr->ServerHostPort,
                                    plptr->VirtualServerHostPortPtr))
            break;
      }
   }

   if (plptr == NULL) return (STS$K_ERROR);

   AUTH_RESET_REQUEST

   /*
      As the path authorization information is potentially volatile
      (i.e. can be reloaded) we need request-local copies of these fields.
   */ 

   TotalLength = plptr->RealmLength +
                 plptr->RealmDescriptionLength +
                 plptr->GroupWriteLength +
                 plptr->GroupReadLength +
                 plptr->GroupRestrictListLength +
                 plptr->WorldRestrictListLength + 6;
   cptr = VmGetHeap (rqptr, TotalLength);

   memcpy (rqptr->AuthRealmPtr = cptr,
           plptr->RealmPtr, plptr->RealmLength+1);
   cptr += plptr->RealmLength+1;
   rqptr->AuthSourceRealm = plptr->SourceRealm;
   rqptr->AuthRealmVmsIdentifier = plptr->RealmVmsIdentifier;

   if (plptr->RealmDescriptionPtr[0])
   {
      memcpy (rqptr->AuthRealmDescriptionPtr = cptr,
              plptr->RealmDescriptionPtr, plptr->RealmDescriptionLength+1);
      cptr += plptr->RealmDescriptionLength+1;
   }
   else
      rqptr->AuthRealmDescriptionPtr = rqptr->AuthRealmPtr;

   memcpy (rqptr->AuthGroupWritePtr = cptr,
           plptr->GroupWritePtr, plptr->GroupWriteLength+1);
   cptr += plptr->GroupWriteLength+1;
   rqptr->AuthSourceGroupWrite = plptr->SourceGroupWrite;
   rqptr->AuthGroupWriteVmsIdentifier = plptr->GroupWriteVmsIdentifier;

   memcpy (rqptr->AuthGroupReadPtr = cptr,
           plptr->GroupReadPtr, plptr->GroupReadLength+1);
   cptr += plptr->GroupReadLength+1;
   rqptr->AuthSourceGroupRead = plptr->SourceGroupRead;
   rqptr->AuthGroupReadVmsIdentifier = plptr->GroupReadVmsIdentifier;

   rqptr->AuthGroupCan = plptr->AuthGroupCan;
   memcpy (rqptr->AuthGroupRestrictListPtr = cptr,
           plptr->GroupRestrictListPtr, plptr->GroupRestrictListLength+1);
   cptr += plptr->GroupRestrictListLength+1;

   rqptr->AuthWorldCan = plptr->AuthWorldCan;
   memcpy (rqptr->AuthWorldRestrictListPtr = cptr,
           plptr->WorldRestrictListPtr, plptr->WorldRestrictListLength+1);
   rqptr->AuthWorldRestrictListPtr = plptr->WorldRestrictListPtr;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
This function just wraps the reporting function, loading a temporary database
if necessary for reporting from the configuration file.
*/ 

AuthPathReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
boolean  UseServerDatabase,
char *VirtualHostNamePort
)
{
   int  status;
   struct AuthConfigStruct  AuthLoad;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthPathReport()\n");

   if (UseServerDatabase)
      AuthPathReportNow (rqptr, UseServerDatabase, &AuthServerConfig,
                         VirtualHostNamePort);
   else
   {
      memset (&AuthLoad, 0, sizeof(struct AuthConfigStruct)); 
      /* indicate it's being used for a report */
      AuthLoad.RequestPtr = rqptr;
      AuthPathConfig (&AuthLoad);
      AuthPathReportNow (rqptr, UseServerDatabase, &AuthLoad,
                         VirtualHostNamePort);
      AuthPathListFree (&AuthLoad);
      if (AuthLoad.ProblemReportPtr != NULL)
         VmFree (AuthLoad.ProblemReportPtr);
   }

   NetWriteBuffered (rqptr, NextTaskFunction, NULL, 0);
}

/*****************************************************************************/
/*
Display all records in the authorization path linked-list. Called from the
admin report.
*/

AuthPathReportNow
(
struct RequestStruct *rqptr,
boolean UseServerDatabase,
struct AuthConfigStruct *acptr,
char *VirtualHostNamePort
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... Path Authorization</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>Path Authorization</H2>\n\
!AZ!AZ!AZ!AZ\
!AZ\n");

   static $DESCRIPTOR (ProblemReportFaoDsc,
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH><FONT COLOR=\"#ff0000\">!UL Report!AZ !AZ</FONT></TH></TR>\n\
<TR><TD><PRE>!AZ</PRE></TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (VirtualFormFaoDsc,
"<FORM ACTION=\"!AZ\">\n\
<INPUT TYPE=hidden NAME=server VALUE=\"!AZ\">\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Virtual Server</TH></TR>\n\
<TR><TD><NOBR>\n\
<SELECT NAME=virtual>\n");

   static $DESCRIPTOR (VirtualOptionFaoDsc, "<OPTION!AZ!AZ>!AZ\n");

   static $DESCRIPTOR (SourceFaoDsc,
"</SELECT>\n\
<INPUT TYPE=submit VALUE=\" select \">\n\
</NOBR>\n\
</TD></TR>\n\
</TABLE>\n\
</FORM>\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Source: &quot;!AZ&quot;</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR>\
<TH ALIGN=RIGHT>File:</TH>\
<TD ALIGN=LEFT>!AZ</TD>\
<TD>&nbsp;</TD><TH>[<A HREF=\"!AZ\">View</A>]</TH>\
</TR>\n\
<TR><TH ALIGN=RIGHT>!AZ</TH>\
<TD ALIGN=LEFT>!AZ</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR>\
<TH>Path</TH><TH>Service</TH><TH>Realm</TH>\
<TH>Group-R+W</TH><TH>Group-R</TH>\
<TH>Group Access</TH><TH>World Access</TH>\
</TR>\n");

   static $DESCRIPTOR (PathFaoDsc,
"<TR>\
<TD><TT>!AZ</TT></TD><TD>!AZ</TD><TD>!AZ!AZ!AZ!AZ!AZ</TD>\
<TD>!AZ!AZ</TD><TD>!AZ!AZ</TD><TD>!AZ</TD><TD>!AZ</TD>\
</TR>\n\
!AZ!AZ!AZ!AZ!AZ!AZ!AZ!AZ");

   static $DESCRIPTOR (NoneFaoDsc,
"<TR ALIGN=CENTER><TD COLSPAN=7><I>(none)</I></TD></TR>\n");

   static $DESCRIPTOR (EndPageFaoDsc,
"</TABLE>\n\
</BODY>\n\
</HTML>\n");

   register unsigned long  *vecptr;
   register struct AuthPathRecordStruct  *plptr, *pplptr;
   register struct ServiceStruct *svptr;

   int  status,
        EntryCount;
   unsigned long  FaoVector [32];
   unsigned short  Length;
   char  *CanStringPtr;
   char  Buffer [2048],
         GroupCanString [128],
         WorldCanString [128];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "AuthPathReportNow() |%s|\n", VirtualHostNamePort);

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;
   if (VirtualHostNamePort[0])
   {
      *vecptr++ = "<H3>";
      *vecptr++ = "Virtual Server: ";
      *vecptr++ = VirtualHostNamePort;
      *vecptr++ = "</H3>\n";
   }
   else
   {
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = "";
      *vecptr++ = "";
   }
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   if (acptr->ProblemReportLength)
   {
      vecptr = FaoVector;
      if ((*vecptr++ = acptr->ProblemCount) == 1)
         *vecptr++ = "";
      else
         *vecptr++ = "s";
      if (acptr == &AuthServerConfig)
         *vecptr++ = "At Startup";
      else
         *vecptr++ = "During Load";
      *vecptr++ = acptr->ProblemReportPtr;

      status = sys$faol (&ProblemReportFaoDsc, &Length, &BufferDsc,
                         &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, FI_LI);
         return;
      }

      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   vecptr = FaoVector;
   *vecptr++ = INTERNAL_REPORT_AUTH_PATHS;
   if (UseServerDatabase)
      *vecptr++ = "yes";
   else
      *vecptr++ = "no";

   status = sys$faol (&VirtualFormFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   vecptr = FaoVector;
   if (VirtualHostNamePort[0])
      *vecptr++ = "";
   else
      *vecptr++ = " SELECTED";
   *vecptr++ = " VALUE=\"\"";
   if (ServiceCount == 1)
      *vecptr++ = "none configured";
   else
      *vecptr++ = "for all services";
   Length = 0;
   status = sys$faol (&VirtualOptionFaoDsc, &Length, &BufferDsc, &FaoVector);
   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);
   if (ServiceCount > 1)
   {
      for (svptr = ServiceListHead; svptr != NULL; svptr = svptr->NextPtr)
      {
         vecptr = FaoVector;
         if (strsame (svptr->ServerHostPort, VirtualHostNamePort, -1))
            *vecptr++ = " SELECTED";
         else
            *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = svptr->ServerHostPort;
         Length = 0;
         status = sys$faol (&VirtualOptionFaoDsc, &Length, &BufferDsc,
                            &FaoVector);
         Buffer[Length] = '\0';
         NetWriteBuffered (rqptr, 0, Buffer, Length);
      }
   }

   vecptr = FaoVector;
   if (UseServerDatabase)
      *vecptr++ = "Server";
   else
      *vecptr++ = "File";
   *vecptr++ = acptr->LoadFileName;
   *vecptr++ = MapVmsPath (acptr->LoadFileName, rqptr);
   if (acptr == &AuthServerConfig)
   {
      *vecptr++ = "Loaded:";
      *vecptr++ = DayDateTime (&acptr->LoadBinTime, 20);
   }
   else
   {
      *vecptr++ = "Revised:";
      *vecptr++ = DayDateTime (&acptr->RevBinTime, 20);
   }

   status = sys$faol (&SourceFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   pplptr = NULL;
   EntryCount = 0;

   for (plptr = acptr->PathListHead; plptr != NULL; plptr = plptr->NextPtr)
   {
      if (Debug)
         fprintf (stdout,
"plptr |%s|%s|%s|%s|%s|%s||%08.08X|%08.08X|\n",
            plptr->VirtualServerHostPortPtr, 
            plptr->PathPtr, plptr->RealmPtr, plptr->RealmDescriptionPtr,
            plptr->GroupWritePtr, plptr->GroupReadPtr,
            plptr->AuthGroupCan, plptr->AuthWorldCan);

      if (VirtualHostNamePort[0])
         if (plptr->VirtualServerHostPortPtr[0] != '*')
            if (strcmp (VirtualHostNamePort, plptr->VirtualServerHostPortPtr))
               continue;

      EntryCount++;

      if ((CanStringPtr =
           AuthCanString (plptr->AuthGroupCan, AUTH_CAN_FORMAT_HTML)) == NULL)
         return;
      strcpy (GroupCanString, CanStringPtr);

      if ((CanStringPtr =
           AuthCanString (plptr->AuthWorldCan, AUTH_CAN_FORMAT_HTML)) == NULL)
         return;
      strcpy (WorldCanString, CanStringPtr);

      vecptr = FaoVector;

      if (pplptr == NULL)
         *vecptr++ = plptr->PathPtr;
      else
      if (strsame (plptr->PathPtr, pplptr->PathPtr, -1))
         *vecptr++ = "";
      else
         *vecptr++ = plptr->PathPtr;
      pplptr = plptr;

      if (*plptr->VirtualServerHostPortPtr == '*')
         *vecptr++ = "";
      else
         *vecptr++ = plptr->VirtualServerHostPortPtr;

      if (plptr->RealmDescriptionPtr[0])
      {
         *vecptr++ = "\"";
         *vecptr++ = plptr->RealmDescriptionPtr;
         *vecptr++ = "\"=";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      *vecptr++ = plptr->RealmPtr;
      *vecptr++ = AuthSourceString (plptr->RealmPtr, plptr->SourceRealm);

      *vecptr++ = plptr->GroupWritePtr;
      *vecptr++ = AuthSourceString (plptr->GroupWritePtr,
                                    plptr->SourceGroupWrite);

      *vecptr++ = plptr->GroupReadPtr;
      *vecptr++ = AuthSourceString (plptr->GroupReadPtr,
                                    plptr->SourceGroupRead);

      *vecptr++ = GroupCanString;
      *vecptr++ = WorldCanString;

      if (plptr->GroupRestrictListPtr[0])
      {
         *vecptr++ = "<TR><TD></TD><TD></TD><TD COLSPAN=5><FONT SIZE=-1><I>";
         if (plptr->GroupWritePtr[0])
            *vecptr++ = "group";
         else
            *vecptr++ = "realm";
         *vecptr++ = " restricted to:</I>&nbsp;&nbsp;";
         *vecptr++ = plptr->GroupRestrictListPtr;
         *vecptr++ = "</FONT></TD></TR>\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      if (plptr->WorldRestrictListPtr[0])
      {
         *vecptr++ =
"<TR><TD></TD><TD></TD><TD COLSPAN=5>\
<FONT SIZE=-1><I>world restricted to:</I>&nbsp;&nbsp;";
         *vecptr++ = plptr->WorldRestrictListPtr;
         *vecptr++ = "</FONT></TD></TR>\n";
      }
      else
      {
         *vecptr++ = "";
         *vecptr++ = "";
         *vecptr++ = "";
      }

      status = sys$faol (&PathFaoDsc, &Length, &BufferDsc, &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$fao()";
         ErrorVmsStatus (rqptr, status, FI_LI);
         return;
      }

      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   if (!EntryCount)
      NetWriteBuffered (rqptr, 0, NoneFaoDsc.dsc$a_pointer,
                                  NoneFaoDsc.dsc$w_length);

   NetWriteBuffered (rqptr, 0, EndPageFaoDsc.dsc$a_pointer,
                              EndPageFaoDsc.dsc$w_length);
}

/*****************************************************************************/
/*
Display all records in the authentication linked-list, binary tree.  Called
by the Control.c or the Request.c module.  For HTTP report, uses blocking
network write.
*/

AuthCacheTreeReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... User Authentication</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>User Authentication</H2>\n\
!AZ\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Since</TH></TR>\n\
<TR><TD>!AZ</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\\
<TR>\
<TH>Realm</TH><TH>Group-R+W</TH><TH>Group-R</TH><TH>User</TH><TH>Access</TH>\
<TH COLSPAN=2>Scheme</TH>\
<TH>Most Recent</TH>\
<TH COLSPAN=2>Authenticated</TH>\
</TR>\n\
<TR>\
<TH COLSPAN=5></TH>\
<TH>Basic</TH><TH>Digest</TH>\
<TH></TH>\
<TH>yes/src</TH><TH>NO</TH>\
</TR>\n");

   static $DESCRIPTOR (NoneFaoDsc,
"<TR ALIGN=CENTER><TD COLSPAN=9><I>(none)</I></TD></TR>\n");

   static $DESCRIPTOR (EndPageFaoDsc,
"</TABLE>\n\
</BODY>\n\
</HTML>\n");

   register unsigned long  *vecptr;
   int  status;
   unsigned long  FaoVector [32];
   unsigned short  Length;
   char  Buffer [2048],
         TimeCurrent [32];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCacheTreeReport()\n");

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;
   *vecptr++ = strcpy (TimeCurrent, DayDateTime (&rqptr->BinaryTime, 20));
   *vecptr++ = DayDateTime (&AuthCacheTreeBinTime, 20);

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$fao()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   AuthDisplayRecordCount = 0;
   status = lib$traverse_tree (&AuthCacheTreeHead,
                               &AuthCacheTreeReportRecord,
                               rqptr);
   if (Debug) fprintf (stdout, "lib$traverse_tree() %%X%08.08X\n", status);


   if (!AuthDisplayRecordCount)
      NetWriteBuffered (rqptr, 0, NoneFaoDsc.dsc$a_pointer,
                                 NoneFaoDsc.dsc$w_length);

   NetWriteBuffered (rqptr, NextTaskFunction,
                    EndPageFaoDsc.dsc$a_pointer,
                    EndPageFaoDsc.dsc$w_length);
}

/*****************************************************************************/
/*
Called by AuthCacheTreeReport().  Display the details of all records in the
linked-list binary tree.  Output the details via blocking network write.
*/

AuthCacheTreeReportRecord
(                      
struct AuthCacheRecordStruct *acrptr,
struct RequestStruct *rqptr
)
{
   static $DESCRIPTOR (ReportFaoDsc,
"<TR>\
<TD>!AZ!AZ</TD><TD>!AZ!AZ</TD><TD>!AZ!AZ</TD>\
<TD>!AZ</TD><TD>!AZ!AZ</TD><TD>!UL</TD><TD>!UL</TD>\
<TD>!17%D</TD>\
<TD>!UL&nbsp;/&nbsp;!UL</TD><TD>!UL</TD>\
</TR>\n");

   register unsigned long  *vecptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [32];
   char  *UserCanStringPtr;
   char  Buffer [256];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCacheTreeReportRecord()\n");

   AuthDisplayRecordCount++;

   if ((UserCanStringPtr =
       AuthCanString (acrptr->AuthUserCan, AUTH_CAN_FORMAT_HTML)) == NULL)
      return (STS$K_ERROR);

   vecptr = FaoVector;

   *vecptr++ = acrptr->Realm;
   *vecptr++ = AuthSourceString (acrptr->Realm, acrptr->SourceRealm);

   *vecptr++ = acrptr->GroupWrite;
   *vecptr++ = AuthSourceString (acrptr->GroupWrite, acrptr->SourceGroupWrite);

   *vecptr++ = acrptr->GroupRead;
   *vecptr++ = AuthSourceString (acrptr->GroupRead, acrptr->SourceGroupRead);

   *vecptr++ = acrptr->UserName;
   *vecptr++ = UserCanStringPtr;
   if (acrptr->HttpsOnly)
      *vecptr++ = " (&quot;https:&quot;&nbsp;only)";
   else
      *vecptr++ = "";

   *vecptr++ = acrptr->BasicCount;
   *vecptr++ = acrptr->DigestCount;
   *vecptr++ = &acrptr->LastAccessBinTime;
   *vecptr++ = acrptr->AccessCount;
   *vecptr++ = acrptr->DataBaseCount;
   *vecptr++ = acrptr->FailureCount;

   status = sys$faol (&ReportFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return (status);
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Set string text according to capability vector in the in-memory authorization
information.  These may be different to the bits in the on-disk database,
reported by HTAdminCanString().
*/

char* AuthCanString
(
unsigned long CanFlags,
int Format
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthCanString()\n");

   if ((CanFlags & HTTP_METHOD_DELETE ||
        CanFlags & HTTP_METHOD_POST ||
        CanFlags & HTTP_METHOD_PUT) &&
       CanFlags & HTTP_METHOD_GET)
   {
      if (Format == AUTH_CAN_FORMAT_HTML)
         return ("read <B>+ write</B>");
      else
      if (Format == AUTH_CAN_FORMAT_CGI)
         return ("READ+WRITE");
      else
         return ("R+W");
   }
   else
   if ((CanFlags & HTTP_METHOD_DELETE ||
        CanFlags & HTTP_METHOD_POST ||
        CanFlags & HTTP_METHOD_PUT))
   {
      if (Format == AUTH_CAN_FORMAT_HTML)
         return ("write-only");
      else
      if (Format == AUTH_CAN_FORMAT_CGI)
         return ("WRITE");
      else
         return ("W");
   }
   else
   if (CanFlags & HTTP_METHOD_GET)
   {
      if (Format == AUTH_CAN_FORMAT_HTML)
         return ("read-only");
      else
      if (Format == AUTH_CAN_FORMAT_CGI)
         return ("READ");
      else
         return ("R");
   }
   else
   {
      if (Format == AUTH_CAN_FORMAT_HTML)
         return ("<I>none!</I>");
      else
      if (Format == AUTH_CAN_FORMAT_CGI)
         return ("");
      else
         return ("-");
   }
   return ("*ERROR*");
}

/*****************************************************************************/
/*
*/ 

char* AuthSourceString
(
char *Name,
unsigned long Value
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthSourceString() |%s| %d\n", Name, Value);

   switch (Value)
   {
      case AUTH_SOURCE_EXTERNAL    : return ("");
      case AUTH_SOURCE_FAIL        : return ("*PROBLEM*");
      case AUTH_SOURCE_HTA         : return ("=hta");
      case AUTH_SOURCE_ID          : return ("=id");
      case AUTH_SOURCE_LIST        : return ("=list");
      case AUTH_SOURCE_NONE        : return ("");
      case AUTH_SOURCE_PROMISCUOUS : return ("");
      case AUTH_SOURCE_VMS :
         if (strsame (Name, AUTH_REALM_VMS, -1)) return ("");
         return ("=vms");
      case AUTH_SOURCE_WASD_ID     : return ("=wasd_id");
      case AUTH_SOURCE_WORLD       : return ("");
      default :
         if (Name[0]) return ("=?");
         return ("");
   }
}

/*****************************************************************************/
/*
*/

AuthControlReload
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>Success 200</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>SUCCESS!!</H1>\n\
<P>Reported by server.\n\
<P>Server !AZ authorization paths reloaded.\n\
</BODY>\n\
</HTML>\n");

   static $DESCRIPTOR (ProblemReportFaoDsc,
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH><FONT COLOR=\"#ff0000\">!UL Report!AZ During Load</FONT></TH></TR>\n\
<TR><TD><PRE>!AZ</PRE></TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (EndPageFaoDsc,
"</BODY>\n\
</HTML>\n");

   register unsigned long  *vecptr;
   register struct AuthPathRecordStruct  *plptr;

   int  status;
   unsigned long  FaoVector [16];
   unsigned short  Length;
   char  Buffer [2048];
   struct AuthConfigStruct  *acptr;
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthControlReload()\n");

   AuthInit ();

   acptr = &AuthServerConfig;

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   if (acptr->ProblemReportLength)
   {
      vecptr = FaoVector;
      if ((*vecptr++ = acptr->ProblemCount) == 1)
         *vecptr++ = "";
      else
         *vecptr++ = "s";
      *vecptr++ = acptr->ProblemReportPtr;

      status = sys$faol (&ProblemReportFaoDsc, &Length, &BufferDsc,
                         &FaoVector);

      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, FI_LI);
         SysDclAst (NextTaskFunction, rqptr);
         return;
      }

      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   NetWriteBuffered (rqptr, NextTaskFunction,
                    EndPageFaoDsc.dsc$a_pointer,
                    EndPageFaoDsc.dsc$w_length);
}

/*****************************************************************************/

